/*++

	Copyright (c) 2009-2010 SMSC. All rights reserved.

	The information contained in this file is confidential and proprietary to
	Standard Microsystems Corporation.	No part of this file may be reproduced
	or distributed, in any form or by any means for any purpose, without the
	express written permission of Standard Microsystems Corporation.


Module Name:

	init.c

Abstract:

	This module contains initialization helper routines called during
	MiniportInitialize.
Revision History:
	01/19/2010	Initial version.

Notes:

--*/

#include "smsc7500.h"

/*++

Routine Description:

	Initialization for the Adapter Context.
 
Arguments:
   
	Adapter 					Pointer to our adapter
	
Return Value:

	NDIS_STATUS_SUCCESS
	NDIS_STATUS_FAILURE
	NDIS_STATUS_RESOURCES										

--*/  
NDIS_STATUS
NICAllocAdapter(
  IN PMP_ADAPTER pAdapter)

{
	PMP_ADAPTER Adapter = pAdapter;
	NDIS_STATUS Status=NDIS_STATUS_SUCCESS;

	SMSC_TRACE0(DBG_INIT,"--> NICAllocAdapter\r\n");

	do
	{		 
		//
		// This event is used to make sure all the asynchronously running routines
		// such as DPCs, workitems, threads have exited before returning from halt
		// handler.
		//
		NdisInitializeEvent(&Adapter->HaltEvent);

			 
		//
		// Workitem used to check link
		//
		NdisInitializeWorkItem(&Adapter->LinkCheckWorkItem,
								NICLinkCheckWorkItemCallBack,
								(PVOID)Adapter);

		
		//
		// Workitem used for USB Error Recovery
		//
		NdisInitializeWorkItem(&Adapter->UsbErrorRecoveryWorkItem,
								NICUsbErrorRecoveryWorkItemCallBack,
								(PVOID)Adapter);


		//
		// Workitem used to service OID requests.
		//
		NdisInitializeWorkItem(&Adapter->WdmNdisRequest.WorkItem,
								NICRequestWorkItemCallback,
								(PVOID)Adapter);

		NdisInitializeEvent(&Adapter->LinkCheckCalledEvent);
		
		NdisAllocateSpinLock(&Adapter->LCSpinlock);

		NdisAllocateSpinLock(&Adapter->RequestLock);

	} while(FALSE);

	//
	// In the failure case, the caller of this routine will end up
	// calling NICFreeAdapter to free all the successfully allocated
	// resources.
	//
	SMSC_TRACE0(DBG_INIT,"<-- NICAllocAdapter\r\n");

	return(Status);

}


/*++

Routine Description:

	Frees resources of our Adapter Context structure.
 
Arguments:

	Adapter    Pointer to our adapter
	
Return Value:

	VOID
	
--*/	
VOID
NICFreeAdapter(
	PMP_ADAPTER Adapter)
{
	SMSC_TRACE0(DBG_SHUT_DOWN, "--> NICFreeAdapter\r\n");

	ASSERT(Adapter);

	ASSERT(!Adapter->RefCount);

	NdisFreeSpinLock(&Adapter->RequestLock);

	NdisFreeSpinLock(&Adapter->LCSpinlock);

    NdisFreeEvent(&Adapter->HaltEvent);
    NdisFreeEvent(&Adapter->LinkCheckCalledEvent);

	SMSC_TRACE0(DBG_SHUT_DOWN, "<-- NICFreeAdapter\r\n");
}


/*++

Routine Description:

	This function gets the device's operational speed (high or full)

Arguments:

	Adapter 		Pointer to our adapter

Return Value:

	VOID

--*/
void
USBDeviceOperationalSpeed(
	IN PMP_ADAPTER		Adapter
	)

{
 
	SMSC_TRACE0(DBG_INIT,"--> USBDeviceOperationalSpeed\r\n");


	// Here's a quick and dirty way of doing it based on the bulk
	// endpoint packet size
	switch (Adapter->BulkOut.wMaxPacketSize)
	{
		case 512:
			SMSC_TRACE0(DBG_INIT,"Device is Operating in USB High Speed!\r\n");
			Adapter->UsbSpeedIsHigh = TRUE;
			break;
		case 64:
			SMSC_TRACE0(DBG_INIT,"Device is Operating in USB Full Speed!\r\n");
			Adapter->UsbSpeedIsHigh = FALSE;
			break;
		default:
			SMSC_TRACE0(DBG_INIT,"Could not determine USB Operating speed!\r\n");
			Adapter->UsbSpeedIsHigh = FALSE;
			break;
	}


	SMSC_TRACE0(DBG_INIT,"<-- USBDeviceOperationalSpeed\r\n");
}


/*++
Routine Description:

	Read device configuration parameters from the registry

Arguments:

	Adapter 						Pointer to our adapter
	WrapperConfigurationContext 	For use by NdisOpenConfiguration

   

Return Value:

	NDIS_STATUS_SUCCESS
	NDIS_STATUS_FAILURE
	NDIS_STATUS_RESOURCES

--*/
NDIS_STATUS
NICReadRegParameters(
	PMP_ADAPTER Adapter,
	NDIS_HANDLE WrapperConfigurationContext)

{
	
	typedef struct {
					NDIS_STRING Name;		// registry item name
					PULONG pValue;			// location to store the value
					NDIS_PARAMETER_TYPE Type;
				   } REGISTRY_ENTRY, *PREGISTRY_ENTRY;

	
	REGISTRY_ENTRY RegistryEntries[] =
	{
		{ NDIS_STRING_CONST("IdReg"),			&Adapter->IdReg 						, NdisParameterInteger },
		{ NDIS_STRING_CONST("MediaType"),		&Adapter->MediaType 					, NdisParameterInteger },
		{ NDIS_STRING_CONST("CustomCapa"),		&Adapter->AdvertisedCapabilites 		, NdisParameterInteger },
		{ NDIS_STRING_CONST("AutoMdix"),		&Adapter->AutoMdix						, NdisParameterInteger },
		{ NDIS_STRING_CONST("SleepPhyOnLinkDown"), &Adapter->SleepPhyOnLinkDown 	  	, NdisParameterInteger },
		{ NDIS_STRING_CONST("LinkWakeModes"),          &Adapter->LinkWakeModes        	, NdisParameterInteger }, 
    	{ NDIS_STRING_CONST("DetachOnLinkDownMode"),   &Adapter->DetachOnLinkDownMode 	, NdisParameterInteger },
    	{ NDIS_STRING_CONST("SleepModeLinkSpeed"),      &Adapter->SleepModeLinkSpeed	, NdisParameterInteger },
    	{ NDIS_STRING_CONST("SDReconnectGpio"),        &Adapter->SDReconnectGpio    	, NdisParameterInteger },
    	{ NDIS_STRING_CONST("SDReconnectGpioPol"),     &Adapter->SDReconnectGpioPol 	, NdisParameterInteger },
		{ NDIS_STRING_CONST("PowerManagement"), &Adapter->PowerManagement				, NdisParameterInteger },
		{ NDIS_STRING_CONST("PhyAddress"),		&Adapter->PHYAddress					, NdisParameterInteger },
		{ NDIS_STRING_CONST("DelayTxComplete"), &Adapter->DelayTxComplete				, NdisParameterInteger },
		{ NDIS_STRING_CONST("WolPMSettings"),	&Adapter->WolPatternMatchSettings		, NdisParameterInteger },
		{ NDIS_STRING_CONST("WolMPSettings"),	&Adapter->WolMagicPacketSettings		, NdisParameterInteger },
		{ NDIS_STRING_CONST("WoLCSettings"),	&Adapter->WoLinkChgSettings 			, NdisParameterInteger },
		{ NDIS_STRING_CONST("NumOfWolPatterns"),&Adapter->NumOfWolPatterns				, NdisParameterInteger },
		{ NDIS_STRING_CONST("FlowControl"), 	&Adapter->FlowControl					, NdisParameterInteger },
		{ NDIS_STRING_CONST("FCTFlowReg"),		&Adapter->FCTFlowReg 					, NdisParameterInteger },
		{ NDIS_STRING_CONST("FCPause"), 		&Adapter->FCPause						, NdisParameterInteger },
		{ NDIS_STRING_CONST("WakeAutoClear"),	&Adapter->WakeAutoClear 				, NdisParameterInteger },
		{ NDIS_STRING_CONST("StopBulkInOnRxEmpty"),&Adapter->StopBulkInOnRxEmpty		, NdisParameterInteger },
		{ NDIS_STRING_CONST("MaxICBs"), 		&Adapter->MaxICBs						, NdisParameterInteger },
		{ NDIS_STRING_CONST("LowWatermarkICBs"),&Adapter->LowWatermarkICBs				, NdisParameterInteger },
		{ NDIS_STRING_CONST("MaxTCBs"), 		&Adapter->MaxTCBs						, NdisParameterInteger },
		{ NDIS_STRING_CONST("MaxRCBs"), 		&Adapter->MaxRCBs						, NdisParameterInteger },
		{ NDIS_STRING_CONST("MaxRxPkts"),		&Adapter->MaxRxPkts 					, NdisParameterInteger },
		{ NDIS_STRING_CONST("MaxFrameSize"),    &Adapter->MaxFrameSize            		, NdisParameterInteger },
    	{ NDIS_STRING_CONST("TxFifoSize"),      &Adapter->TxFifoSize              		, NdisParameterInteger },
    	{ NDIS_STRING_CONST("RxFifoSize"),      &Adapter->RxFifoSize 					, NdisParameterInteger },	
		{ NDIS_STRING_CONST("LowWatermarkRCBs"),&Adapter->LowWatermarkRCBs				, NdisParameterInteger },
		{ NDIS_STRING_CONST("TxOnePktperTcb"),	&Adapter->TxOnePktperTcb				, NdisParameterInteger },
		{ NDIS_STRING_CONST("TxBatching"),		&Adapter->TxBatching					, NdisParameterInteger },
		{ NDIS_STRING_CONST("TxBurstMax"),		&Adapter->TxBurstMax					, NdisParameterInteger },
		{ NDIS_STRING_CONST("RxOnePktperRcb"),	&Adapter->RxOnePktperRcb				, NdisParameterInteger },
		{ NDIS_STRING_CONST("RxBurstMax"),		&Adapter->RxBurstMax					, NdisParameterInteger },
		{ NDIS_STRING_CONST("RxBurstCap"),		&Adapter->RxBurstCap					, NdisParameterInteger },
		{ NDIS_STRING_CONST("MEFBulkInDelay"),	&Adapter->MEFBulkInDelay				, NdisParameterInteger },
		{ NDIS_STRING_CONST("LinkActLedCfg"),		 &Adapter->LinkActLedCfg			, NdisParameterInteger },
		{ NDIS_STRING_CONST("PHYDriveStrength"),&Adapter->PHYDriveStrength				, NdisParameterInteger },
		{ NDIS_STRING_CONST("VlanID"),			&Adapter->VlanID						, NdisParameterInteger },
		{ NDIS_STRING_CONST("PriorityVlanTag"),	&Adapter->PriorityVlanTag				, NdisParameterInteger },
		{ NDIS_STRING_CONST("TcpTaskOffload"),   &Adapter->TcpTaskOffload            	, NdisParameterInteger },
		{ NDIS_STRING_CONST("IpCsoCap"),         &Adapter->IpCsoCap                   	, NdisParameterInteger },
    	{ NDIS_STRING_CONST("MaxFrameSize"),   &Adapter->MaxFrameSize 					, NdisParameterInteger },
    	{ NDIS_STRING_CONST("TcpCsoCap"),               &Adapter->TcpCsoCap	            , NdisParameterInteger },
    	{ NDIS_STRING_CONST("TCPChecksumOffloadIPv4"),  &Adapter->TCPChecksumOffloadIPv4, NdisParameterInteger },
    	{ NDIS_STRING_CONST("TCPChecksumOffloadIPv6"),  &Adapter->TCPChecksumOffloadIPv6, NdisParameterInteger },
    	{ NDIS_STRING_CONST("UdpCsoCap"),              &Adapter->UdpCsoCap            	, NdisParameterInteger },
    	{ NDIS_STRING_CONST("UDPChecksumOffloadIPv4"),  &Adapter->UDPChecksumOffloadIPv4, NdisParameterInteger },
    	{ NDIS_STRING_CONST("UDPChecksumOffloadIPv6"),  &Adapter->UDPChecksumOffloadIPv6, NdisParameterInteger },
    	{ NDIS_STRING_CONST("LsoV1IPv4"),               &Adapter->LsoV1IPv4             , NdisParameterInteger },	
		{ NDIS_STRING_CONST(""),				NULL								, 0UL } // must be the last item
	};


	NDIS_STATUS 	Status = NDIS_STATUS_SUCCESS;
	NDIS_HANDLE 	ConfigurationHandle;
	PUCHAR			NetworkAddress;
	PREGISTRY_ENTRY pEntry;
	UINT			Length;
	PNDIS_CONFIGURATION_PARAMETER pParameterValue;
	NDIS_STRING 	strMiniportName = NDIS_STRING_CONST("MiniportName");

	SMSC_TRACE0(DBG_INIT,"--> NICReadRegParameters\r\n");

	
	Adapter->PHYAddress 		= 0x01;
	Adapter->MediaType = MEDIA_TYPE_AUTO_FULL;

	Adapter->StopBulkInOnRxEmpty = FALSE;	// Default to low latency mode.
  	Adapter->MEFBulkInDelay = ETH_BULK_IN_DLY_AUTO;
	Adapter->TxBatching = TRUE;
	Adapter->TxOnePktperTcb=FALSE;
	Adapter->RxOnePktperRcb=FALSE;
	Adapter->DelayTxComplete=FALSE;

	Adapter->MaxICBs = NIC_MAX_BUSY_INTS_DEFAULT;
	Adapter->LowWatermarkICBs = NIC_MIN_VAL_FOR_INTS_LOW_WM	;
	Adapter->MaxTCBs = NIC_MAX_BUSY_SENDS_DEFAULT;
	Adapter->MaxRCBs = NIC_MAX_BUSY_RECVS_DEFAULT;
	Adapter->LowWatermarkRCBs=NIC_MIN_VAL_FOR_RECVS_LOW_WM;
	Adapter->MaxRxPkts=32;
	
	Adapter->TxBurstMax = Adapter->UsbSpeedIsHigh ? BUFFER_SIZE_16KB : BUFFER_SIZE_10KB;
	Adapter->RxBurstCap = Adapter->UsbSpeedIsHigh ? BUFFER_SIZE_16KB : BUFFER_SIZE_10KB; 	   
	Adapter->RxBurstMax = Adapter->UsbSpeedIsHigh ? BUFFER_SIZE_16KB : BUFFER_SIZE_10KB; 

//	WakeAutoClear:			- Controls if/how Rev B wake auto clearing logic is setup:
//								+ WakeAutoClear = 0: Neither wake status or enable bits are auto cleared 
//								  after resume signaling (Rev A behavior)
//								+ WakeAutoClear = 1: Wake Enable bits are auto cleared after resume signaling.
//								+ WakeAutoClear = 2: Wake Status bits are auto cleared after resume signaling.
//								+ WakeAutoClear = 3: Wake Enable and Status bits are auto cleared after resume 
//													 signaling.
//							  Defaults to WakeAutoClear = 2.

	// Power Management Defaults
	//0 DISABLED, 1 ENABLED, 2 FORCED
	Adapter->PowerManagement		= ENABLED;
	Adapter->WolPatternMatchSettings			= ENABLED;
	Adapter->WolMagicPacketSettings			= ENABLED;
	Adapter->WoLinkChgSettings			= ENABLED;

	
	Adapter->WakeAutoClear				= WAKE_CLEAR_STS;
	Adapter->FlowControl				= FC_SYMM;
	Adapter->FCTFlowReg=0;
	Adapter->FCPause=0;
	
	

				
    Adapter->NumOfWolPatterns=8;		//by default is ff, Driver will auto detect it. 

	//AMDIX_DISABLE_STRAIGHT          0x00000000UL
	//AMDIX_DISABLE_CROSSOVER         0x00000001UL
	//AMDIX_ENABLE                    0x00000002UL
	//AMDIX_USE_STRAP                 0x00000003UL
    Adapter->AutoMdix=3;

	
	Adapter->LinkWakeModes=LINK_WAKE_MODE_DEFAULT;
	Adapter->DetachOnLinkDownMode=DOLD_ENABLED;
	Adapter->SDReconnectGpio=12;
	Adapter->SDReconnectGpioPol=DRIVEN_LOW;

	Adapter->SleepModeLinkSpeed=SLEEP_MODE_AUTO_10_100;

    Adapter->SleepPhyOnLinkDown=0;	//Put PHY to EDPD(Energy Detect Power Down) Mode when the Link is down

    Adapter->PHYDriveStrength=0xff;         
    Adapter->LinkActLedCfg=0;            

//    Adapter->LinkLedOnGpio=0xc;      	//Software control LED(0-10)
//    Adapter->LinkLedOnGpioBufType=0;
//    Adapter->LinkLedOnGpioPolarity=0;

    Adapter->VlanID=0;			//Vlan Tag ID
    Adapter->PriorityVlanTag=0;

	Adapter->TxFifoSize=MAX_TX_FIFO_SIZE;
	Adapter->RxFifoSize=MAX_RX_FIFO_SIZE;
	
// Setup Tx/Rx FIFO End Value
    Adapter->TxFifoEnd = (Adapter->TxFifoSize - 512UL)/512UL;
    Adapter->RxFifoEnd = (Adapter->RxFifoSize - 512UL)/512UL;

//OFFLOAD_DISABLE         0
//TX_OFFLOAD_EN           1
//RX_OFFLOAD_EN           2
//RXTX_OFFLOAD_EN         3

	Adapter->TcpTaskOffload=RXTX_OFFLOAD_EN;
	Adapter->IpCsoCap=RXTX_OFFLOAD_EN;			
    Adapter->TcpCsoCap=RXTX_OFFLOAD_EN;	            
    Adapter->TCPChecksumOffloadIPv4=RXTX_OFFLOAD_EN;
    Adapter->TCPChecksumOffloadIPv6=RXTX_OFFLOAD_EN;
    Adapter->UdpCsoCap=RXTX_OFFLOAD_EN;            	
    Adapter->UDPChecksumOffloadIPv4=RXTX_OFFLOAD_EN;
    Adapter->UDPChecksumOffloadIPv6=RXTX_OFFLOAD_EN;
    Adapter->LsoV1IPv4=1;             
	Adapter->MaxFrameSize=MAX_802_3_FRAME_SIZE; 

	Adapter->AnegCompleteTiomeout=10;
	Adapter->MpRstCompleteTimeout=10;
	//
	// Open the registry for this adapter to read advanced
	// configuration parameters stored by the INF file.
	//
	NdisOpenConfiguration(
		&Status,
		&ConfigurationHandle,
		WrapperConfigurationContext);
	if(Status != NDIS_STATUS_SUCCESS)
	{
		SMSC_TRACE0(DBG_ERROR,"SMSC: NdisOpenConfiguration failed\r\n");
		return NDIS_STATUS_FAILURE;
	}

	// Read registry entries
	pEntry = RegistryEntries;
	while(pEntry->pValue)
	{
		SMSC_TRACE1(DBG_INIT,"%s = ", pEntry->Name.Buffer);

		NdisReadConfiguration(
								&Status,
								&pParameterValue,
								ConfigurationHandle,
								&pEntry->Name,
								pEntry->Type
							 );

		if(Status == NDIS_STATUS_SUCCESS)
		{
			*pEntry->pValue = pParameterValue->ParameterData.IntegerData;
			#ifdef SMSC_DBG
			if (g_DebugMode & DBG_INIT)
				RETAILMSG(1,(TEXT("0x%08x (read from registry)\r\n"), *pEntry->pValue));
			#endif
		 }
		 else
		{
			#ifdef SMSC_DBG
			if (g_DebugMode & DBG_INIT)
				RETAILMSG(1,(TEXT("0x%08x (not found in registry, using default)\r\n"), *pEntry->pValue));
			#endif
		}

		pEntry++;
	} // while
	
	
	   // Adjust interdependent values	
	   // Synchronizing Standard Keywords with Advanced Properties Parameters if they are not disable
	   if (Adapter->IPChecksumOffloadIPv4 == RXTX_OFFLOAD_EN) // Standard keyword is default
	   {
		   if (Adapter->IpCsoCap != RXTX_OFFLOAD_EN)
		   {
			   SMSC_TRACE0(DBG_INIT,"Overriding Standard Keyword (*IPChecksumOffloadIPv4) with Advanced Properties Parameter \r\n");
			   Adapter->IPChecksumOffloadIPv4 = Adapter->IpCsoCap;
		   }
	   }
	   
	   if (Adapter->TCPChecksumOffloadIPv4 == RXTX_OFFLOAD_EN) // Standard keyword is default
	   {
		   if (Adapter->TcpCsoCap != RXTX_OFFLOAD_EN)
		   {
			   SMSC_TRACE0(DBG_INIT,"Overriding Standard Keyword (*TCPChecksumOffloadIPv4) with Advanced Properties Parameter \r\n");
			   Adapter->TCPChecksumOffloadIPv4 = Adapter->TcpCsoCap;
		   }
	   }
	
	   if (Adapter->TCPChecksumOffloadIPv6 == RXTX_OFFLOAD_EN) // Standard keyword is default
	   {
		   if (Adapter->TcpCsoCap != RXTX_OFFLOAD_EN)
		   {
			   SMSC_TRACE0(DBG_INIT,"Overriding Standard Keyword (*TCPChecksumOffloadIPv6) with Advanced Properties Parameter \r\n");
			   Adapter->TCPChecksumOffloadIPv6 = Adapter->TcpCsoCap;
		   }
	   }
	
	   if (Adapter->UDPChecksumOffloadIPv4 == RXTX_OFFLOAD_EN) // Standard keyword is default
	   {
		   if (Adapter->UdpCsoCap != RXTX_OFFLOAD_EN)
		   {
			   SMSC_TRACE0(DBG_INIT,"Overriding Standard Keyword (*UDPChecksumOffloadIPv4) with Advanced Properties Parameter \r\n");
			   Adapter->UDPChecksumOffloadIPv4 = Adapter->UdpCsoCap;
		   }
	   }
	
	   if (Adapter->UDPChecksumOffloadIPv6 == RXTX_OFFLOAD_EN) // Standard keyword is default
	   {
		   if (Adapter->UdpCsoCap != RXTX_OFFLOAD_EN)
		   {
			   SMSC_TRACE0(DBG_INIT,"Overriding Standard Keyword (*UDPChecksumOffloadIPv6) with Advanced Properties Parameter \r\n");
			   Adapter->UDPChecksumOffloadIPv6 = Adapter->UdpCsoCap;
		   }
	   }
	
	   // Set up Tx/Rx Task offload using registry values
	   Adapter->TxTaskOffload = 0;
	   if (Adapter->IPChecksumOffloadIPv4 & TX_OFFLOAD_EN)
	   {
		   Adapter->TxTaskOffload |= IPV4_CSUM_IP_OPT_EN_;
		   Adapter->TxTaskOffload |= IPV4_CSUM_IP_CSUM_EN_;
	   }
	
	   if (Adapter->TCPChecksumOffloadIPv4 & TX_OFFLOAD_EN)
	   {
		   Adapter->TxTaskOffload |= IPV4_CSUM_TCP_OPT_EN_;
		   Adapter->TxTaskOffload |= IPV4_CSUM_TCP_CSUM_EN_;
	   }
	
	   if (Adapter->UDPChecksumOffloadIPv4 & TX_OFFLOAD_EN)
	   {
		   Adapter->TxTaskOffload |= IPV4_CSUM_UDP_CSUM_EN_;
	   }
	   
	   if (Adapter->TxTaskOffload & (IPV4_CSUM_UDP_CSUM_EN_ | IPV4_CSUM_TCP_CSUM_EN_ | IPV4_CSUM_IP_CSUM_EN_))
	   {
		   Adapter->TxTaskOffload |= IPV4_CSUM_EN_;
	   }
	
	   if (Adapter->TCPChecksumOffloadIPv6 & TX_OFFLOAD_EN)
	   {
		   Adapter->TxTaskOffload |= IPV6_CSUM_IP_EXT_EN_;
		   Adapter->TxTaskOffload |= IPV6_CSUM_TCP_OPT_EN_;
		   Adapter->TxTaskOffload |= IPV6_CSUM_TCP_CSUM_EN_;
	   }
	
	   if (Adapter->UDPChecksumOffloadIPv6 & TX_OFFLOAD_EN)
	   {
		   Adapter->TxTaskOffload |= IPV6_CSUM_IP_EXT_EN_;
		   Adapter->TxTaskOffload |= IPV6_CSUM_UDP_CSUM_EN_;
	   }
	   
	   if (Adapter->TxTaskOffload & (IPV6_CSUM_UDP_CSUM_EN_ | IPV6_CSUM_TCP_CSUM_EN_))
	   {
		   Adapter->TxTaskOffload |= IPV6_CSUM_EN_;
	   }
	
	   if (Adapter->LsoV1IPv4)
	   {
		   Adapter->TxTaskOffload |= IPV4_LSOV1_EN_;
		   Adapter->TxTaskOffload |= IPV4_LSOV1_TCP_OPT_EN_;
		   Adapter->TxTaskOffload |= IPV4_LSOV1_IP_OPT_EN_;
	   }
	
	   Adapter->RxTaskOffload = 0;
	   if (Adapter->IPChecksumOffloadIPv4 & RX_OFFLOAD_EN)
	   {
		   Adapter->RxTaskOffload |= IPV4_CSUM_IP_OPT_EN_;
		   Adapter->RxTaskOffload |= IPV4_CSUM_IP_CSUM_EN_;
	   }
	
	   if (Adapter->TCPChecksumOffloadIPv4 & RX_OFFLOAD_EN)
	   {
		   Adapter->RxTaskOffload |= IPV4_CSUM_TCP_OPT_EN_;
		   Adapter->RxTaskOffload |= IPV4_CSUM_TCP_CSUM_EN_;
	   }
	
	   if (Adapter->UDPChecksumOffloadIPv4 & RX_OFFLOAD_EN)
	   {
		   Adapter->RxTaskOffload |= IPV4_CSUM_UDP_CSUM_EN_;
	   }
	   
	   if (Adapter->RxTaskOffload & (IPV4_CSUM_UDP_CSUM_EN_ | IPV4_CSUM_TCP_CSUM_EN_ | IPV4_CSUM_IP_CSUM_EN_))
	   {
		   Adapter->RxTaskOffload |= IPV4_CSUM_EN_;
	   }
	
	   if (Adapter->TCPChecksumOffloadIPv6 & RX_OFFLOAD_EN)
	   {
		   Adapter->RxTaskOffload |= IPV6_CSUM_IP_EXT_EN_;
		   Adapter->RxTaskOffload |= IPV6_CSUM_TCP_OPT_EN_;
		   Adapter->RxTaskOffload |= IPV6_CSUM_TCP_CSUM_EN_;
	   }
	
	   if (Adapter->UDPChecksumOffloadIPv6 & RX_OFFLOAD_EN)
	   {
		   Adapter->RxTaskOffload |= IPV6_CSUM_IP_EXT_EN_;
		   Adapter->RxTaskOffload |= IPV6_CSUM_UDP_CSUM_EN_;
	   }
	   
	   if (Adapter->RxTaskOffload & (IPV6_CSUM_UDP_CSUM_EN_ | IPV6_CSUM_TCP_CSUM_EN_))
	   {
		   Adapter->RxTaskOffload |= IPV6_CSUM_EN_;
	   }


	// If our configuration descriptor claims we don't support remote
	// wakeup, override our registry selections and claim support for
	// NO WAKEUPS at all.
	// Same if we made a registry overide to disable all PM support
	// Also print warnings if the registry had selected otherwise.
	if (!(Adapter->UsbDeviceSupportsWakes) || (Adapter->PowerManagement == DISABLED))
	{
		// If the descriptor says we don't support it and we have one enabled, warn !
		if (
			(Adapter->WolPatternMatchSettings != DISABLED) ||
			(Adapter->WolMagicPacketSettings != DISABLED) ||
			(Adapter->WoLinkChgSettings != DISABLED)
		   )
		{
			if (!(Adapter->UsbDeviceSupportsWakes))
			{
				SMSC_WARNING0("The device is not configed to support remote wakes , Over write to disable remote wakes\r\n");
			}

			if (Adapter->PowerManagement == DISABLED)
			{
				SMSC_WARNING0("The PowerManagement Parameter is disabled , Over write to disable remote wakes\r\n");
			}
		}

		Adapter->WolPatternMatchSettings	= DISABLED;
		Adapter->WolMagicPacketSettings 	= DISABLED;
		Adapter->WoLinkChgSettings			= DISABLED;
		Adapter->WakeAutoClear = WAKE_CLEAR_DISABLED;
	}


	// Verify Wake Clear settings
	if (Adapter->WakeAutoClear > WAKE_CLEAR_EN_AND_STS)
	{
		Adapter->WakeAutoClear = WAKE_CLEAR_STS;
		SMSC_TRACE1(DBG_ERROR, "Invalid WakeAutoClear value. Overriding to %d (WAKE_CLEAR_STS)\r\n", Adapter->WakeAutoClear);
	}

	// Verify Flow control settings are within valid range
	if (Adapter->FlowControl > FC_ASYM_RESP)
	{
		SMSC_TRACE0(DBG_INIT,"Invalid Flow Control value. Overriding to 1 (FC_SYMM)\r\n");
		Adapter->FlowControl = FC_SYMM;
	}


	 // If generation of pause frames is enabled setup FCTFlowReg and pause.
    if ((Adapter->FlowControl == FC_SYMM) || (Adapter->FlowControl == FC_ASYM_GEN))
    {
        if (Adapter->FCTFlowReg == 0)
        {
            DWORD dwWaterMark;
            
            dwWaterMark = (((Adapter->RxFifoSize * 2)/(10 * 512)) & 0x7FUL);  // Set Hi Water Mark to 20% of the Rx Fifo Size (Used Space)
            dwWaterMark <<= 8UL;
            dwWaterMark |= (((Adapter->RxFifoSize * 8)/(10 * 512)) & 0x7FUL); // Set Lo Water Mark to 80% of the Rx Fifo Size (Used Space)
            Adapter->FCTFlowReg = dwWaterMark;
            SMSC_TRACE1(DBG_INIT,"Overiding FCT_FLOW Reg Value (was 0) to 0x%08lx\r\n", Adapter->FCTFlowReg);
        }

        if (Adapter->FCPause == 0)
        {
            Adapter->FCPause = 0xFFFF;  // Use the max, since we'll send a 0 to overide it when we don't 
                                        // need it anymore
        }
    }
    else
    {
        if (Adapter->FCTFlowReg != 0)
        {
            SMSC_TRACE1(DBG_INIT,"Overiding FCT_FLOW Reg Value (was 0x%08lx) to 0 since FlowControl pause generation feature is disabled\r\n", Adapter->FCTFlowReg);
            Adapter->FCTFlowReg = 0;
        }
    }

	
	if (Adapter->MaxRCBs < NIC_MIN_VAL_FOR_MAX_BUSY_RECVS)
	{
		SMSC_WARNING2("MaxRCBs = < %d. Overiding to %d\r\n",
							NIC_MIN_VAL_FOR_MAX_BUSY_RECVS,
							NIC_MIN_VAL_FOR_MAX_BUSY_RECVS
							);
		Adapter->MaxRCBs = NIC_MIN_VAL_FOR_MAX_BUSY_RECVS;
	}

	if (Adapter->MaxICBs < NIC_MIN_VAL_FOR_MAX_BUSY_INTS)
	{
		SMSC_WARNING2("MaxICBs = < %d. Overiding to %d\r\n",
							NIC_MIN_VAL_FOR_MAX_BUSY_INTS,
							NIC_MIN_VAL_FOR_MAX_BUSY_INTS
							);
		Adapter->MaxICBs = NIC_MIN_VAL_FOR_MAX_BUSY_INTS;
	}

	if (Adapter->MaxTCBs < NIC_MIN_VAL_FOR_MAX_BUSY_SENDS)
	{
		SMSC_WARNING2("MaxTCBs = < %d. Overiding to %d\r\n",
							NIC_MIN_VAL_FOR_MAX_BUSY_SENDS,
							NIC_MIN_VAL_FOR_MAX_BUSY_SENDS
							);
		Adapter->MaxTCBs = NIC_MIN_VAL_FOR_MAX_BUSY_SENDS;
	}

	// Setup RCB low Watermarks
	if (Adapter->LowWatermarkRCBs == 0)
	{
		Adapter->LowWatermarkRCBs = Adapter->MaxRCBs / 4;
		if (Adapter->LowWatermarkRCBs < NIC_MIN_VAL_FOR_RECVS_LOW_WM)
			Adapter->LowWatermarkRCBs = NIC_MIN_VAL_FOR_RECVS_LOW_WM;
	}

	// Setup ICB low Watermarks
	if (Adapter->LowWatermarkICBs == 0)
	{
		Adapter->LowWatermarkICBs = Adapter->MaxICBs / 4;
		if (Adapter->LowWatermarkICBs < NIC_MIN_VAL_FOR_INTS_LOW_WM)
			Adapter->LowWatermarkICBs = NIC_MIN_VAL_FOR_INTS_LOW_WM;
	}
/*
	// Validate single / multi pkt mode parameters and overide them if bad.
    if ((Adapter->LsoV1IPv4) &&
        (Adapter->TxBurstMax < (64*1024)))
    {
        SMSC_TRACE3(DBG_INIT,"LSOv1 is enable but TxBurstMax = %d < %d. Overiding to %d\r\n", Adapter->TxBurstMax, (64*1024), (64*1024));
        Adapter->TxBurstMax = (64*1024);
    }
*/
	 // Validate single / multi pkt mode parameters and overide them if bad.
    if (Adapter->TxBurstMax < NIC_BUFFER_SIZE)
    {
        SMSC_TRACE3(DBG_INIT,"TxBurstMax = %d < NIC_BUFFER_SIZE =%d. Overiding to %d\n", Adapter->TxBurstMax, NIC_BUFFER_SIZE, NIC_BUFFER_SIZE);
        Adapter->TxBurstMax = NIC_BUFFER_SIZE;
    }

	if (Adapter->TxOnePktperTcb)
	{
		 // Single Eth-Tx packets per each Bulk Out transfer
        if (Adapter->TxBurstMax > NIC_BUFFER_SIZE)
        {
            SMSC_TRACE3(DBG_INIT,"TxOnePktperTcb = TRUE, but TxBurstMax = %d > NIC_BUFFER_SIZE = %d is a waste. Overiding TxBurstMax = %d\r\n", Adapter->TxBurstMax, NIC_BUFFER_SIZE, NIC_BUFFER_SIZE);
            Adapter->TxBurstMax = NIC_BUFFER_SIZE;
        }

		if (Adapter->TxBatching)
		{
			SMSC_WARNING0("TxOnePktperTcb = TRUE, but TxBatching = TRUE is incompatible. Overiding TxBatching = FALSE\r\n");
			Adapter->TxBatching = FALSE;
		}
	}
	else
	{
		// Multiple Eth-Tx packets per each Bulk Out transfer
		if (Adapter->DelayTxComplete)
		{
			SMSC_WARNING0("DelayTxComplete = TRUE but TxOnePktperTcb = FALSE. Multipacket incompatible with Delayed completions. Overiding DelayTxComplete = FALSE\r\n");
			Adapter->DelayTxComplete = FALSE;
		}
	}


	 if (Adapter->TxOnePktperTcb)
    {
        if (!(Adapter->LsoV1IPv4))
        {
            // Single Eth-Tx packets per each Bulk Out transfer
            if (Adapter->TxBurstMax > NIC_BUFFER_SIZE)
            {
                SMSC_TRACE3(DBG_INIT,"TxOnePktperTcb = TRUE, but TxBurstMax = %d > NIC_BUFFER_SIZE = %d is a waste. Overiding TxBurstMax = %d\n", Adapter->TxBurstMax, NIC_BUFFER_SIZE, NIC_BUFFER_SIZE);
                Adapter->TxBurstMax = NIC_BUFFER_SIZE;
            }
        }
    } 

	if (Adapter->RxBurstMax < NIC_BUFFER_SIZE)
    {
        SMSC_TRACE3(DBG_INIT,"RxBurstMax = %d < NIC_BUFFER_SIZE =%d. Overiding to %d\r\n", Adapter->RxBurstMax, NIC_BUFFER_SIZE, NIC_BUFFER_SIZE);
        Adapter->RxBurstMax = NIC_BUFFER_SIZE;
    }

	// If device does not support MEF, we can only do single packet receives.
	if ((!NICSupportsMEF(Adapter)) && (!Adapter->RxOnePktperRcb))
	{
		SMSC_WARNING0("Driver set to multipacket Rx but device does not support MEF so overriding RxOnePktperRcb to TRUE\r\n");
		Adapter->RxOnePktperRcb = TRUE;
	}

	if (Adapter->RxOnePktperRcb)
	{
		// Multipkt mode is disabled, therefore do not allocate Split Rx Pkts and turn off Burst Cap
		if (Adapter->MaxRxPkts)
		{
			SMSC_WARNING1("MaxRxPkts = %d, but multipkt mode is disabled so overiding to 0\r\n", Adapter->MaxRxPkts);
			Adapter->MaxRxPkts = 0;
		}

		if (Adapter->RxBurstCap)
		{
			SMSC_WARNING1("RxBurstCap = %d, but multipkt mode is disabled so overiding to 0\r\n", Adapter->RxBurstCap);
			Adapter->RxBurstCap = 0;
		}

		 if (Adapter->RxBurstMax > NIC_BUFFER_SIZE)
        {
            SMSC_TRACE3(DBG_INIT,"RxBurstMax = %d > %d, but multipkt mode is disabled so it's a waste. Overiding to %d\r\n", 
            Adapter->RxBurstMax, NIC_BUFFER_SIZE, NIC_BUFFER_SIZE);
            Adapter->RxBurstMax = NIC_BUFFER_SIZE;
       	 }
	}
	else
	{
		// Multipkt mode is enabled, lets see if params are OK
		if (Adapter->UsbSpeedIsHigh)
            {
                // High speed
                if (Adapter->RxBurstCap > (255 * 512))
                {
                    SMSC_TRACE2(DBG_INIT,"Invalid RxBurstCap = %d > 130560 (255 * 512). Overiding to RxBurstCap = %d\r\n", 
                                        Adapter->RxBurstCap,
                                        255 * 512);
                    Adapter->RxBurstCap = 255 * 512;
                }

                if (Adapter->RxBurstCap < Adapter->MaxFrameSize)
                {
                    SMSC_TRACE3(DBG_INIT,"RxBurstCap = %d < MaxFrameSize = %d, results untoward operation. Overiding to RxBurstCap = %d\r\n", 
                                    Adapter->RxBurstCap, Adapter->MaxFrameSize, BUFFER_SIZE_16KB);

                    Adapter->RxBurstCap = BUFFER_SIZE_16KB;
                }

                // Do we have enough URB buffer size for this Burst Cap size ?
                if (Adapter->RxBurstMax < Adapter->RxBurstCap)
                {
                    SMSC_TRACE3(DBG_INIT,"RxBurstMax = %d < %d = RxBurstCap. Overriding to RxBurstMax = %d\r\n", 
                                        Adapter->RxBurstMax, 
                                        Adapter->RxBurstCap, 
                                        Adapter->RxBurstCap );
                
                    Adapter->RxBurstMax = Adapter->RxBurstCap;
                }
            }
            else
            {
                // Full speed
                if (Adapter->RxBurstCap > (255 * 64))
                {
                    SMSC_TRACE2(DBG_INIT,"Invalid RxBurstCap = %d > 16320 (255 * 64). Overiding to RxBurstCap = %d\r\n", 
                                        Adapter->RxBurstCap,
                                        255 * 64);
                    Adapter->RxBurstCap = 255 * 64;
                }

                if (Adapter->RxBurstCap < Adapter->MaxFrameSize)
                {
                    SMSC_TRACE3(DBG_INIT,"RxBurstCap = %d < MaxFrameSize = %d, results untoward operation. Overiding to RxBurstCap = %d\r\n", 
                                    Adapter->RxBurstCap, Adapter->MaxFrameSize, BUFFER_SIZE_10KB);

                    Adapter->RxBurstCap = BUFFER_SIZE_10KB;
                }

                // Do we have enough URB buffer size for this Burst Cap size ?
                if (Adapter->RxBurstMax < Adapter->RxBurstCap)
                {
                    SMSC_TRACE3(DBG_INIT,"RxBurstMax = %d < %d = RxBurstCap. Overriding to RxBurstMax = %d\r\n", 
                                        Adapter->RxBurstMax, 
                                        Adapter->RxBurstCap, 
                                        Adapter->RxBurstCap);
                
                    Adapter->RxBurstMax = Adapter->RxBurstCap;
                }
            }

		// Do we have enough Rx Split packets ?
		if (Adapter->MaxRxPkts < Adapter->MaxRCBs)
		{
			SMSC_WARNING3("MaxRxPkts = %d < %d = MaxRCBs. Overriding to MaxRxPkts = %d\r\n", Adapter->MaxRxPkts, Adapter->MaxRCBs, 4 * Adapter->MaxRCBs);
			Adapter->MaxRxPkts = 8 * Adapter->MaxRCBs;
		}
	}

	
	// Auto MEF delay setup ?
	   if (Adapter->MEFBulkInDelay == ETH_BULK_IN_DLY_AUTO)
	   {
		   // Lets setup the best value for each case 
		   if (Adapter->MaxRCBs == 1)
		   {
			   Adapter->MEFBulkInDelay = ETH_BULK_IN_DLY_DEFAULT_1_RCB;
		   }
		   else
		   {
			   Adapter->MEFBulkInDelay = ETH_BULK_IN_DLY_DEFAULT_N_RCBS;
		   }
	   }

/*
	// Validate NumOfWolPatterns parameter and overide them if bad.
	if (!((Adapter->NumOfWolPatterns == 0xff) ||
		  (Adapter->NumOfWolPatterns == 4)	  ||
		  (Adapter->NumOfWolPatterns == 8)))
	{
		SMSC_TRACE1(DBG_INIT,"NumOfWolPatterns = %d (wrong value). Overiding to 0xff\r\n", Adapter->NumOfWolPatterns);
		// Override to default autodetect mode
		Adapter->NumOfWolPatterns = 0xff;
	}
*/
	// Display parameters that may have changed
	SMSC_TRACE0(DBG_INIT,"Parameter values after checks/overides:\r\n");
	SMSC_TRACE1(DBG_INIT,"TxFifoSize           = %d\r\n", Adapter->TxFifoSize);
    SMSC_TRACE1(DBG_INIT,"RxFifoSize           = %d\r\n", Adapter->RxFifoSize);	
	SMSC_TRACE1(DBG_INIT,"MaxTCBs			   = %d\r\n", Adapter->MaxTCBs);
	SMSC_TRACE2(DBG_INIT,"MaxRCBs			   = %d, LowWatermarkRCBs = %d\r\n", Adapter->MaxRCBs, Adapter->LowWatermarkRCBs);
	SMSC_TRACE2(DBG_INIT,"MaxICBs			   = %d, LowWatermarkICBs = %d\r\n", Adapter->MaxICBs, Adapter->LowWatermarkICBs);
	SMSC_TRACE1(DBG_INIT,"DelayTxComplete	   = %d\r\n", Adapter->DelayTxComplete);
	SMSC_TRACE1(DBG_INIT,"TxBatching		   = %d\r\n", Adapter->TxBatching);
	SMSC_TRACE1(DBG_INIT,"TxBurstMax		   = %d\r\n", Adapter->TxBurstMax); 
	SMSC_TRACE1(DBG_INIT,"StopBulkInOnRxEmpty  = %d\r\n", Adapter->StopBulkInOnRxEmpty);	
	SMSC_TRACE1(DBG_INIT,"MaxRxPkts 		   = %d\r\n", Adapter->MaxRxPkts);
	SMSC_TRACE1(DBG_INIT,"RxBurstMax		   = %d\r\n", Adapter->RxBurstMax);
	SMSC_TRACE1(DBG_INIT,"RxBurstCap		   = %d\r\n", Adapter->RxBurstCap);
	SMSC_TRACE1(DBG_INIT,"TxTaskOffload        = %08lX\r\n", Adapter->TxTaskOffload);
    SMSC_TRACE1(DBG_INIT,"RxTaskOffload        = %08lX\r\n", Adapter->RxTaskOffload);
    SMSC_TRACE1(DBG_INIT,"SleepPhyOnLinkDown        = %d\r\n", Adapter->SleepPhyOnLinkDown);
    SMSC_TRACE1(DBG_INIT,"DetachOnLinkDownMode        = %d\r\n", Adapter->DetachOnLinkDownMode);
    
	//
	// Read the Miniport Name.
	//
	NdisReadConfiguration(&Status,
						  &pParameterValue,
						  ConfigurationHandle,
						  &strMiniportName,
						  NdisParameterString);

	if (Status != NDIS_STATUS_SUCCESS) {

		SMSC_TRACE0(DBG_ERROR,"SMSC: NdisReadConfiguration for miniport name failed\r\n");

	} else {
		Length = min(NIC_ADAPTER_NAME_SIZE-1,
		pParameterValue->ParameterData.StringData.Length/sizeof(WCHAR));

	}

	// Read a possible Node Address override
	NdisReadNetworkAddress(
							&Status,
							(PVOID *) &NetworkAddress,
							&Length,
							ConfigurationHandle
						  );

	if((Status == NDIS_STATUS_SUCCESS) && (Length == ETH_LENGTH_OF_ADDRESS))
	{
		
		SMSC_TRACE0(DBG_INIT,"Node Address Override !\r\n");
		SMSC_TRACE0(DBG_INIT,"Node Address read from registry: ");
		#ifdef SMSC_DBG
		if (g_DebugMode & DBG_INIT) {
			UINT i;
			for(i=0;i<6;i++)
			{
				RETAILMSG(1,(TEXT(" %02X"), NetworkAddress[i]));
			}
			RETAILMSG(1,(TEXT("\r\n")));
		}
		#endif

		if ((ETH_IS_MULTICAST(NetworkAddress) || ETH_IS_BROADCAST(NetworkAddress))
			|| !ETH_IS_LOCALLY_ADMINISTERED (NetworkAddress))
		{
			SMSC_TRACE0(DBG_INIT,"NetworkAddress is invalid, ignoring !\r\n");

		}
		else
		{
			ETH_COPY_NETWORK_ADDRESS(Adapter->CurrentNodeAddress, NetworkAddress);
		}
	}
	else {

			SMSC_TRACE0(DBG_INIT,"Fail to read netwrk Address !\r\n");

	}

	//
	// Close the configuration registry
	//
	NdisCloseConfiguration(ConfigurationHandle);

	SMSC_TRACE0(DBG_INIT,"<-- NICReadRegParameters\r\n");

	return NDIS_STATUS_SUCCESS;
}


/*++
Routine Description:

	Initialize the adapter.

Arguments:

	Adapter 	Pointer to our adapter
	WrapperConfigurationContext - Pointer to the configuration context
								  only valid during MiniportInitialize.
Return Value:

	NDIS_STATUS_SUCCESS
	NDIS_STATUS_ADAPTER_NOT_FOUND

--*/
NDIS_STATUS
NICInitializeAdapter(
	IN	PMP_ADAPTER  Adapter
	)
{
	NDIS_STATUS 		Status = NDIS_STATUS_SUCCESS;
	NTSTATUS			ntStatus;

	SMSC_TRACE0(DBG_INIT,"--> NICInitializeAdapter\r\n");

#ifdef NDIS50_MINIPORT
	//
	// Register a shutdown handler for NDIS50 or earlier miniports
	// For NDIS51 miniports, set AdapterShutdownHandler.
	//
	NdisMRegisterAdapterShutdownHandler(
		Adapter->AdapterHandle,
		(PVOID) Adapter,
		(ADAPTER_SHUTDOWN_HANDLER) MPShutdown);
#endif


	ntStatus = NICInitAdapterWorker(Adapter);

	NT_STATUS_TO_NDIS_STATUS(ntStatus, &Status);

	SMSC_TRACE1(DBG_INIT,"<-- NICInitializeAdapter, Status=%x\r\n", Status);

	return Status;

}


/*++
Routine Description:

	This is the worker routine for doing all the initialization work.
	It opens the target device, allocates send & receive side resources
	and starts the pipes.

Arguments:

	Adapter 	Pointer to our adapter

Return Value:

	NT Status code

--*/
NTSTATUS
NICInitAdapterWorker(
	PMP_ADAPTER Adapter
	)

{
	NTSTATUS			ntStatus = STATUS_SUCCESS;
	NDIS_STATUS 		Status;


	SMSC_TRACE1(DBG_INIT,"--> NICInitAdapterWorker %p\r\n", Adapter);

	do {

//		Adapter->TargetDeviceObject = Adapter->NextDeviceObject;
	
		//
		// Read Advanced configuration information from the registry
		// We do it here after USB Initialization since some default
		// values depend on USB speed.
		//

		Status = NICReadRegParameters(Adapter, Adapter->WrapperConfigurationContext);
		if(Status != NDIS_STATUS_SUCCESS)
		{
			SMSC_TRACE0(DBG_ERROR,"Failed while reading Registry params\r\n");
			break;
		}


		#ifdef USE_INT_PIPE
		Status = NICAllocIntResources(Adapter);
		if(Status != NDIS_STATUS_SUCCESS)
		{
			ntStatus = STATUS_INSUFFICIENT_RESOURCES;
			SMSC_TRACE0(DBG_ERROR,"Failed to allocate interrupt resources\r\n");
			break;
		}
		#endif

		Status = NICAllocSendResources(Adapter);
		if(Status != NDIS_STATUS_SUCCESS)
		{
			ntStatus = STATUS_INSUFFICIENT_RESOURCES;
			SMSC_TRACE0(DBG_ERROR,"Failed to allocate send side resources\r\n");
			break;
		}

		Status = NICAllocRecvResources(Adapter);
		if(Status != NDIS_STATUS_SUCCESS)
		{
			ntStatus = STATUS_INSUFFICIENT_RESOURCES;
			SMSC_TRACE0(DBG_ERROR,"Failed to allocate receive side resources\r\n");
			break;
		}


		if (Adapter->WolPatternMatchSettings != DISABLED)
		{
			Status = NICAllocWolResources(Adapter);
			if(Status != NDIS_STATUS_SUCCESS)
			{
				ntStatus = STATUS_INSUFFICIENT_RESOURCES;
				SMSC_TRACE0(DBG_ERROR,"Failed to allocate Wol resources\n");
				break;
			}
		}



		// Set the receive mask to disable all receives, NDIS will call us with
		// enabling value later.
		Adapter->PacketFilter = 0;
		
		 /* Reset the multicast table (16 DWORD entries) */
    	NdisZeroMemory((PVOID)Adapter->MulticastHashTable, (16*sizeof(DWORD)));
		
		// Determine the Number of WOL Patterns supported by this device
//		NICGetWOLPatternCount(Adapter);

		MP_SET_FLAG(Adapter, fMP_DISCONNECTED);
		SMSC_TRACE1(DBG_INIT,"after setting fMP_DISCONNECTED, Adapter->Flags = 0x%x\r\n", Adapter->Flags);

		Status = NICResetAdapter(Adapter);
		if(Status != NIC_STATUS_SUCCESS)
		{
			ntStatus = STATUS_INSUFFICIENT_RESOURCES;
			SMSC_TRACE0(DBG_ERROR,"Failed to Reset Adapter\r\n");
			break;
		}

		MP_CLEAR_FLAG(Adapter, fMP_INIT_IN_PROGRESS);	// So that pipes can start

		if (NIC_STATUS_FAILURE == NICStartTxPath(Adapter))
		{
			ntStatus = STATUS_UNSUCCESSFUL;
			break;
		}

		if (NIC_STATUS_FAILURE == NICStartRxPath(Adapter))
		{
			ntStatus = STATUS_UNSUCCESSFUL;
			break;
		}

		#ifdef USE_INT_PIPE
		if (Adapter->StopBulkInOnRxEmpty) {
			if (NIC_STATUS_FAILURE == NICStartInterruptPipe(Adapter))
			{
				ntStatus = STATUS_UNSUCCESSFUL;
				break;
			}
		}
		#endif

		// Put Link check in motion
		StartLinkCheck(Adapter);

		ntStatus =	NDIS_STATUS_SUCCESS;

	}while(FALSE);

	SMSC_TRACE1(DBG_INIT, "<-- NICInitAdapterWorker %x\r\n", ntStatus);

	return ntStatus;
}


/*++
Routine Description:

	Allocate resources required to Send packets to the device

Arguments:

	Adapter    Pointer to our adapter
	

Return Value:

	NDIS_STATUS_SUCCESS
	NDIS_STATUS_FAILURE
	NDIS_STATUS_RESOURCES

--*/
NDIS_STATUS
NICAllocSendResources(
	PMP_ADAPTER Adapter)

{
	NDIS_STATUS Status;
	LONG index;
	PUCHAR pTCBMem = NULL, pTCBData = NULL;
	NDIS_PHYSICAL_ADDRESS HighestAcceptableMax;

	SMSC_TRACE0(DBG_INIT,"--> NICAllocSendResources\r\n");

	// Allocate adapter context memory.
    	HighestAcceptableMax.LowPart = 0xFFFFFFFFUL;        // DWORD
    	HighestAcceptableMax.HighPart = -1L;                // LONG

	do {

		
		//
		// Workitem used to process the send TCB pending and/or waiting queues
		//
		NdisInitializeWorkItem(&Adapter->SendsWaitingWorkItem,
								NICPostSendsWaitingWorkItemCallBack,
								(PVOID)Adapter);
				
		NdisInitializeEvent(&Adapter->SendsWaitingWorkItemReady);
		
		//
		// Workitem used to process TCB completion queue.
		//
		NdisInitializeWorkItem(&Adapter->SendsCompletionWorkItem,
								SendsCompletionWorkItemCallBack,
								(PVOID)Adapter);
		NdisInitializeEvent(&Adapter->IsSendsCompletionWorkItemReady);
		
		//
		// Following are the lists to hold packets at different stages of processing.
		// SendPktWaitList	  - Original send packets waiting to be processed
		// SendTcbFreeList	  - TCBs available for send operation
		// SendTcbPendingList - TCBs waiting to be sent to USB
		// SendTcbBusyList	  - TCBs sent to USB
		// SendTcbcompletionList  -TCBs return from completion routine
		// SendLock is used to synchronize access to these lists.
		//
		NdisInitializeListHead(&Adapter->SendPktWaitList);
		NdisInitializeListHead(&Adapter->SendTcbFreeList);
		NdisInitializeListHead(&Adapter->SendTcbPendingList);
		NdisInitializeListHead(&Adapter->SendTcbBusyList);
		NdisInitializeListHead(&Adapter->SendTcbcompletionList);	
		NdisAllocateSpinLock(&Adapter->SendLock);

		//
		// Let us set the flag to indicate resources are allocated.
		// NICFreeSendResources will check this flag to determine whether there
		// is any resource that needs to be freed.
		//
		MP_SET_FLAG(Adapter, fMP_SEND_SIDE_RESOURCE_ALLOCATED);

		
		//
		// Allocate memory for all TCB's
		//
		Status = NdisAllocateMemory(
			&pTCBMem,
			sizeof(TCB) * Adapter->MaxTCBs,
			NDIS_ALLOC_FLAG,
			HighestAcceptableMax);

		if(Status != NDIS_STATUS_SUCCESS)
		{
			SMSC_TRACE0(DBG_ERROR,"Failed to allocate memory for TCB's\r\n");
			break;
		}

		NdisZeroMemory(pTCBMem, sizeof(TCB) * Adapter->MaxTCBs);
		Adapter->TCBMem = pTCBMem;

		SMSC_TRACE1(DBG_INIT,"TCB Base = 0x%p\r\n", Adapter->TCBMem);

		//
		// Allocate the data buffers for the TCBs
		//
		Status = NdisAllocateMemory(
			&pTCBData,
			Adapter->TxBurstMax * Adapter->MaxTCBs,
			NDIS_ALLOC_FLAG, 
			HighestAcceptableMax);

		if(Status != NDIS_STATUS_SUCCESS)
		{
			SMSC_TRACE0(DBG_ERROR,"Failed to allocate memory for TCB's data buffers\r\n");
			break;
		}
		NdisZeroMemory(pTCBData, Adapter->TxBurstMax * Adapter->MaxTCBs);
		Adapter->TCBData = pTCBData;

		SMSC_TRACE1(DBG_INIT,"TCB Data Buffers Base = 0x%p\r\n", Adapter->TCBData);

		//
		// Divide the TCBMem blob into TCBs and create a buffer
		// descriptor for the Data portion of the TCBs. 
		//
		for(index=0; index < Adapter->MaxTCBs; index++)
		{
			PTCB pTCB = (PTCB) pTCBMem;

			pTCB->Adapter = Adapter;
			pTCB->pData = pTCBData;
			pTCB->SyncEvent=CreateEvent( NULL, TRUE, FALSE, NULL);
			if ( !pTCB->SyncEvent ) {
					Status = NDIS_STATUS_FAILURE;
					break;
				}
/*
			ASSERT(Adapter->TargetDeviceObject);

			if(!Adapter->TargetDeviceObject) {
				Status = NDIS_STATUS_FAILURE;
				break;
			}
*/
			//
			// Insert TCB blocks into FreeList.
			//
			NdisInterlockedInsertTailList(
				&Adapter->SendTcbFreeList,
				&pTCB->List,
				&Adapter->SendLock);

			SMSC_TRACE1(DBG_INIT,"TCB 0x%p\r\n", pTCB);
			SMSC_TRACE1(DBG_INIT,"	Data Buffer: 0x%p\r\n", pTCBData);

			pTCBMem = pTCBMem + sizeof(TCB);
			pTCBData += Adapter->TxBurstMax;
		}
	} while(FALSE);

	SMSC_TRACE1(DBG_INIT,"<-- NICAllocSendResources %x\r\n", Status);

	return Status;
}


/*++
Routine Description:

  Free resources allocated for send operation

Arguments:

	Adapter 	Pointer to our adapter
	
Return Value:

--*/

VOID
NICFreeSendResources(
	PMP_ADAPTER Adapter)
{
	PTCB		   pTCB;

	SMSC_TRACE0(DBG_SHUT_DOWN,"--> NICFreeSendResources\r\n");

	//
	// Did we allocate any resources for send?
	//
	if(!MP_TEST_FLAG(Adapter, fMP_SEND_SIDE_RESOURCE_ALLOCATED)){
		return;
	}

	while(!IsListEmpty(&Adapter->SendTcbFreeList))
	{
		pTCB = (PTCB) NdisInterlockedRemoveHeadList(
					 &Adapter->SendTcbFreeList,
					 &Adapter->SendLock);
		if(!pTCB) {
			break;
		}

		if(pTCB->SyncEvent){
			if (!CloseHandle(pTCB->SyncEvent))
			SMSC_TRACE0(DBG_ERROR,"In NicFreeSendResources, close pTCB event fail!\r\n");
		}
	}

	if(Adapter->TCBData) {
		NdisFreeMemory(Adapter->TCBData, Adapter->TxBurstMax * Adapter->MaxTCBs, NDIS_ALLOC_FLAG);
		Adapter->TCBData = NULL;
	}

	if(Adapter->TCBMem) {
		NdisFreeMemory(Adapter->TCBMem, sizeof(TCB) * Adapter->MaxTCBs, NDIS_ALLOC_FLAG);
		Adapter->TCBMem = NULL;
	}

	#ifdef USE_THREAD
	if (Adapter->SendThreadEvent){
		if (!CloseHandle(Adapter->SendThreadEvent))
			SMSC_TRACE0(DBG_ERROR,"In NicFreeSendResources, close SendThreadEvent fail!\r\n");
	}
	#endif

    #ifdef USE_WORK_ITEM_NO_OUT
    NdisFreeEvent(&Adapter->SendWorkItemReady);
    #endif
	
	ASSERT(IsListEmpty(&Adapter->SendTcbFreeList));
	ASSERT(IsListEmpty(&Adapter->SendTcbcompletionList));
	ASSERT(IsListEmpty(&Adapter->SendPktWaitList));

    NdisFreeEvent(&Adapter->SendsWaitingWorkItemReady);
    NdisFreeEvent(&Adapter->IsSendsCompletionWorkItemReady);
	
	NdisFreeSpinLock(&Adapter->SendLock);

	MP_CLEAR_FLAG(Adapter, fMP_SEND_SIDE_RESOURCE_ALLOCATED);
	SMSC_TRACE0(DBG_SHUT_DOWN,"<-- NICFreeSendResources\r\n");

}

#ifdef USE_INT_PIPE

/*++
Routine Description:

	Allocate resources required to process interrupts from the USB

Arguments:

	Adapter - Pointer to our adapter

Return Value:

	NDIS_STATUS_SUCCESS
	NDIS_STATUS_FAILURE
	NDIS_STATUS_RESOURCES

--*/
NDIS_STATUS
NICAllocIntResources(
	PMP_ADAPTER Adapter)
{
	PUCHAR pICBMem = NULL;
	LONG index;
	NDIS_STATUS Status;
	NDIS_PHYSICAL_ADDRESS HighestAcceptableMax;

	SMSC_TRACE0(DBG_INIT, "--> NICAllocIntResources\r\n");

	// Allocate adapter context memory.
    	HighestAcceptableMax.LowPart = 0xFFFFFFFFUL;        // DWORD
    	HighestAcceptableMax.HighPart = -1L;                // LONG

	do{

		
		//
		// Workitem used to process interrupt status.
		//
		NdisInitializeWorkItem(&Adapter->IntStatWorkItem,
								NICIntStatWorkItemCallBack,
								(PVOID)Adapter);

		//
		// Workitem used to post interrupt requests.
		//
		NdisInitializeWorkItem(&Adapter->PostIntsWorkItem,
								NICPostIntsWorkItemCallBack,
								(PVOID)Adapter);

		//
		// Workitem used to process Ints completion queue.
		//
		NdisInitializeWorkItem(&Adapter->IntsCompletionWorkItem,
								IntsCompletionWorkItemCallBack,
							   (PVOID)Adapter);
		NdisInitializeEvent(&Adapter->IsIntsCompletionWorkItemReady);

		
		//
		// Following is the lists to hold int resources at different stages of processing.
		// IntFreeList		- ICBs available for posting to USB
		// IntBusyList		- ICBs posted to USB
		// IntProcessedList - ICBs that have been processed and are awaiting Freeing.
		// IntCompletionList   -ICBs return from completion routine
		// IntLock is used to synchronize access to these lists.
		//
		NdisInitializeListHead(&Adapter->IntFreeList);
		NdisInitializeListHead(&Adapter->IntBusyList);
		NdisInitializeListHead(&Adapter->IntProcessedList);
		NdisInitializeListHead(&Adapter->IntCompletionList);
		NdisAllocateSpinLock(&Adapter->IntLock);

		//
		// Let us set the flag to indicate that resources are allocated.
		// NICFreeIntResources will check this flag to determine weather there
		// is any resource that needs to be freed.
		//
		MP_SET_FLAG(Adapter, fMP_INT_RESOURCE_ALLOCATED);

		//
		// Allocate a block of memory for all ICB's
		//
		Status = NdisAllocateMemory(
			&pICBMem,
			sizeof(ICB) * Adapter->MaxICBs,
			NDIS_ALLOC_FLAG, 
			HighestAcceptableMax);

		if(Status != NDIS_STATUS_SUCCESS)
		{
			SMSC_TRACE0(DBG_ERROR,"In NicAlloIntResources, Failed to allocate memory for ICB's\r\n");
			break;
		}

		NdisZeroMemory(pICBMem, sizeof(ICB) * Adapter->MaxICBs);
		Adapter->ICBMem = pICBMem;

		SMSC_TRACE1(DBG_INT,"ICB Base = 0x%p\r\n", Adapter->ICBMem);

		//
		// Divide the ICBMem blob into ICBs.
		//
		for(index=0; index < Adapter->MaxICBs; index++)
		{
			PICB pICB = (PICB) pICBMem;

			//
			// Initialize the RCB structure.
			//
			pICB->pData = (PUCHAR) &pICB->Data[0];
			pICB->Adapter = Adapter;

			pICB->SyncEvent=CreateEvent( NULL, TRUE, FALSE, NULL);
			if ( !pICB->SyncEvent ) {
					Status = NDIS_STATUS_FAILURE;
					break;
				}
	  
			NdisInterlockedInsertTailList(
				&Adapter->IntFreeList,
				&pICB->List,
				&Adapter->IntLock);

			SMSC_TRACE1(DBG_INIT, "ICB 0x%p\r\n", pICB);
			SMSC_TRACE1(DBG_INIT, "   Data Buffer: 0x%p\r\n", pICB->pData);

			pICBMem = pICBMem + sizeof(ICB);

		}
	}while(FALSE);

	SMSC_TRACE1(DBG_INIT, "<-- NICAllocIntResources %x\r\n", Status);

	return Status;

}


/*++
Routine Description:

	Free resources allocated for interrupt operation

Arguments:

		Adapter - Pointer to our adapter	

Return Value:

	None

--*/
VOID
NICFreeIntResources(
	PMP_ADAPTER Adapter)
{
	PICB		   pICB;

	SMSC_TRACE0(DBG_SHUT_DOWN, "--> NICFreeIntResources\r\n");

	//
	// Free all the resources we allocated for interrupts.
	//
	if(!MP_TEST_FLAG(Adapter, fMP_INT_RESOURCE_ALLOCATED)){
		return;
	}

	while(!IsListEmpty(&Adapter->IntFreeList))
	{
		pICB = (PICB) NdisInterlockedRemoveHeadList(
					 &Adapter->IntFreeList,
					 &Adapter->IntLock);
		if(!pICB){
			break;
		}

		if(pICB->SyncEvent){
			if (!CloseHandle(pICB->SyncEvent))
			SMSC_TRACE0(DBG_ERROR,"In NicFreeIntResources, close pICB event fail!\r\n");
		}
	}

	if(Adapter->ICBMem)
	{
		NdisFreeMemory(Adapter->ICBMem, sizeof(ICB) * Adapter->MaxICBs, NDIS_ALLOC_FLAG);
	}

	#ifdef USE_THREAD
	if(Adapter->IntThreadEvent){
		if (!CloseHandle(Adapter->IntThreadEvent))
		SMSC_TRACE0(DBG_ERROR,"In NicFreeIntResources, close IntThreadEvent fail!\r\n");
	}
	#endif

    #ifdef USE_WORK_ITEM_NO_OUT
    NdisFreeEvent(&Adapter->IntsWorkItemReady);
    #endif

	ASSERT(IsListEmpty(&Adapter->IntFreeList));
	ASSERT(IsListEmpty(&Adapter->IntProcessedList));
	ASSERT(IsListEmpty(&Adapter->IntBusyList));
	ASSERT(IsListEmpty(&Adapter->IntCompletionList));

    NdisFreeEvent(&Adapter->IsIntsCompletionWorkItemReady);

	NdisFreeSpinLock(&Adapter->IntLock);

	MP_CLEAR_FLAG(Adapter, fMP_INT_RESOURCE_ALLOCATED);

   SMSC_TRACE0(DBG_SHUT_DOWN, "<-- NICFreeIntResources\r\n");
	return;
}
#endif


/*++
Routine Description:

	Allocate resources required to receive packets

Arguments:

	Adapter - Pointer to our adapter

	
Return Value:

	NDIS_STATUS_SUCCESS
	NDIS_STATUS_FAILURE
	NDIS_STATUS_RESOURCES

--*/
NDIS_STATUS
NICAllocRecvResources(
	PMP_ADAPTER Adapter)
{
	PUCHAR pRCBMem = NULL, pRCBData = NULL, pRxData = NULL;
	PNDIS_PACKET Packet = NULL;
	PNDIS_BUFFER Buffer = NULL;
	PRCB pRCB = NULL;
	LONG index;
	NDIS_STATUS Status;
	NDIS_PHYSICAL_ADDRESS HighestAcceptableMax;
	
	SMSC_TRACE0(DBG_INIT,"--> NICAllocRecvResources\r\n");

	// Allocate adapter context memory.
    	HighestAcceptableMax.LowPart = 0xFFFFFFFFUL;        // DWORD
    	HighestAcceptableMax.HighPart = -1L;                // LONG
	do{

		
		//
		// Workitem used to post read requests.
		//
		NdisInitializeWorkItem(&Adapter->ReadWorkItem,
								NICPostReadsWorkItemCallBack,
								(PVOID)Adapter);
		
		//
		// Event to ensure only one thread at a time is posting reads
		//
		NdisInitializeEvent(&Adapter->ReadWorkItemReady);
		
		//
		// Workitem used to resume Bulk In processing.
		//
		NdisInitializeWorkItem(&Adapter->BulkInResumeWorkItem,
								NICBulkInResumeWorkItemCallBack,
								(PVOID)Adapter);
		
		NdisInitializeEvent(&Adapter->BulkInResumeWorkItemReady);
		
		//
		// Workitem used to process Bulk in Completion queue.
		//
		NdisInitializeWorkItem(&Adapter->BulkInCompletionWorkItem,
								BulkInCompletionWorkItemCallBack,
								(PVOID)Adapter);
		
		NdisInitializeEvent(&Adapter->BulkInCompletionWorkItemReady);

		//
		// Following are the lists to hold USB RCBs at different stages of processing.
		// BulkInFreeList - RCBs available for posting to USB
		// BulkInBusyList - RCBs posted to USB
		// BulkInWaitList - RCBs with completed URBs waiting for
		//					Rx Packets to copy data into
		// BulkInProcessedList - RCBs that have been processed
		//						 (either bad rx or indicated to ndis)
		//						 and are awaiting Freeing.
		//BulkInCompletionList	 -RCBs return from completion routine
		// RCBProcessLock is used to synchronize access to these lists.
		//
		NdisInitializeListHead(&Adapter->BulkInFreeList);
		NdisInitializeListHead(&Adapter->BulkInBusyList);
		NdisInitializeListHead(&Adapter->BulkInCompletionList);
		NdisInitializeListHead(&Adapter->BulkInWaitList);
		NdisInitializeListHead(&Adapter->BulkInProcessedList);
		NdisAllocateSpinLock(&Adapter->BulkInLock);
		NdisAllocateSpinLock(&Adapter->RCBProcessLock);

		
		if (!Adapter->RxOnePktperRcb)
		{
			SMSC_TRACE0(DBG_INIT,"allocate memory for Multi Rx Pkt's data buffers\r\n");
			//
			// Following are the lists to hold Rx split Pkt RCBs at
			// different stages of processing:
			//
			// RxPktFreeList - Packets available for receive operation
			// RxPktBusyList - Packets posted  to the lower WDM stack
			// RxPktLock is used to synchronize access to these lists.
			//
			NdisInitializeListHead(&Adapter->RxPktFreeList);
			NdisInitializeListHead(&Adapter->RxPktBusyList);
			NdisAllocateSpinLock(&Adapter->RxPktLock);
		}


		//
		// Let us set the flag to indicate that resources are allocated.
		// NICFreeRecvResources will check this flag to determine whether there
		// is any resource that needs to be freed.
		//

		MP_SET_FLAG(Adapter, fMP_RECV_SIDE_RESOURCE_ALLOCATED);

		//
		// Allocate a huge block of memory for all RCB's (both the ones
		// used for URBs and the ones used only for Rx Pkts
		//
		Status = NdisAllocateMemory(
			&pRCBMem,
			sizeof(RCB) * (Adapter->MaxRCBs + Adapter->MaxRxPkts),
			NDIS_ALLOC_FLAG, 
			HighestAcceptableMax);

		if(Status != NDIS_STATUS_SUCCESS)
		{
			SMSC_TRACE0(DBG_ERROR,"In NicAllocRecvResources, Failed to allocate memory for RCB's\r\n");
			break;
		}

		NdisZeroMemory(pRCBMem, sizeof(RCB) * (Adapter->MaxRCBs + Adapter->MaxRxPkts));
		Adapter->RCBMem = pRCBMem;

		SMSC_TRACE1(DBG_INIT,"RCB Base = 0x%p\r\n", Adapter->RCBMem);

		//
		// Allocate the BulkInCompletionArray
		//
		Status = NdisAllocateMemory(
			(PVOID*) &Adapter->BulkInCompletionArray,
			sizeof(PRCB) * Adapter->MaxRCBs,
			NDIS_ALLOC_FLAG, 
			HighestAcceptableMax);

		if(Status != NDIS_STATUS_SUCCESS)
		{
			SMSC_TRACE0(DBG_ERROR,"In NicAllocRecvResources,Failed to allocate memory for BulkInCompletionArray\r\n");
			Adapter->BulkInCompletionArray = NULL;
			break;
		}

		NdisZeroMemory(Adapter->BulkInCompletionArray, sizeof(PRCB) * Adapter->MaxRCBs);

		//
		// Allocate the data buffers for the USB RCBs
		//
		Status = NdisAllocateMemory(
			&pRCBData,
			Adapter->RxBurstMax * Adapter->MaxRCBs,
			NDIS_ALLOC_FLAG, 
			HighestAcceptableMax);

		if(Status != NDIS_STATUS_SUCCESS)
		{
			SMSC_TRACE0(DBG_ERROR,"In NicAllocRecvResources, Failed to allocate memory for RCB's data buffers\r\n");
			break;
		}
		NdisZeroMemory(pRCBData, Adapter->RxBurstMax * Adapter->MaxRCBs);
		Adapter->RCBData = pRCBData;

		SMSC_TRACE1(DBG_INIT,"RCB Data Buffers Base = 0x%p\r\n", Adapter->RCBData);

		//
		// Allocate a buffer pool for RCB (USB) recv buffers.
		//

		NdisAllocateBufferPool(
			&Status,
			&Adapter->RCBBufferPoolHandle,
			Adapter->MaxRCBs);
		if(Status != NDIS_STATUS_SUCCESS)
		{
			SMSC_TRACE0(DBG_ERROR,"In NicAllocRecvResources, NdisAllocateBufferPool for RCB recv buffers failed\r\n");
			break;
		}

		//
		// Allocate RCB packet pool for (USB) receive indications
		//
		NdisAllocatePacketPool(
			&Status,
			&Adapter->RCBPacketPoolHandle,
			Adapter->MaxRCBs,
			PROTOCOL_RESERVED_SIZE_IN_PACKET);

		if(Status != NDIS_STATUS_SUCCESS)
		{
			SMSC_TRACE0(DBG_ERROR,"In NicAllocRecvResources, NdisAllocatePacketPool for RCB recv packets failed\r\n");
			break;
		}

		//
		// Divide the first part of the RCBMem blob into RCBs for
		// the Bulk in USBs and create a buffer descriptor for the
		// Data portion of the RCBs.
		//

		for(index=0; index < Adapter->MaxRCBs; index++)
		{
			pRCB = (PRCB) pRCBMem;

			//
			// Initialize the RCB structure.
			//
			pRCB->pData = pRCBData;
			pRCB->Adapter = Adapter;
			pRCB->BulkInRCB=1;


			//
			// Create a buffer descriptor for the Data portion of the RCBs.
			// Buffer descriptors are nothing but MDLs on NT systems.
			//
			NdisAllocateBuffer(
				&Status,
				&Buffer,
				Adapter->RCBBufferPoolHandle,
				(PVOID) (pRCBData + 10),
				(Adapter->RxBurstMax - 10UL));
			if(Status != NDIS_STATUS_SUCCESS)
			{
				SMSC_TRACE0(DBG_ERROR,"In NicAllocRecvResources, NdisAllocateBuffer for RCB Recv failed\r\n");
				break;
			}

			pRCB->Buffer = Buffer;

			pRCB->SyncEvent=CreateEvent( NULL, TRUE, FALSE, NULL);

			if ( !pRCB->SyncEvent ) {
					Status = NDIS_STATUS_FAILURE;
					break;
				}
 
			//
			// Initialize receive packets.
			// Allocate a packet descriptor for receive packets
			// from a preallocated pool.
			//
			NdisAllocatePacket(
				&Status,
				&Packet,
				Adapter->RCBPacketPoolHandle);
			if(Status != NDIS_STATUS_SUCCESS)
			{
				SMSC_TRACE0(DBG_ERROR,"In NicAllocRecvResources, NdisAllocatePacket for RCB failed\r\n");
				break;
			}

			pRCB->Packet = Packet;

			NDIS_SET_PACKET_HEADER_SIZE(Packet, ETH_HEADER_SIZE);

			NdisInterlockedInsertTailList(
				&Adapter->BulkInFreeList,
				&pRCB->List,
				&Adapter->BulkInLock);

			SMSC_TRACE1(DBG_INIT,"Bulk RCB 0x%p\r\n", pRCB);
			SMSC_TRACE1(DBG_INIT,"	 Ndis Packet: 0x%p\r\n", Packet);
			SMSC_TRACE1(DBG_INIT,"	 Ndis Buffer: 0x%p\r\n", Buffer);
			SMSC_TRACE1(DBG_INIT,"	 Data Buffer: 0x%p\r\n", pRCBData);

			pRCBMem = pRCBMem + sizeof(RCB);
			pRCBData += Adapter->RxBurstMax;

		}

		// If operating in multipacket receive mode we need a separate pool of
		// receive packets to split the packets in the RCB buffer before indication

		if (!Adapter->RxOnePktperRcb)
		{
			

			//
			// Allocate the data buffers for the split Rx Packets
			//

			Status = NdisAllocateMemory(
				&pRxData,
				Adapter->MaxRxPkts * NIC_BUFFER_SIZE,
				NDIS_ALLOC_FLAG, 
				HighestAcceptableMax);

			if(Status != NDIS_STATUS_SUCCESS)
			{
				SMSC_TRACE0(DBG_ERROR,"In NicAllocRecvResources, Failed to allocate memory for split Rx Pkt's data buffers\r\n");
				break;
			}
			NdisZeroMemory(pRxData, Adapter->MaxRxPkts * NIC_BUFFER_SIZE);
			Adapter->RxData = pRxData;

			SMSC_TRACE1(DBG_INIT,"Rx Data Buffers Base = 0x%p\r\n", Adapter->RxData);

			//
			// Allocate a buffer pool for split Rx buffers.
			//

			NdisAllocateBufferPool(
				&Status,
				&Adapter->RxBufferPoolHandle,
				Adapter->MaxRxPkts);
			if(Status != NDIS_STATUS_SUCCESS)
			{
				SMSC_TRACE0(DBG_ERROR,"In NicAllocRecvResources, NdisAllocateBufferPool for split Rx buffers failed\r\n");
				break;
			}

			//
			// Allocate packet pool for split Rx Packet indications
			//
			NdisAllocatePacketPool(
				&Status,
				&Adapter->RxPacketPoolHandle,
				Adapter->MaxRxPkts,
				PROTOCOL_RESERVED_SIZE_IN_PACKET);

			if(Status != NDIS_STATUS_SUCCESS)
			{
				SMSC_TRACE0(DBG_ERROR,"In NicAllocRecvResources, NdisAllocatePacketPool for split Rx packets failed\r\n");
				break;
			}

			//
			// Split the Second RCB blob / RxData memory blob and create a RCB/buffer/packet
			// linkeage for each piece
			//
			for(index=0; index < Adapter->MaxRxPkts; index++)
			{
				pRCB = (PRCB) pRCBMem;

				//
				// Initialize the RCB structure.
				//
				pRCB->pData = pRxData;
				pRCB->Adapter = Adapter;
				pRCB->BulkInRCB=0;

				//
				// Divide the second part of the RCBMem blob into RCBs for
				// the split RxPkts and create a buffer descriptor for each
				// of the Data portion of the RCBs.
				//
				NdisAllocateBuffer(
					&Status,
					&Buffer,
					Adapter->RxBufferPoolHandle,
					 (PVOID) (pRxData + 10UL),
                    			(NIC_BUFFER_SIZE - 10UL));   // We won't ever use that much, but that's the abs max.
				if(Status != NDIS_STATUS_SUCCESS)
				{
					SMSC_TRACE0(DBG_ERROR,"In NicAllocRecvResources, NdisAllocateBuffer for Rx Split Packet failed\r\n");
					break;
				}

				pRCB->Buffer = Buffer;

				//
				// Initialize receive packets.
				// Allocate a packet descriptor for receive packets
				// from a preallocated pool.
				//
				NdisAllocatePacket(
					&Status,
					&Packet,
					Adapter->RxPacketPoolHandle);
				if(Status != NDIS_STATUS_SUCCESS)
				{
					SMSC_TRACE0(DBG_ERROR,"In NicAllocRecvResources, NdisAllocatePacket for Rx Split packets failed\r\n");
					break;
				}

				pRCB->Packet = Packet;

				NDIS_SET_PACKET_HEADER_SIZE(Packet, ETH_HEADER_SIZE);

				NdisInterlockedInsertTailList(
					&Adapter->RxPktFreeList,
					&pRCB->List,
					&Adapter->RxPktLock);

				SMSC_TRACE1(DBG_INIT,"Rx Split Pkt RCB 0x%p\r\n", pRCB);
				SMSC_TRACE1(DBG_INIT,"	 Ndis Packet: 0x%p\r\n", Packet);
				SMSC_TRACE1(DBG_INIT,"	 Ndis Buffer: 0x%p\r\n", Buffer);
				SMSC_TRACE1(DBG_INIT,"	 Data Buffer: 0x%p\r\n", pRxData);

				pRxData +=	NIC_BUFFER_SIZE;
				pRCBMem = pRCBMem + sizeof(RCB);
			}
		}

	}while(FALSE);

	SMSC_TRACE1(DBG_INIT,"<-- NICAllocRecvResources %x\r\n", Status);

	return Status;
}



/*++
Routine Description:

	Free resources allocated for receive operation

Arguments:

		Adapter - Pointer to our adapter

Return Value:
	None

--*/

VOID
NICFreeRecvResources(
	PMP_ADAPTER Adapter)

{
	
	PRCB		   pRCB;

	SMSC_TRACE0(DBG_SHUT_DOWN,"--> NICFreeRecvResources\r\n");

	//
	// Free all the resources we allocated for receive.
	//
	if(!MP_TEST_FLAG(Adapter, fMP_RECV_SIDE_RESOURCE_ALLOCATED)){
		return;
	}

	while(!IsListEmpty(&Adapter->BulkInFreeList))
	{
		pRCB = (PRCB) NdisInterlockedRemoveHeadList(
					 &Adapter->BulkInFreeList,
					 &Adapter->BulkInLock);
		if(!pRCB){
			break;
		}

		if(pRCB->Buffer){
			NdisFreeBuffer(pRCB->Buffer);
		}
		if(pRCB->Packet) {
			NdisFreePacket(pRCB->Packet);
		}

		if(pRCB->SyncEvent){
			if (!CloseHandle(pRCB->SyncEvent))
			SMSC_TRACE0(DBG_ERROR,"In NICFreeRecvResources, close pRCB event fail!\r\n");
		}

	}

	if(Adapter->RCBPacketPoolHandle)
	{
		NdisFreePacketPool(Adapter->RCBPacketPoolHandle);
		Adapter->RCBPacketPoolHandle = NULL;
	}

	if(Adapter->RCBBufferPoolHandle)
	{
		NdisFreeBufferPool(Adapter->RCBBufferPoolHandle);
		Adapter->RCBBufferPoolHandle = NULL;
	}

	if(Adapter->RCBData) {
		NdisFreeMemory(Adapter->RCBData, Adapter->RxBurstMax * Adapter->MaxRCBs, NDIS_ALLOC_FLAG);
		Adapter->RCBData = NULL;
	}

	if (Adapter->BulkInCompletionArray) {
		NdisFreeMemory(Adapter->BulkInCompletionArray, sizeof(PRCB) * Adapter->MaxRCBs, NDIS_ALLOC_FLAG);
		Adapter->BulkInCompletionArray = NULL;
	}

	ASSERT(IsListEmpty(&Adapter->BulkInFreeList));
	ASSERT(IsListEmpty(&Adapter->BulkInProcessedList));
	ASSERT(IsListEmpty(&Adapter->BulkInWaitList));
	ASSERT(IsListEmpty(&Adapter->BulkInBusyList));
	ASSERT(IsListEmpty(&Adapter->BulkInCompletionList));
	NdisFreeSpinLock(&Adapter->BulkInLock);
	NdisFreeSpinLock(&Adapter->RCBProcessLock);
    NdisFreeEvent(&Adapter->ReadWorkItemReady);
    NdisFreeEvent(&Adapter->BulkInResumeWorkItemReady);
    NdisFreeEvent(&Adapter->BulkInCompletionWorkItemReady);

	// If operating in multipacket mode ...

	if (!Adapter->RxOnePktperRcb)
	{
		while(!IsListEmpty(&Adapter->RxPktFreeList))
		{
			pRCB = (PRCB) NdisInterlockedRemoveHeadList(
						 &Adapter->RxPktFreeList,
						 &Adapter->RxPktLock);
			if(!pRCB){
				break;
			}

			if(pRCB->Buffer){
				NdisFreeBuffer(pRCB->Buffer);
			}
			if(pRCB->Packet) {
				NdisFreePacket(pRCB->Packet);
			}
		}

		if(Adapter->RxPacketPoolHandle)
		{
			NdisFreePacketPool(Adapter->RxPacketPoolHandle);
			Adapter->RxPacketPoolHandle = NULL;
		}

		if(Adapter->RxBufferPoolHandle)
		{
			NdisFreeBufferPool(Adapter->RxBufferPoolHandle);
			Adapter->RxBufferPoolHandle = NULL;
		}

		if (Adapter->RxData)
		{
			NdisFreeMemory(Adapter->RxData, Adapter->MaxRxPkts * NIC_BUFFER_SIZE, NDIS_ALLOC_FLAG);
			Adapter->RxData = NULL;
		}

		ASSERT(IsListEmpty(&Adapter->RxPktFreeList));
		ASSERT(IsListEmpty(&Adapter->RxPktBusyList));

		NdisFreeSpinLock(&Adapter->RxPktLock);
	}

	// Do this at the end because it holds both RCBs for Bulk In and split packets.
	if(Adapter->RCBMem)
	{
		NdisFreeMemory( Adapter->RCBMem,
						sizeof(RCB) * (Adapter->MaxRCBs + Adapter->MaxRxPkts),
						NDIS_ALLOC_FLAG);
		Adapter->RCBMem = NULL;
	}

	#ifdef USE_THREAD
	if(Adapter->RecvThreadEvent){
		if (!CloseHandle(Adapter->RecvThreadEvent))
		SMSC_TRACE0(DBG_ERROR,"In NicFreeRecvResources,close RecvThreadEvent fail!\r\n");
	}
	#endif

    #ifdef USE_WORK_ITEM_NO_OUT
    NdisFreeEvent(&Adapter->RecvWorkItemReady);
    #endif

	MP_CLEAR_FLAG(Adapter, fMP_RECV_SIDE_RESOURCE_ALLOCATED);
	SMSC_TRACE0(DBG_SHUT_DOWN,"<-- NICFreeRecvResources\r\n");
	return;
}



/*++
Routine Description:

	Allocate resources required to process Wake up patterns

Arguments:

	Adapter - Pointer to our adapter

	

Return Value:

	NDIS_STATUS_SUCCESS
	NDIS_STATUS_FAILURE
	NDIS_STATUS_RESOURCES

--*/
NDIS_STATUS
NICAllocWolResources(
	PMP_ADAPTER Adapter)

{
	NDIS_STATUS 			Status;
	PLM_WOL_SAMPLE_STRUCT	pLmWOLStruct;
	BYTE					*pLmMask;
	BYTE					*pLmSample;
	DWORD					WolMemSize;
	BYTE					*pWOLMem;
	int 					Count;
	NDIS_PHYSICAL_ADDRESS HighestAcceptableMax;

	SMSC_TRACE0(DBG_INIT,"--> NICAllocWolResources\r\n");

	// Allocate adapter context memory.
    	HighestAcceptableMax.LowPart = 0xFFFFFFFFUL;        // DWORD
    	HighestAcceptableMax.HighPart = -1L;                // LONG
    	
	// We allocate space for each pattern's (struct + data + mask) + a scratchpad
	WolMemSize = NUMBER_OF_WOL_PATTERNS_8 * (sizeof(LM_WOL_SAMPLE_STRUCT) + (SIZE_OF_WOL_SAMPLE/8) + SIZE_OF_WOL_SAMPLE) + SIZE_OF_WOL_SAMPLE/8 + SIZE_OF_WOL_SAMPLE;

	Status = NdisAllocateMemory(
						(PVOID) &Adapter->NicWolMemBlock,
						(UINT) WolMemSize,
						NDIS_ALLOC_FLAG, 
						HighestAcceptableMax );

	if(Status != NDIS_STATUS_SUCCESS)
	{
		SMSC_TRACE0(DBG_ERROR,"In NicAllocWolResources, Failed to allocate memory for ICB's\r\n");
		goto Exit_NICAllocWolResources;
	}

	NdisZeroMemory((PVOID)Adapter->NicWolMemBlock, WolMemSize);

	Adapter->NicWolMemSize = WolMemSize;

	// Initialize WOL memory. Links between samples stucture and data storage.

	pWOLMem =		Adapter->NicWolMemBlock;
	pLmWOLStruct =	(PLM_WOL_SAMPLE_STRUCT) pWOLMem;
	pLmMask =		pWOLMem + NUMBER_OF_WOL_PATTERNS_8 * sizeof(LM_WOL_SAMPLE_STRUCT);
	pLmSample = 	pWOLMem + NUMBER_OF_WOL_PATTERNS_8 * ( sizeof(LM_WOL_SAMPLE_STRUCT) + (SIZE_OF_WOL_SAMPLE/8) );

	if ((pLmWOLStruct != NULL) &&
        (pLmMask != NULL) &&
        (pLmSample != NULL))
    {
		for(Count = 0; Count < NUMBER_OF_WOL_PATTERNS_8; Count++)
		{
			pLmWOLStruct->pWOLSample = pLmSample;
			pLmWOLStruct->pWOLMask = pLmMask;

			#ifdef SMSC_DBG
			if (g_DebugMode &DBG_INIT) {
				RETAILMSG(1,(TEXT("pLmWOLStruct[%d] = %p, "),Count,pLmWOLStruct));
				RETAILMSG(1,(TEXT("pLmMask[%d] = %p, "),Count,pLmMask));
				RETAILMSG(1,(TEXT("pLmSample[%d] = %p\r\n"),Count,pLmSample));
			}
			#endif

			pLmWOLStruct++;
			pLmMask += (SIZE_OF_WOL_SAMPLE/8);
			pLmSample += SIZE_OF_WOL_SAMPLE;
		}
	}

	Adapter->pWolMaskScratchpad = pLmSample;
	Adapter->pWolSampleScratchpad = pLmSample + (SIZE_OF_WOL_SAMPLE/8);
	SMSC_TRACE1(DBG_INIT,"pWolMaskScratchpad = %p\r\n",pLmSample);
	SMSC_TRACE1(DBG_INIT,"pWolSampleScratchpad = %p\r\n",pLmSample+(SIZE_OF_WOL_SAMPLE/8));

Exit_NICAllocWolResources:
	SMSC_TRACE0(DBG_INIT,"<-- NICAllocWolResources\r\n");

	return Status;
}

/*++
Routine Description:

   Free resources allocated for wake up pattern operation

Arguments:

	Adapter - Pointer to our adapter
	
Return Value:
	None


--*/

VOID
NICFreeWolResources(
	PMP_ADAPTER Adapter)

{
	SMSC_TRACE0(DBG_SHUT_DOWN,"--> NICFreeWolResources\r\n");
	if (Adapter->NicWolMemBlock)
		NdisFreeMemory(Adapter->NicWolMemBlock,Adapter->NicWolMemSize, NDIS_ALLOC_FLAG);
	SMSC_TRACE0(DBG_SHUT_DOWN,"<-- NICFreeWolResources\r\n");
}

/*++

Routine Description:

		This routine performs a reset of the adapter. The USB Interface is NOT reset.
	The Receive and Transmit paths and interrupts are not started (will be done
	later together with corrsponding USB pipes).

Arguments:

	Adapter 	Pointer to our adapter

Return Value:

	NIC_STATUS

--*/
NIC_STATUS
NICResetAdapter(
	IN PMP_ADAPTER Adapter
	)

{
	DWORD Value32,dwTemp;
	UINT Timeout;
	NIC_STATUS Status;	
	DWORD HwCfgBeforeLrst;

	SMSC_TRACE0(DBG_RESET,"--> NICResetAdapter\r\n");

	/*
	 *	See if LAN7500 is present by reading and verifying it's ID
	 */
	if(!NICIsCardPresent(Adapter))
	{
		Status = NIC_STATUS_FAILURE;
		SMSC_TRACE0(DBG_INIT,"Card is not present first time!!!!\r\n");
		goto Exit_NICResetAdapter;
	}
/*
	// Check the preset MAC address in ADDRH, and ADDRL Regs. during initialization.
	// The preset network address may have been written into ADDRL & ADDRH by BIOS
	// to avoid the necessity of EEPROM in the design.
	if (MP_TEST_FLAG(Adapter, fMP_INIT_IN_PROGRESS))
	{	 
		WORD *pWord1 = (WORD *) Adapter->CurrentNodeAddress;

		// Make sure CurrentNodeAddress[] has not been loaded by Node Address registry key
		if(!(*pWord1 | *(pWord1+1) | *(pWord1+2)))
		{
			Status = NICReadRegister( Adapter, ADDRL, ((DWORD *) Adapter->CurrentNodeAddress) );
			Status |= NICReadRegister( Adapter, ADDRH, &Value32 );
			if (Status != NIC_STATUS_SUCCESS)
				goto Exit_NICResetAdapter;
			*((WORD *) (&Adapter->CurrentNodeAddress[4])) = (WORD)Value32;
			// See if whatever we read, is a valid address 
			{
				int i;
				BYTE BcastCheck = 0xff;
				
				for (i=0;i<6;i++) 
				{
					BcastCheck &= Adapter->CurrentNodeAddress[i];
				}

				if (BcastCheck == 0xff)
				{
					SMSC_TRACE0(DBG_INIT,"ADDRL & ADDRH are default values\r\n");

					// Zero out CurrentNodeAddress[] now and it then will be loaded w/ PermanentNodeAddress[]
					memset(Adapter->CurrentNodeAddress,0,ETH_LENGTH_OF_ADDRESS);
				}
			}
		}
	}
*/
	Status = NICReadRegister(Adapter, ETH_HW_CFG, &Value32);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICResetAdapter;

	 // Keep a copy of ETH_HW_CFG register
	HwCfgBeforeLrst = Value32;
	 
/*
	Adapter->PhyIsExternal = (BOOLEAN)((Value32 & ETH_HW_CFG_PSEL_) == ETH_HW_CFG_PSEL_);
	SMSC_TRACE1(DBG_INIT,"Phy in use is %s\r\n", Adapter->PhyIsExternal ? "EXTERNAL" : "INTERNAL");
	if (Adapter->PhyIsExternal)
	{
		SMSC_TRACE0(DBG_INIT,"Disabling 'WakeOnLink' and 'SleepPhyOnLinkDown' features\r\n");
		Adapter->WoLinkChgSettings = DISABLED;
		Adapter->SleepPhyOnLinkDown = FALSE;
	}
*/
	//No External PHY support in LAN7500
	Adapter->PhyIsExternal =0;

	/*
	 *	Reset LAN7500 to get it to a well know state before we start using it
	 */

	/*
	 *	We cannot do a full software reset because the device will
	 *	"disappear" when the USB stuff resets. Use lite reset instead.
	 */

	// Now issue the lite reset and wait for bit to self clear, should not take more
	// than a few ms (in reality ns).
	

	dwTemp= Value32 | ETH_HW_CFG_LRST_;
	Status=NICWriteRegister( Adapter, ETH_HW_CFG, &dwTemp);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICResetAdapter;
	
	

	Timeout = 0;
	Status=NICReadRegister(Adapter, ETH_HW_CFG, &Value32);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICResetAdapter;
	
	

	while ( (Value32 & ETH_HW_CFG_LRST_) && (Timeout < NIC_1_SEC_TIMEOUT))
	{
		NdisMSleep(NDIS_SLEEP_1_MS);		 /* wait for 1 ms before trying again */
		Timeout++;
		Status=NICReadRegister(Adapter, ETH_HW_CFG, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
		{
			break;
		}
	}

	if ((Timeout >= NIC_1_SEC_TIMEOUT) || (Status != NIC_STATUS_SUCCESS))
	{
		// Should never happen.
		SMSC_TRACE1(DBG_ERROR,"Timeout waiting for completion of Lite Reset after %d second\r\n", NIC_1_SEC_TIMEOUT/1000);
		SMSC_TRACE1(DBG_ERROR,"After LRST: ETH_HW_CFG   = 0x%08lx\r\n", Value32);//debug only
        NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);//debug only
        SMSC_TRACE1(DBG_ERROR,"After LRST: ETH_PMT_CTL   = 0x%08lx\r\n", Value32);//debug only
        NICReadRegister(Adapter, ETH_MAC_CR, &Value32);//debug only
        SMSC_TRACE1(DBG_ERROR,"After LRST: ETH_MAC_CR   = 0x%08lx\r\n", Value32);//debug only
        
		ASSERT(FALSE);
		Status = NIC_STATUS_FAILURE;
		goto Exit_NICResetAdapter;
	}
	
		
	Status=NICReadRegister(Adapter, ETH_ID_REV, &Value32);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICResetAdapter;
	SMSC_TRACE1(DBG_RESET,"ID_Reg	= 0x%08lx\r\n", Value32);
	Status=NICReadRegister(Adapter, ETH_FPGA_REV, &Value32);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICResetAdapter;
	SMSC_TRACE1(DBG_RESET,"FPGA_Reg = 0x%08lx\r\n", Value32);
	

	/*
	 *	Make sure LAN7500 is still present by reading and verifying it's ID once again
	 */

	if(!NICIsCardPresent(Adapter))
	{
		Status = NIC_STATUS_FAILURE;
		SMSC_TRACE0(DBG_RESET,"Card is not present second time!!!!\r\n");
		goto Exit_NICResetAdapter;
	}

	/*
	 *	OK LAN7500 is now in a well known state
	 *	Time to start doing some real NIC initialization stuff now.
	 */

	// If this reset is part of the initialization lets get the
	// Permanent and current node addresses, if not just skip this.
	SMSC_TRACE1(DBG_RESET,"Adapter->Flags 0x%x\r\n",Adapter->Flags);

	if (MP_TEST_FLAG(Adapter, fMP_INIT_IN_PROGRESS))
	{
		WORD *pWord1;
		WORD *pWord2;
		// This is where the permanent and current address will be stored
		// once this block of code executes. Also the entry values before
		// execution are specified.
		//
		// PermanentNodeAddress  Is the permanent address. Starts as 0
		// CurrentNodeAddress	 Is the current address. Starts as 0
		//						 unless there was a node override via
		//						 registry in which case it is loaded
		//						 with that address

		//	Get the Permanent node address from the EEPROM/Flash/Chip
		Status=NICGetPermanentNodeAddress(Adapter);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;
		

		// See if there was a node ID override via the registry and act
		// accordingly.
		pWord1 = (WORD *) Adapter->CurrentNodeAddress;
		pWord2 = (WORD *) Adapter->PermanentNodeAddress;

		if(*pWord1 | *(pWord1+1) | *(pWord1+2))
        {
        //  There was a node override, so copy this current node 
        //  address into the chip's registers.
		//	There was a node override, so copy this current node
		//	address into the chip's registers.
		memcpy(&dwTemp, Adapter->CurrentNodeAddress, sizeof(DWORD));
		Status = NICWriteRegister( Adapter, ETH_RX_ADDRL, &dwTemp);
		memcpy(&dwTemp, &Adapter->CurrentNodeAddress[4], 2);
		Status |= NICWriteRegister( Adapter, ETH_RX_ADDRH, &dwTemp);
			
      	// Assign the first MAF to adapter MAC address
      	memcpy(&dwTemp, Adapter->CurrentNodeAddress, sizeof(DWORD));
            	Status = NICWriteRegister( Adapter, ETH_MAF_LO_0, &dwTemp );
		memcpy(&dwTemp, &Adapter->CurrentNodeAddress[4], 2);
		dwTemp|=ETH_MAF_HI_VALID_;
            	Status |= NICWriteRegister( Adapter, ETH_MAF_HI_0, &dwTemp);
        }
        else
        {
            //  There was no node override, so copy the permanent node 
            //  address into the current node address and into the chip 
            //  registers
            memcpy(&dwTemp, Adapter->PermanentNodeAddress, sizeof(DWORD));
			Status = NICWriteRegister( Adapter, ETH_RX_ADDRL, &dwTemp);
			memcpy(&dwTemp, &Adapter->PermanentNodeAddress[4], 2);
			Status |= NICWriteRegister( Adapter, ETH_RX_ADDRH, &dwTemp);
			
      		// Assign the first MAF to adapter MAC address
      		memcpy(&dwTemp, Adapter->PermanentNodeAddress, sizeof(DWORD));
            	Status = NICWriteRegister( Adapter, ETH_MAF_LO_0, &dwTemp );
			memcpy(&dwTemp, &Adapter->PermanentNodeAddress[4], 2);
			dwTemp|=ETH_MAF_HI_VALID_;
            	Status |= NICWriteRegister( Adapter, ETH_MAF_HI_0, &dwTemp);

            *pWord1 = *pWord2;

            pWord1++;
            pWord2++;
            *pWord1 = *pWord2;

            pWord1++;
            pWord2++;
            *pWord1 = *pWord2;
        }

		
		
		if (g_DebugMode & DBG_RESET) {
			
			int i;
			SMSC_TRACE0(DBG_RESET,"Permanent Node Address:");
			for(i = 0; i < 6; i++)
				RETAILMSG(1,(TEXT(" %02x"), Adapter->PermanentNodeAddress[i]));
			RETAILMSG(1,(TEXT("\r\n")));

			SMSC_TRACE0(DBG_RESET,"Current Node Address:");
			for(i = 0; i < 6; i++)
				RETAILMSG(1,(TEXT(" %02x"), Adapter->CurrentNodeAddress[i]));
			RETAILMSG(1,(TEXT("\r\n")));
		}
	   
		
	}
	else
	{
		// The reset cleared the address registers, restore them

		memcpy(&dwTemp, Adapter->CurrentNodeAddress, sizeof(DWORD));
		Status = NICWriteRegister( Adapter, ETH_RX_ADDRL, &dwTemp);
		memcpy(&dwTemp, &Adapter->CurrentNodeAddress[4], 2);
		Status |= NICWriteRegister( Adapter, ETH_RX_ADDRH, &dwTemp);
			
      		// Assign the first MAF to adapter MAC address
      		memcpy(&dwTemp, Adapter->CurrentNodeAddress, sizeof(DWORD));
            	Status = NICWriteRegister( Adapter, ETH_MAF_LO_0, &dwTemp );
		memcpy(&dwTemp, &Adapter->CurrentNodeAddress[4], 2);
		dwTemp|=ETH_MAF_HI_VALID_;
            	Status |= NICWriteRegister( Adapter, ETH_MAF_HI_0, &dwTemp);
			
		if (g_DebugMode & DBG_RESET) {
			int j;

			SMSC_TRACE0(DBG_RESET,"Permanent Node Address hardcoded in the driver :");
			for(j = 0; j < 6; j++)
				RETAILMSG(1,(TEXT(" %02x"), Adapter->CurrentNodeAddress[j]));
			RETAILMSG(1,(TEXT("\r\n")));
		}
		
		
	}

	if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;

	// Set the HW_CFG options according to desired options.
	Status=NICReadRegister( Adapter, ETH_HW_CFG, &Value32 );
	if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;

	if (!Adapter->StopBulkInOnRxEmpty)
	{
		// Before we start the receive pipe, make sure we NAK when no data
		// if we'll use the "Bulk only" / low latency mode.
		Value32 |= ETH_HW_CFG_BIR_; 
		SMSC_TRACE0(DBG_RESET,"Setting device to NAK when no RX Data\r\n");
		
	}

	if (!Adapter->RxOnePktperRcb)
	{
		// Enable Multipacket Receives
		Value32 |= ETH_HW_CFG_MEF_;
		SMSC_TRACE0(DBG_RESET,"Setting device to Enable Multipkt Ethernet per USB transaction on Rx\r\n");
		dwTemp=Adapter->MEFBulkInDelay;
		Status=NICWriteRegister( Adapter, ETH_BULK_IN_DLY, &dwTemp);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;

		if (Adapter->RxBurstCap)
		{
			Value32 |= ETH_HW_CFG_BCE_;
			dwTemp=Adapter->RxBurstCap/(Adapter->UsbSpeedIsHigh ? 512 : 64);
			Status=NICWriteRegister( Adapter, ETH_BURST_CAP,&dwTemp);
			if (Status != NIC_STATUS_SUCCESS)
				goto Exit_NICResetAdapter;
			SMSC_TRACE2(DBG_RESET,"Setting Device to Enable Burst for Rx in HW_CFG, and setting BURST_CAP = 0x%08lx (%d bytes)\r\n",
							Adapter->RxBurstCap/(Adapter->UsbSpeedIsHigh ? 512 : 64),
							Adapter->RxBurstCap);
		}
	}

	
	{
		// Lets first clear the field
		Value32 &= (~ETH_HW_CFG_PHY_BOOST_);
		
		if (Adapter->PHYDriveStrength == 0xff) // Default PHY_BOOST setting
		{
			Value32 |= (HwCfgBeforeLrst & ETH_HW_CFG_PHY_BOOST_);
		}
		else // Forced by registry
		{
			Value32 |= (Adapter->PHYDriveStrength << 8);
		}
	}

	Status=NICWriteRegister( Adapter, ETH_HW_CFG, &Value32) ;
	if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;
	
	
	Status=NICReadRegister( Adapter, ETH_HW_CFG, &Value32 );
	if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;
	SMSC_TRACE1(DBG_RESET,"HW_CFG = 0x%08lx\r\n", Value32);
	Status=NICReadRegister( Adapter, ETH_BURST_CAP, &Value32 );
	if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;
	SMSC_TRACE1(DBG_RESET,"ETH_BURST_CAP = 0x%08lx\r\n", Value32);
	Status=NICReadRegister( Adapter, ETH_BULK_IN_DLY, &Value32 );
	if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;
	SMSC_TRACE1(DBG_RESET,"ETH_BULK_IN_DLY = 0x%08lx\r\n", Value32);

		
	// Enable the LEDs by default
	dwTemp= ETH_LED_GPIO_CFG0_GPIOEN_ | \
            ETH_LED_GPIO_CFG0_LED2_FUN_SEL_ | \
            ETH_LED_GPIO_CFG0_SPDLED_FUN_SEL_ ;

	Status=NICWriteRegister( Adapter, ETH_LED_GPIO_CFG0, &dwTemp) ;
	if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;

	#ifdef USE_INT_PIPE
	if (Adapter->StopBulkInOnRxEmpty) {
	// Enable desired interrupts.
	Value32 =	 ETH_INT_EP_TXE_EN_         | \
                ETH_INT_EP_TDFU_EN_        | \
                ETH_INT_EP_TDFO_EN_;
	
	// Only request data available interrupt if we need it.
	if (Adapter->StopBulkInOnRxEmpty) {
		SMSC_TRACE1(DBG_INIT,"StopBulkInOnRxEmpty  = %d\r\n", Adapter->StopBulkInOnRxEmpty);
		Value32 |= ETH_INT_EP_RX_FIFO_EN_;
		SMSC_TRACE1(DBG_INIT,"Value32  = 0x%x\r\n", Value32);
	}

	Status=NICWriteRegister( Adapter, ETH_INT_EP_CTL, &Value32 );
	if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;
	
	Status=NICReadRegister( Adapter, ETH_INT_EP_CTL, &Value32 );
	if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;
	SMSC_TRACE1(DBG_INIT,"Turning on capabilities of some events to generate interrupts. ETH_INT_EP_CTL = 0x%08lx\r\n", Value32);
	
	
	}
	#endif


	/* Set up MAC for auto speed/auto duplex mode detection */
    Status = NICReadRegister(Adapter, ETH_MAC_CR, &Value32);
    if (Status != NIC_STATUS_SUCCESS)
        goto Exit_NICResetAdapter;
    Value32 |= (ETH_MAC_CR_AUTO_SPEED_ | ETH_MAC_CR_AUTO_DUPLEX_);
    
    Status = NICWriteRegister(Adapter, ETH_MAC_CR, &Value32);
    if (Status != NIC_STATUS_SUCCESS)
        goto Exit_NICResetAdapter;

    
    Status = NICReadRegister(Adapter, ETH_MAC_CR, &Value32);
    if (Status != NIC_STATUS_SUCCESS)
        goto Exit_NICResetAdapter;
    SMSC_TRACE1(DBG_INIT,"ETH_MAC_CR = 0x%08lx\r\n", Value32 );


	

	// Determine line speed and media type
	Status = NICInitializeMedia(Adapter);
	if(Status != NIC_STATUS_SUCCESS)
	{

		goto Exit_NICResetAdapter;
	}

	// Restore WOL registers
	if(Adapter->PowerManagement != DISABLED)
	{
	
		DWORD Temp32;
		USHORT i;
		 // Restore WOL Configs if any was set.
        	for (i=0;i<Adapter->NumOfWolPatterns;i++)
        	{
        			Temp32=Adapter->dwWUFFCfg[i];
            		Status = NICWriteRegister(Adapter, (ETH_WUF_CFGx + i*4), &Temp32);
            		if (Status != NIC_STATUS_SUCCESS)
                		goto Exit_NICResetAdapter;
        	}

        	// Restore WOL Filters if any was set.
        	for (i=0;i<(Adapter->NumOfWolPatterns*4);i++)
        	{
        		Temp32=Adapter->dwWUFFMask[i];
            		Status = NICWriteRegister(Adapter, (ETH_WUF_MASKx + i*4UL), &Temp32);
            		if (Status != NIC_STATUS_SUCCESS)
                		goto Exit_NICResetAdapter;
        	}
	}

	 /* 
     * Reset VLAN filtering RAM (VHF RAM)
     */
    {
        DWORD i;
        
        Status = NIC_STATUS_SUCCESS;
        SMSC_TRACE0(DBG_INIT,"Reseting VLAN filtering RAM \r\n");
        for (i=0; i<128; i++)
        {
            Status |= NICWriteVlanRam32(Adapter, i, 0UL);
        }

        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICResetAdapter;
    }


	if (Adapter->PriorityVlanTag != 0)
	{
		// Setup the ETH_VLAN_TYPE register to accept VLAN packets /
		dwTemp=NDIS_ETH_TYPE_802_1Q;
        Status = NICWriteRegister(Adapter, ETH_VLAN_TYPE, &dwTemp);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICResetAdapter;

			
		// Setup VLAN Frame Size Enforcement 
		Status = NICReadRegister(Adapter, ETH_MAC_RX, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;

		Value32 |= ETH_MAC_RX_VLAN_FSE_;
		Status = NICWriteRegister(Adapter, ETH_MAC_RX, &Value32 );
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;

        // Enabling VLAN tag stripping & Vlan Filtering 
        Status = NICReadRegister(Adapter, ETH_RFE_CTL, &Value32);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICResetAdapter;

		dwTemp=Value32 | ETH_RFE_CTL_VLAN_STRIP_ | ETH_RFE_CTL_VLAN_FILTER_;
        Status = NICWriteRegister(Adapter, ETH_RFE_CTL,&dwTemp);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICResetAdapter;

       	// Set associated bit in VLAN filtering RAM
        Status = NICReadVlanRam32(Adapter, (Adapter->VlanID >> 5), &Value32);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICResetAdapter;

        Value32 |= (0x1UL << (Adapter->VlanID & 0x1FUL));
        
        Status = NICWriteVlanRam32(Adapter, (Adapter->VlanID >> 5), Value32);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICResetAdapter;
		
        NICReadVlanRam32(Adapter, (Adapter->VlanID >> 5), &Value32);
       	SMSC_TRACE1(DBG_INIT,"Adapter->VlanID=%d\r\n", Adapter->VlanID);
        SMSC_TRACE2(DBG_INIT,"after set OID_GEN_VLAN_ID, VlanRam[%d]=0x%08lX\r\n", (Adapter->VlanID >> 5), Value32);
        
	}

/*
	// Initialize Flow Control

	if(Adapter->FlowControl != DISABLED)
	{
		BOOLEAN bGenerate = FALSE;
		BOOLEAN bRespond = FALSE;

		switch(Adapter->FlowControl)
		{
			case FC_SYMM:
				bRespond = TRUE;
				bGenerate = TRUE;
				break;

			case FC_ASYM_GEN:
				bGenerate = TRUE;
				break;

			case FC_ASYM_RESP:
				bRespond = TRUE;
				break;
		}

		if (bGenerate)
		{
			// Enable device to send pauses
			SMSC_TRACE2(DBG_RESET,"Enabling sending Flow Control frames with Pause = 0x%04X and AFC_REG = 0x%08lx\r\n",
							Adapter->FCPause, Adapter->AFCCfgReg);
		}

		if (bRespond)
		{
			// Enable device to respond to pauses
			SMSC_TRACE0(DBG_RESET,"Enabling responding to Flow Control frames\r\n");
		}
		dwTemp=((DWORD)(Adapter->FCPause << 16)) | (bRespond ? FLOW_FCEN_ : 0);
		Status=NICWriteRegister(Adapter, FLOW, &dwTemp);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;
		dwTemp=Adapter->AFCCfgReg;
		Status=NICWriteRegister(Adapter, ETH_AFC_CFG, &dwTemp);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;
	}

*/

	 /* 
     * Set multicast filter bits according to the current multicast filter
     * (which was reset at init time and set via OID_802_3_MULTICAST_LIST)
     */
    {
        DWORD i;
        
        Status = NIC_STATUS_SUCCESS;
        for (i=0; i<16; i++)
        {
//            SMSC_TRACE2(DBG_RESET,"HASH[%d] = 0x%08lx\n", i, Adapter->MulticastHashTable[i]);
            Status |= NICWriteHashRam32(Adapter, i, Adapter->MulticastHashTable[i]);
        }
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICResetAdapter;
    }

	Status=NICSetReceiveMask(Adapter, Adapter->PacketFilter);
	if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICResetAdapter;
	SMSC_TRACE1(DBG_RESET,"Adapter->ReceiveMask 0x%x\r\n",Adapter->ReceiveMask);

	// Reset Power state variable
	Adapter->CurrentPowerState = 0;

	Status = NIC_STATUS_SUCCESS;

Exit_NICResetAdapter:
	SMSC_TRACE0(DBG_RESET,"<-- NICResetAdapter\r\n");
	return Status;
} /* NICResetAdapter */

/*++

Routine Description:

	This routine determines if the adapter is present by reading the
	ID Register

Arguments:

	Adapter 	Pointer to our adapter

Return Value:

	TRUE if present, FALSE if not

--*/

BOOLEAN NICIsCardPresent(
	IN PMP_ADAPTER Adapter
	)

{
	DWORD Temp32;

	if (Adapter->CurrentPowerState)
	{
		SMSC_TRACE1(DBG_INIT,"<= NICIsCardPresent called in D%d =>\r\n", Adapter->CurrentPowerState);
		return FALSE;
	}

	if (NIC_STATUS_SUCCESS != NICReadRegister(Adapter, ETH_ID_REV, &Temp32))
	{
		return FALSE;
	}

	Adapter->DevIdReg = Temp32;
	
	if (GetChipIdFromID_REV(Temp32) == ETH_ID_REV_CHIP_ID_7500_)
		{
			return TRUE;
		}
		else if (GetChipIdFromID_REV(Temp32) == ETH_ID_REV_CHIP_ID_7505_)
		{
			return TRUE;
		}
		else
		{
			return FALSE;
		}

}


/*
void
NICGetWOLPatternCount(
	IN PMP_ADAPTER Adapter
	) 
{
	DWORD Temp32,Value32, i;
	SMSC_TRACE0(DBG_POWER,"---> NICGetWOLPatternCount()\r\n");
	   
	// If we got soemthing from the registry use that.
	if (Adapter->NumOfWolPatterns == 0xff)
	{
		// Assume it has 4, will correct if not later.
		Adapter->NumOfWolPatterns = 4;

		
			//*  Autodetect by 21 consecutive writing to WUFF w/ incremental pattern.
			//*  Therefore if we read the first pattern in the first WUFF mask, the chip has	
			//*  8 patterns, if not it has 4
			
		for (i=0;i<21;i++)
		{
			NICWriteRegister(Adapter, ETH_WUFF, &i);
		}

		//Reset the WUFF pointer
		Value32=ETH_WUCSR_WFF_PTR_RST;
		NICWriteRegister(Adapter, ETH_WUCSR, &Value32);
		
		NICReadRegister(Adapter, ETH_WUCSR, &Temp32);
		if (Temp32 & WUCSR_WFF_PTR_RST)
		{
			SMSC_TRACE0(DBG_POWER,"Failed to Reset WUFF Pointer (1)\r\n");
			goto Exit_NICGetWOLPatternCount;
		}

		//Read Filter0 Byte Mask 0
		NICReadRegister(Adapter, WUFF, &Temp32);
		if (!Temp32) // 8 filters
		{
			SMSC_TRACE0(DBG_POWER,"Device has 8 filters !\r\n");
			Adapter->NumOfWolPatterns = 8;
		}
		else
		{
			SMSC_TRACE0(DBG_POWER,"Device has 4 filters !\r\n");
		}
	}

	// Cleanup
	for (i=0;i<Adapter->NumOfWolPatterns*5;i++)
	{
		Value32=0;
		NICWriteRegister(Adapter, WUFF, &Value32);
	}

	//Reset the WUFF pointer
	Value32= WUCSR_WFF_PTR_RST;
	NICWriteRegister(Adapter, WUCSR, &Value32);
	
	NICReadRegister(Adapter, WUCSR, &Temp32);
	if (Temp32 & WUCSR_WFF_PTR_RST)
	{
		SMSC_TRACE0(DBG_POWER,"Failed to Reset WUFF Pointer (2)\r\n");
		goto Exit_NICGetWOLPatternCount;
	}

Exit_NICGetWOLPatternCount:

	SMSC_TRACE0(DBG_POWER,"<--- NICGetWOLPatternCount()\r\n");

	
}
*/

/*++

Routine Description:

	This routine retrieves the Permanent Node Address form the EEPROM

Arguments:

	Adapter 	Pointer to our adapter

Return Value:

	None

--*/

NIC_STATUS 
NICGetPermanentNodeAddress(
	IN PMP_ADAPTER Adapter
	)
{
	BYTE i;
	BYTE ZeroCheck = 0, BcastCheck = 0xff;
	BYTE Signature;
	NIC_STATUS Status=NIC_STATUS_FAILURE;
	DWORD	dwAddrl;
	DWORD	dwAddrh;
	DWORD	dwHwCfg = 0;
	WORD	wTemp[3] = {0xFFFF, 0xFFFF, 0xFFFF};
	
	SMSC_TRACE0(DBG_INIT,"--> NICGetPermanentNodeAddress\r\n");	
	
	// See if whatever is already in ADDFRL & ADDRH is valid 
	// (would be the case for Bothwell EEPROM-less case)
	SMSC_TRACE0(DBG_INIT,"Values read from ADDR Regs\r\n");
	Status = NICReadRegister(Adapter, ETH_RX_ADDRH, &dwAddrh);
	SMSC_TRACE1(DBG_INIT,"ETH_RX_ADDRH = 0x%08lx\r\n", dwAddrh);
	Status |= NICReadRegister(Adapter, ETH_RX_ADDRL, &dwAddrl);
	SMSC_TRACE1(DBG_INIT,"ETH_RX_ADDRL = 0x%08lx\r\n", dwAddrl);
	Status |= NICReadRegister(Adapter, ETH_HW_CFG, &dwHwCfg);
	SMSC_TRACE1(DBG_INIT,"ETH_HW_CFG = 0x%08lx\r\n", dwHwCfg);
	
	wTemp[0] = (WORD) dwAddrh;
	wTemp[1] = (WORD) dwAddrl;
	wTemp[2] = (WORD) (dwAddrl >> 16);
	
	if (
		((wTemp[0] | wTemp[1] | wTemp[2]) != 0) &&
		((wTemp[0] & wTemp[1] & wTemp[2]) != 0xFFFF) &&
		(dwHwCfg & ETH_HW_CFG_RST_PROTECT_) &&
		(Status == NIC_STATUS_SUCCESS)
		)
	{
		// Good Address, Exit
		SMSC_TRACE0(DBG_INIT,"EEPROM-less Mac address used!\r\n");
		*((DWORD *)&Adapter->PermanentNodeAddress[0]) = dwAddrl;
		*((WORD *)&Adapter->PermanentNodeAddress[4]) = (WORD)dwAddrh;
	
		goto Exit_NICGetPermanentNodeAddress;
	}

	Status = NICReadEepromByte(Adapter, 0, &Signature);
	if ((Signature == 0xA5) && (Status == NIC_STATUS_SUCCESS))
	{
		SMSC_TRACE0(DBG_INIT,"Values read from EEPROM\r\n");

		Status = NICReadEepromBlock(Adapter, 1, 6, Adapter->PermanentNodeAddress);
				
		
		if (Status == NIC_STATUS_SUCCESS)
		{
			SMSC_TRACE0(DBG_INIT,"Node Address =");
			#ifdef SMSC_DBG
			if (g_DebugMode & DBG_INIT) {
				for (i=0; i<6; i++)
				{
					RETAILMSG(1,(TEXT(" %02x"), Adapter->PermanentNodeAddress[i]));
				}
				RETAILMSG(1,(TEXT("\r\n")));
			}
			#endif
		}
		else
		{
			SMSC_TRACE0(DBG_INIT,"\r\nFailure to Read Node Address from EEPROM\r\n");
			Status=NIC_STATUS_FAILURE;
			goto Exit_NICGetPermanentNodeAddress;
		}
		
	}


// See if whatever we read is a valid address

	for (i=0;i<6;i++)
	{
		ZeroCheck |= Adapter->PermanentNodeAddress[i];
		BcastCheck &= Adapter->PermanentNodeAddress[i];
	}

	if (
		(ZeroCheck == 0) ||
		(BcastCheck == 0xff) ||
		ETH_IS_MULTICAST(Adapter->PermanentNodeAddress)
	   )
	{
	

		/*
		 *	Hard code some value (0081071170nn, with nn the instance
		 *	number loaded)
		 */

		Adapter->PermanentNodeAddress[0] = 0x00;
		Adapter->PermanentNodeAddress[1] = 0x80;
		Adapter->PermanentNodeAddress[2] = 0x0F;
		Adapter->PermanentNodeAddress[3] = 0x11;
		Adapter->PermanentNodeAddress[4] = 0x70;
		Adapter->PermanentNodeAddress[5] = (BYTE) (Adapter->AdapterNumber & 0xff);

		#ifdef SMSC_DBG
		if (g_DebugMode & DBG_INIT) {
			int j;

			SMSC_TRACE0(DBG_INIT,"Permanent Node Address hardcoded in the driver :");
			for(j = 0; j < 6; j++)
				RETAILMSG(1,(TEXT(" %02x"), Adapter->PermanentNodeAddress[j]));
			RETAILMSG(1,(TEXT("\r\n")));
		}
		#endif
		

	}
	else
		
		SMSC_TRACE0(DBG_INIT,"Valid node address read from permanent storage \r\n");

	#ifdef SMSC_DBG
	if (g_DebugMode & DBG_INIT) {
		int j;

		SMSC_TRACE0(DBG_INIT,"Node Address returned as the permanent address:");
		for(j = 0; j < 6; j++)
			RETAILMSG(1,(TEXT(" %02x"),Adapter->PermanentNodeAddress[j]));
		RETAILMSG(1,(TEXT("\r\n")));
	}
	#endif
	

Exit_NICGetPermanentNodeAddress:
	SMSC_TRACE0(DBG_INIT,"<-- NICGetPermanentNodeAddress\r\n");

	return Status;
}
void 
StartLinkCheck(
    IN PMP_ADAPTER Adapter
    )
/*++

Routine Description:

    This routine starts the periodic link check

Arguments:

    Adapter - pointer to our Adapter

Return Value:

--*/
{
    
	SMSC_TRACE0(DBG_PHY,"--> StartLinkCheck\r\n");
    

	MP_SET_FLAG(Adapter, fMP_PERIODIC_LINK_CHECK);
	NdisMSetPeriodicTimer(
							&Adapter->LinkCheckTimer,
							LINK_CHECK_INTERVAL
						 );
	
	Adapter->Timercancelled=FALSE;
	SMSC_TRACE0(DBG_PHY,"<--StartLinkCheck\r\n");
}


NDIS_STATUS
SafeCancelTimer(
	IN PMP_ADAPTER Adapter,
	IN PNDIS_SPIN_LOCK Spinlock,
	IN PNDIS_MINIPORT_TIMER Timer,
	IN PNDIS_EVENT Event,
	IN UINT Timeout
	)
/*
Routine Description:

    This routine safely cancels the link timer waiting for it to fire it the 
    cancellation does not succeed.

Arguments:

    Adapter - pointer to our Adapter
    Timeout - amount of time to wait for Event signalling.

Return Value:

--*/
{
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;
    BOOLEAN           TimerCancelled;

	SMSC_TRACE0(DBG_SHUT_DOWN,"--> SafeCancelTimer\r\n");

	if (Adapter->Timercancelled)
		goto Exit_SafeCancelTimer;

	NdisAcquireSpinLock(Spinlock);

	NdisResetEvent(Event);
	NdisMCancelTimer(Timer, &TimerCancelled);

	NdisReleaseSpinLock(Spinlock);

	if (!TimerCancelled)
	{
		if (!NdisWaitEvent(Event, Timeout))
		{
			SMSC_TRACE1(DBG_ERROR,"ERROR: Timer Callback did not happen within %d milliseconds (2 * LINK_CHECK_INTERVAL) of it's timer cancellation!\n", Timeout);
			Status = NDIS_STATUS_FAILURE;
			goto Exit_SafeCancelTimer;
		}
	}
	else 
		Adapter->Timercancelled=TRUE;

Exit_SafeCancelTimer:
	SMSC_TRACE0(DBG_SHUT_DOWN,"<-- SafeCancelTimer\r\n");
    return Status;

}


/*++

Routine Description:

    This routine safely cancels the periodic link check (timer and workitem)

Arguments:

    Adapter - pointer to our Adapter

Return Value:

--*/

NDIS_STATUS
SafeCancelLinkCheck(
	IN PMP_ADAPTER Adapter
	)
{
    UINT Timeout = 0;
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;

    SMSC_TRACE0(DBG_SHUT_DOWN,"--> SafeCancelLinkCheck\r\n");

    // First make sure we will not service it any longer.
    MP_CLEAR_FLAG(Adapter, fMP_PERIODIC_LINK_CHECK);

	if (NDIS_STATUS_SUCCESS != SafeCancelTimer( Adapter,
												&Adapter->LCSpinlock,
												&Adapter->LinkCheckTimer,
												&Adapter->LinkCheckCalledEvent,
												LINK_CHECK_INTERVAL * 2))
	{
		Status = NDIS_STATUS_FAILURE;
		goto Exit_SafeCancelLinkCheck;
	}

	// Check if the workitem has finished and if not wait for it.
	while (MP_TEST_FLAG(Adapter, fMP_LINK_CHECK_IN_PROGRESS))
	{
		NdisMSleep(NDIS_SLEEP_1_MS);
		if (++Timeout >= NIC_2_SEC_TIMEOUT)
		{
			SMSC_TRACE0(DBG_ERROR,"ERROR: Timer Callback workitem did not happen complete within 2 seconds\r\n");
			Status = NDIS_STATUS_FAILURE;
			goto Exit_SafeCancelLinkCheck;
		}
	}

Exit_SafeCancelLinkCheck:
	SMSC_TRACE0(DBG_SHUT_DOWN,"<-- SafeCancelLinkCheck\r\n");

    return Status;
}

/*++

Routine Description:

	This routine initializes the PHY registers to enable or disable
	interrupt generation capabilities

Arguments:

	Adapter    Pointer to our adapter

Return Value:

	NIC_STATUS

--*/
NIC_STATUS
NICSetPHYInterruptCapability(
	IN PMP_ADAPTER Adapter,
	WORD IntMask
	)
{
	NIC_STATUS Status = NIC_STATUS_SUCCESS;
    DWORD Value32,dwTemp;

    if (!Adapter->PhyIsExternal)
    {
        // Write to the Interrupt source register to clear all status.
        dwTemp=0xFFFF;
        Status = NICWritePhy(Adapter, PHY_INT_SRC_FLG, dwTemp);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICSetPHYInterruptCapability;

        Status = NICReadPhy(Adapter, PHY_INT_SRC_FLG, &Value32);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICSetPHYInterruptCapability;
        SMSC_TRACE1(DBG_PHY,"1nd Read PHY_INT_SRC_FLG = 0x%04lx.\r\n", Value32);
        
        if (Value32)
        {
            // Write to the Interrupt source register to clear all status.
            dwTemp=0xFFFF;
            Status = NICWritePhy(Adapter, PHY_INT_SRC_FLG,dwTemp );
            if (Status != NIC_STATUS_SUCCESS)
                goto Exit_NICSetPHYInterruptCapability;

            Status = NICReadPhy(Adapter, PHY_INT_SRC_FLG, &Value32);
            if (Status != NIC_STATUS_SUCCESS)
                goto Exit_NICSetPHYInterruptCapability;
            SMSC_TRACE1(DBG_PHY,"2nd Read PHY_INT_SRC_FLG = 0x%04lx.\r\n", Value32);
        }        

        // Now setup the interrupt mask

        SMSC_TRACE1(DBG_PHY,"Writing 0x%04x to PHY_INT_MASK\r\n", IntMask);
        Status = NICWritePhy(Adapter, PHY_INT_MASK, IntMask);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICSetPHYInterruptCapability;
    }
    
Exit_NICSetPHYInterruptCapability:
   
	return(Status);
}

/*++

Routine Description:

	This routine indicates media connections and disconnections if we
	are not running from initilization

Arguments:

	Adapter 	Pointer to our adapter

Return Value:

	None.

--*/
NIC_STATUS
NICIndicateMediaStatus(
	IN PMP_ADAPTER Adapter
	)
{
	NDIS_STATUS Status;
//	BOOLEAN bCompleteReset = FALSE;
	NIC_STATUS	NicStatus=NIC_STATUS_SUCCESS;
	DWORD Value32;
		

	SMSC_TRACE0(DBG_PHY,"--> NICIndicateMediaStatus\r\n");

	Status = NDIS_STATUS_MEDIA_CONNECT;

    if (Adapter->SecondsElapsedAfterMPInitializeExit < Adapter->AnegCompleteTiomeout )
             Adapter->SecondsElapsedAfterMPInitializeExit = Adapter->AnegCompleteTiomeout;

	if (MP_TEST_FLAG(Adapter, fMP_DISCONNECTED))
	{
		Status = NDIS_STATUS_MEDIA_DISCONNECT;
		// Link disconnected
//		NICSetGpo(Adapter, Adapter->LinkLedOnGpio, LED_OFF);
		
		SMSC_TRACE0(DBG_INIT,"Indicating Disconnection\r\n");
	}
	else
	{
		Status = NDIS_STATUS_MEDIA_CONNECT;
		 // Link connected
//		NICSetGpo(Adapter, Adapter->LinkLedOnGpio, LED_ON);
        		
		SMSC_TRACE0(DBG_INIT,"Indicating Connection \r\n");
         

	}
/*
	// If we are NOT initializing check to see if we need to complete a reset
	if(!MP_TEST_FLAG(Adapter, fMP_INIT_IN_PROGRESS))
	{
		// If we exit MP_Reset PENDING and the link is up now then change the hardware
		// status to ready and complete the reset
		SMSC_TRACE0(DBG_PHY,"Not fMP_INIT_IN_PROGRESS\r\n");
		NdisAcquireSpinLock(&Adapter->RequestLock);
		if (
			(!MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS)) &&
			(Adapter->HardwareStatus == NdisHardwareStatusReset) &&
			(!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED))
		   )
		{
			Adapter->HardwareStatus = NdisHardwareStatusReady;

			bCompleteReset = TRUE;
		}
		NdisReleaseSpinLock(&Adapter->RequestLock);
	}
*/
	if(MP_TEST_FLAG(Adapter, fMP_REQUEST_PENDING))
		NICCompleteOidQueryIfPending(Adapter);

	if (Status == NDIS_STATUS_MEDIA_CONNECT) {
		SMSC_TRACE0(DBG_PHY,"NdisMIndicateStatus with NDIS_STATUS_MEDIA_CONNECT \r\n");
	}
	else {
		SMSC_TRACE0(DBG_PHY,"NdisMIndicateStatus with NDIS_STATUS_MEDIA_DISCONNECT \r\n");
	}


	NdisMIndicateStatus(
						Adapter->AdapterHandle,
						Status,
						NULL,
						0
					   );

	NdisMIndicateStatusComplete(Adapter->AdapterHandle);

	
	// Restart or stop Bulk in before indicating the link. This is particularly
	// important when the link is re-established, since there can be receives
	// in the context of the media indication
	if (!Adapter->StopBulkInOnRxEmpty && (Adapter->HardwareStatus == NdisHardwareStatusReady))
	{
		

		 if (!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED))
		{
			//
			//	Now we call NICStartRxPath().
			//

			NicStatus = NICStartRxPath(Adapter);
			if (NicStatus != NIC_STATUS_SUCCESS)
				goto Exit_NICIndicateMediaStatus;

		}
		else
		{
			// Link is down, stop the receive path. 		
			NicStatus = NICStopAndFlushRxPath(Adapter);
			// Leave it a clean state for when we need to restart. 
			
			if (NicStatus != NIC_STATUS_SUCCESS)
						 goto Exit_NICIndicateMediaStatus;
			Adapter->BulkInExpectedSeq = 0;
			Adapter->BulkInSubmitedSeq = 0;
			MP_CLEAR_FLAG(Adapter, fMP_POST_READS);

            if ((Adapter->SecondsElapsedAfterMPInitializeExit >= (Adapter->AnegCompleteTiomeout -1)) &&
				Adapter->SleepPhyOnLinkDown && (!Adapter->PhyIsExternal))
			{
				/*
						* Put the PHY in ED PD so that we can save some power.
						*/

				// If we are doing NWAY turn it off now
				if(Adapter->MediaType < MEDIA_TYPE_10HD_FORCED)
				{
					NicStatus = NICWritePhy(Adapter, PHY_CTRL, 0);
					if (NicStatus != NIC_STATUS_SUCCESS)
						goto Exit_NICIndicateMediaStatus;
					SMSC_TRACE0(DBG_PHY,"Turn off NWAY\r\n");
				}

				// Turn off Auto MDIX
				NicStatus = NICReadPhy(Adapter, PHY_MOD_CTL_STS, &Value32);
				if(NicStatus!= NIC_STATUS_SUCCESS)
					goto Exit_NICIndicateMediaStatus;
				Adapter->SaveAutoMdix = Value32;
				
				Value32 |= PHY_MOD_CTL_STS_AMDIX_DIS_;
				NicStatus = NICWritePhy(Adapter, PHY_MOD_CTL_STS, Value32);
				if(NicStatus != NIC_STATUS_SUCCESS)
					goto Exit_NICIndicateMediaStatus;

				NicStatus = NICReadPhy(Adapter, PHY_EXT_CTL_STS, &Value32);
                		if(NicStatus != NIC_STATUS_SUCCESS)
                    			goto Exit_NICIndicateMediaStatus;
				Value32 &= (~(PHY_EXT_CTL_STS_MDIX_AB_ | PHY_EXT_CTL_STS_MDIX_CD_));		
               		NicStatus = NICWritePhy(Adapter, PHY_EXT_CTL_STS, Value32 );
                		if(NicStatus != NIC_STATUS_SUCCESS)
                    			goto Exit_NICIndicateMediaStatus;
                		Adapter->SaveMdixState = Value32;

				SMSC_TRACE0(DBG_PHY,"Turn off AMDIX\r\n");
				

				// Clear any pending Phy interrupt and enable the mask for it
                NicStatus = NICSetPHYInterruptCapability(Adapter, PHY_INT_MASK_ENERGY_ON_);
                if (NicStatus != NIC_STATUS_SUCCESS)
                    goto Exit_NICIndicateMediaStatus;

                // Clear the WUPS bits to avoid spurious wake up
                NicStatus = NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
                if(NicStatus != NIC_STATUS_SUCCESS)
                    goto Exit_NICIndicateMediaStatus;

				Value32 |= ETH_PMT_CTL_WUPS_MASK_;
                NicStatus = NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
                if(NicStatus != NIC_STATUS_SUCCESS)
                    goto Exit_NICIndicateMediaStatus;

				// Set phy in ED low power mode.
				NicStatus = NICReadPhy(Adapter, PHY_MOD_CTL_STS, &Value32);
				if (NicStatus != NIC_STATUS_SUCCESS)
					goto Exit_NICIndicateMediaStatus;

				Value32 |= PHY_MOD_CTL_STS_EDPWRDOWN_;
				NicStatus = NICWritePhy(Adapter, PHY_MOD_CTL_STS, Value32);
				if (NicStatus != NIC_STATUS_SUCCESS)
					goto Exit_NICIndicateMediaStatus;

				Adapter->PhyLinkPowerDown = TRUE;
		  	}

            if ((Adapter->SecondsElapsedAfterMPInitializeExit >= (Adapter->AnegCompleteTiomeout-1)) &&
                (Adapter->DetachOnLinkDownMode == DOLD_ENABLED))
            {

                SMSC_TRACE0(DBG_INIT,"Setting up Smart Detach\r\n");

                if (!Adapter->SleepPhyOnLinkDown && (!Adapter->PhyIsExternal))
                {
                SMSC_TRACE0(DBG_PHY,"Don't sleep phy on link down\r\n");
                    // Phy interrupts as reconnect source ?
                    if (Adapter->LinkWakeModes & 0xFF)
                    {
                        // Clear any pending Phy interrupt and enable the mask for it
                        SMSC_TRACE0(DBG_PHY,"Enabling PHY interrupts as reconnect Source\r\n");

                        NicStatus = NICSetPHYInterruptCapability(Adapter, ((WORD)(Adapter->LinkWakeModes & 0xFFFF)));

                        // Clear the WUPS bits to avoid spurious wake up
                        NicStatus = NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
                        if(NicStatus != NIC_STATUS_SUCCESS)
                            goto Exit_NICIndicateMediaStatus;
						Value32 |= ETH_PMT_CTL_WUPS_MASK_;
                        NicStatus = NICWriteRegister(Adapter, ETH_PMT_CTL,&Value32 );
                        if(NicStatus != NIC_STATUS_SUCCESS)
                            goto Exit_NICIndicateMediaStatus;

                        // Set all supported media in NWAY Adver. Register in case it's disabled by forced link
                        // 10/100 FD/HD
                        NicStatus = NICReadPhy(Adapter, PHY_AUTONEG_ADV, &Value32);
                        if(NicStatus != NIC_STATUS_SUCCESS)
                            goto Exit_NICIndicateMediaStatus;
						Value32 |= NWAY_AR_ALL_CAPS_;
                        NicStatus = NICWritePhy(Adapter, PHY_AUTONEG_ADV,Value32 );
                        if(NicStatus != NIC_STATUS_SUCCESS)
                            goto Exit_NICIndicateMediaStatus;
                    
                        // 1000 FD
                        NicStatus = NICReadPhy(Adapter, PHY_1000T_CTRL, &Value32);
                        if (NicStatus != NIC_STATUS_SUCCESS)
                            goto Exit_NICIndicateMediaStatus;

                        Value32 &= (~CR_1000T_HD_CAPS_);
                        Value32 |= CR_1000T_FD_CAPS_;

                        NicStatus = NICWritePhy(Adapter, PHY_1000T_CTRL, Value32);
                        if (NicStatus != NIC_STATUS_SUCCESS)
                            goto Exit_NICIndicateMediaStatus;
						Value32=MII_CR_AUTO_NEG_EN_ | MII_CR_RESTART_AUTO_NEG_;
                        // Enable NWAY in PHY_BMC in case it's disabled by forced link
                        NicStatus = NICWritePhy(Adapter, PHY_CTRL, Value32);
                        if(NicStatus != NIC_STATUS_SUCCESS)
                            goto Exit_NICIndicateMediaStatus;
                    }
                    

                    // Link interrupt as reconnect source ?
                    if (Adapter->LinkWakeModes & ETH_GPIO_WAKE_PHY_LINKUP_EN_)
                    {
                        SMSC_TRACE1(DBG_PHY,"Enabling Link on GPIO%d as reconnect Source\r\n",
                                                   Adapter->SDReconnectGpio == 12 ? 7 : Adapter->SDReconnectGpio);

                        // Setup Linkup for either internal (GPIO7) or external GPIO that carries link.
                        // Do it while preserving values for every other bit in this register.
                        NicStatus = NICReadRegister(Adapter, ETH_GPIO_WAKE, &Value32);
                        if (Adapter->SDReconnectGpio == 12)
                        {
                            // Enable wake on link for GPIO7 with polarity set to "driven low"
                            Value32 = ( 
                                        ETH_GPIO_WAKE_PHY_LINKUP_EN_ |
                                        (Value32| ETH_GPIO_WAKE_GPIOPOL7_) | 
                                        ETH_GPIO_WAKE_GPIOWK7_ 
                                      );
                        }
                        else
                        {
                           
                           // Enable wake on GPIO<n> with polarity set to "driven low"
                            // Clear wake on link and polarity bit for this GPIO.
                            Value32 &= (~(ETH_GPIO_WAKE_PHY_LINKUP_EN_ | (ETH_GPIO_WAKE_GPIOPOL0_ << Adapter->SDReconnectGpio)));

                            // Enable GPIO for wakes and set the registry defined polarity for it.
                            Value32 |= ((ETH_GPIO_WAKE_GPIOWK0_ | (Adapter->SDReconnectGpioPol << 16) ) << Adapter->SDReconnectGpio); 
                        }

                        NicStatus = NICWriteRegister(Adapter, ETH_GPIO_WAKE, &Value32);
                        if (NicStatus != NIC_STATUS_SUCCESS)
                            goto Exit_NICIndicateMediaStatus;

                        // Set GPIO7 as input
                        NicStatus = NICReadRegister( Adapter, ETH_GPIO_CFG, &Value32);
                        if (NicStatus != NIC_STATUS_SUCCESS)
                            goto Exit_NICIndicateMediaStatus;
                        
                        if (Adapter->SDReconnectGpio == 12)
                        {
                            // Enable GPIO7 as input; data and type are don't care.
                            Value32 &= (~((ETH_GPIO_CFG_GPO4_EN_ | ETH_GPIO_CFG_GPO4_DIR_) << 3));  
                        }
                        else
                        {
                            if (Adapter->SDReconnectGpio >= 4)
                            {
                                // Enable GPIO<n> as input; data and type are don't care.
                                Value32 &= (~((ETH_GPIO_CFG_GPO4_EN_ | ETH_GPIO_CFG_GPO4_DIR_) << (Adapter->SDReconnectGpio - 4))); 
                            }
                            else
                            {
                                DWORD dwTemp32;
                                
                                NicStatus = NICReadRegister( Adapter, ETH_LED_GPIO_CFG0, &dwTemp32);
                                if (NicStatus != NIC_STATUS_SUCCESS)
                                    goto Exit_NICIndicateMediaStatus;

                                // Enable GPIO<n> as input; data and type are don't care.
                                dwTemp32 &= (~((ETH_LED_GPIO_CFG0_GPIOEN0_ | ETH_LED_GPIO_CFG0_GPDIR0_) << Adapter->SDReconnectGpio)); 

                                NicStatus = NICWriteRegister( Adapter, ETH_LED_GPIO_CFG0, &dwTemp32);
                                if (NicStatus != NIC_STATUS_SUCCESS)
                                    goto Exit_NICIndicateMediaStatus;

                            }
                        }

                        NicStatus = NICWriteRegister( Adapter, ETH_GPIO_CFG, &Value32);
                        if (NicStatus != NIC_STATUS_SUCCESS)
                            goto Exit_NICIndicateMediaStatus;
                        
                        // Clear All GPIO Wake Events
                        Value32= ETH_INT_STS_GPIO_;
                        NicStatus = NICWriteRegister( Adapter, ETH_INT_STS, &Value32);
                    }
                }

                               
                
                NicStatus = NICReadRegister(Adapter, ETH_HW_CFG, &Value32);
                if (NicStatus != NIC_STATUS_SUCCESS)
                {
                    SMSC_TRACE0(DBG_PHY,"Error reading ETH_HW_CFG in NICIndicateMediaStatus\r\n");
                    goto Exit_NICIndicateMediaStatus;
                }

                
                SMSC_TRACE0(DBG_INIT,"Enabling Smart Detach now...\r\n");

                // Now do the Smart Detach !
                // USB bus will be disconnected as soon as ETH_HW_CFG_SMDET_EN_ bit is set. 
                // So from this point all USB transactions will fail including the following register write.
                Value32 |= ETH_HW_CFG_SMDET_EN_;
                NicStatus = NICWriteRegister(Adapter, ETH_HW_CFG,&Value32);
                if (NicStatus != NIC_STATUS_SUCCESS)
                {
                    // At this point we don't know if NICWriteRegister failed for any reason other than Smart Detach.
                    // Let's try the following read to make sure USB bus is really detached
                    NicStatus = NICReadRegister(Adapter, ETH_HW_CFG, &Value32);
                    if (NicStatus != NIC_STATUS_SUCCESS)
                    {
                        NicStatus = NIC_STATUS_SUCCESS; // sucessfully detached
                    }
                    else
                    {
                        SMSC_TRACE0(DBG_PHY,"USB bus is still attached after setting Smart Detach Bit!!!\r\n");
                        SMSC_TRACE1(DBG_PHY,"ETH_HW_CFG=0x%08lX\r\n", Value32);
                        NicStatus = NIC_STATUS_FAILURE; // Detach failed
                    }
                    goto Exit_NICIndicateMediaStatus;
                }
            }
		}
	}
		


	#ifdef USE_INT_PIPE
	   if (Adapter->StopBulkInOnRxEmpty && (Adapter->HardwareStatus == NdisHardwareStatusReady))
	   {
		   
	
			if (!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED))
		   {
			   
			   MP_SET_FLAG(Adapter, fMP_POST_INTS);
			   NICStartInterruptPipe(Adapter);
		   }
		   else
		   {
			   // Link is down, stop submitting bulk in requests.
			   MP_CLEAR_FLAG(Adapter, fMP_POST_INTS);
			   NICStopInterruptPipe(Adapter);
		   }
	   }
	#endif
	

	// We've already indicated status, so skip the "indicate if disconnected
	// after two seconds of returning from MP_Initialize"
//	Adapter->SecondsElapsedAfterMPInitializeExit = Adapter->AnegCompleteTiomeout;
/*
	if (bCompleteReset)
	{

	SMSC_TRACE1(DBG_RESET,"NicIndicateSattus Reset = 0x%x \r\n", NDIS_STATUS_SUCCESS);
		NdisMResetComplete(
							Adapter->AdapterHandle,
							NDIS_STATUS_SUCCESS,
							FALSE
						  );

		Adapter->SecondsElapsedAfterMPResetExit = Adapter->MpRstCompleteTimeout;
	}
*/
	SMSC_TRACE0(DBG_PHY,"<-- NICIndicateMediaStatus\r\n");
	
Exit_NICIndicateMediaStatus:
	return NicStatus;
}

NIC_STATUS 
NICSetReceiveMask(
    IN PMP_ADAPTER Adapter,
    DWORD PacketFilter
    )
/*++

Routine Description:

    This routine Sets the receive Mask 

Arguments:

    Adapter         Pointer to our adapter
    PacketFilter    The ndis packet filter used to set the Rx mask

Return Value:

    TRUE if read successful, FALSE otherwise.

--*/
{
    DWORD Value32;
    NIC_STATUS Status = NIC_STATUS_SUCCESS;

    SMSC_TRACE0(DBG_REQUEST,"--->NICSetReceiveMask\r\n");
	
    Status = NICReadRegister(Adapter, ETH_RFE_CTL, &Value32);
    if (Status != NIC_STATUS_SUCCESS)
    {
        goto Exit_NICSetReceiveMask;
    }

    if(PacketFilter & NDIS_PACKET_TYPE_PROMISCUOUS)
    {
        // No filter in promiscuous mode
        Value32 &= (~(ETH_RFE_CTL_DA_PERFECT_ | ETH_RFE_CTL_MCAST_HASH_));
        Value32 |= (ETH_RFE_CTL_UCAST_EN_ | ETH_RFE_CTL_MCAST_EN_ | ETH_RFE_CTL_BCAST_EN_);
    }
    else
    {
        // Clear all filter types
        Value32 &= (~(ETH_RFE_CTL_UCAST_EN_ | ETH_RFE_CTL_MCAST_EN_ | ETH_RFE_CTL_BCAST_EN_ | ETH_RFE_CTL_MCAST_HASH_));
        
        // Setup filters
        // At the minimum, NIC receives it's own MAC address w/ DA perfect filtering
        Value32 |= ETH_RFE_CTL_DA_PERFECT_;
        
        // ACCEPT_BROADCAST
        if (PacketFilter & NDIS_PACKET_TYPE_BROADCAST)
        {
            Value32 |= ETH_RFE_CTL_BCAST_EN_;
        }

        // ACCEPT_MULTI_PROM
        if (PacketFilter & NDIS_PACKET_TYPE_ALL_MULTICAST)
        {
            Value32 |= ETH_RFE_CTL_MCAST_EN_;
        }
        else
        {
            // only ACCEPT_MULTICAST
            if (PacketFilter & NDIS_PACKET_TYPE_MULTICAST)
            {
                    #ifdef DO_NOT_USE_MC_HASH_FILTER
                        // If driver can not use HASH filtering, then it should receive all MC packet and 
                        // filter it in the software
                        // -- fixme It can still use perfect DA filtering
                        Value32 |= ETH_RFE_CTL_MCAST_EN_;
                    #else
                        { // --fixme we may not need HASH filter setting here
                            int hIndex;
                            
                            for (hIndex=0; hIndex<16; hIndex++)
                            {
//                                SMSC_TRACE2(DBG_REQUEST,"HASH[%d] = 0x%08lx\r\n", hIndex, Adapter->MulticastHashTable[hIndex]);
                                Status = NICWriteHashRam32(Adapter, hIndex, Adapter->MulticastHashTable[hIndex]);
                                if (Status != NIC_STATUS_SUCCESS)
                                {
                                    goto Exit_NICSetReceiveMask;
                                }
                            }
                        }
                        Value32 |= ETH_RFE_CTL_MCAST_HASH_;
                    #endif    
            }
        }
    }

    Status = NICWriteRegister(Adapter, ETH_RFE_CTL, &Value32);
    if (Status != NIC_STATUS_SUCCESS)
    {
        goto Exit_NICSetReceiveMask;
    }

    // Check Rx BAD Packet filter
    Status = NICReadRegister(Adapter, ETH_FCT_RX_CTL, &Value32);
    if (Status != NIC_STATUS_SUCCESS)
    {
        goto Exit_NICSetReceiveMask;
    }

    if(Adapter->RxBadPkts)
        Value32 |= ETH_FCT_RX_CTL_SBF_;
    else
        Value32 &= (~ETH_FCT_RX_CTL_SBF_);

    Status = NICWriteRegister(Adapter, ETH_FCT_RX_CTL, &Value32);
    if (Status != NIC_STATUS_SUCCESS)
    {
        goto Exit_NICSetReceiveMask;
    }

    #ifdef SMSC_DBG
    Status = NICReadRegister(Adapter, ETH_RFE_CTL, &Value32);
    if (Status != NIC_STATUS_SUCCESS)
    {
        goto Exit_NICSetReceiveMask;
    }
    SMSC_TRACE1(DBG_REQUEST,"ETH_RFE_CTL = 0x%08lx\r\n", Value32);

    Status = NICReadRegister(Adapter, ETH_FCT_RX_CTL, &Value32);
    if (Status != NIC_STATUS_SUCCESS)
    {
        goto Exit_NICSetReceiveMask;
    }
    SMSC_TRACE1(DBG_REQUEST,"ETH_FCT_RX_CTL = 0x%08lx\r\n", Value32);
    #endif

    // Save the new packet filter value                                                               
    Adapter->PacketFilter = PacketFilter;

Exit_NICSetReceiveMask:
    SMSC_TRACE1(DBG_REQUEST,"<---NICSetReceiveMask, %d\r\n", Status);

    return (Status); 
} /* NICSetReceiveMask */

NIC_STATUS shape_phy(PMP_ADAPTER Adapter)
{
        int delay = 500;
		DWORD dwTemp,dwValue32;
		NIC_STATUS ret=NIC_STATUS_FAILURE;
        
        //set bank3
        dwTemp=0x000000FF;
        if (NICWritePhy(Adapter, PHY_SPCL_TST_WRITE, dwTemp)==NIC_STATUS_FAILURE)
			goto Done;
		
		dwValue32=(7 | PHY_SPCL_TST_CNTL_WRITE_);
        if (NICWritePhy(Adapter, PHY_SPCL_TST_CNTL, dwValue32)==NIC_STATUS_FAILURE)
			goto Done;
		
        do
        {
                NdisMSleep(100);
                ret=NICReadPhy(Adapter, PHY_SPCL_TST_CNTL, &dwTemp);
                if(!(dwTemp & PHY_SPCL_TST_CNTL_WRITE_) || (ret==NIC_STATUS_FAILURE))
                {
                        break;
                }
        }while(--delay);

		ret=NICReadPhy(Adapter, PHY_SPCL_TST_CNTL, &dwTemp);
		dwValue32=(dwTemp | PHY_SPCL_TST_CNTL_ADVANCE_REG_);
        ret=NICWritePhy(Adapter, PHY_SPCL_TST_CNTL, dwValue32);
		
        if(!delay || (ret==NIC_STATUS_FAILURE))
        {
                SMSC_TRACE0(DBG_PHY,"PHY shape Failed\r\n");
        }

		
		
    	ret = NIC_STATUS_SUCCESS;
Done:		
        return ret;

}


/*++

Routine Description:

	This routine initializes the MII registers, determines line speed,
	and media type.

Arguments:

	Adapter 	Pointer to our adapter

Return Value:

	TRUE if successful, FALSE otherwise.

--*/
NIC_STATUS NICInitializeMedia(
	IN PMP_ADAPTER Adapter
	)
{
	int Count=0;
	NIC_STATUS Status=NIC_STATUS_FAILURE;
	DWORD Value32,dwTemp;	
	DWORD dwId1, dwId2;

	SMSC_TRACE0(DBG_INIT,"--> NICInitializeMedia\r\n");

	NdisMSleep(NDIS_SLEEP_10_MS);	// As per Phy requirement on resume.

	
	Status = NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
	SMSC_TRACE1(DBG_INIT,"ETH_PMT_CTL = 0x%x\r\n",Value32);
			
	
	// Reset the PHY using it's (hw) reset signal from the MAC's PM_CTRL reg

	Status = NICReadRegister(Adapter, ETH_PMT_CTL, &Value32 );
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICInitializeMedia;

    //dwTemp|= ETH_PMT_CTL_PHY_RST_;
	dwTemp=(Value32 & (~ETH_PMT_CTL_WUPS_MASK_)) | ETH_PMT_CTL_PHY_RST_;
	Status = NICWriteRegister(Adapter, ETH_PMT_CTL, &dwTemp);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICInitializeMedia;

	Adapter->PhyLinkPowerDown = FALSE;

	for(Count=0;Count<NIC_HALF_SEC_TIMEOUT;Count++)
	{
		Status = NICReadRegister(Adapter, ETH_PMT_CTL, &Value32 );
		if (Status != NIC_STATUS_SUCCESS)
			break;

		if(!(Value32 & ETH_PMT_CTL_PHY_RST_))
			break;
		NdisMSleep(NDIS_SLEEP_1_MS);
	}

	SMSC_TRACE3(DBG_INIT,"PHY HW Reset took <=%d ticks of %d ms each. Aprox = %d ms.\r\n",
						Count+1, NDIS_SLEEP_1_MS/1000, (Count+1)*NDIS_SLEEP_1_MS/1000 );


	if ((Count==NIC_HALF_SEC_TIMEOUT) || (Status != NIC_STATUS_SUCCESS))
	{
		SMSC_TRACE0(DBG_INIT,"MII PHY did not complete reset");
		Status = NIC_STATUS_FAILURE;
		goto Exit_NICInitializeMedia;
	}

	 // Enable MII management Interface

	if (!Adapter->PhyIsExternal)
	{
		Adapter->PHYAddress = 1;
		SMSC_TRACE1(DBG_INIT,"Internal PHY. PHY Address = %d\r\n", Adapter->PHYAddress);
	   
	}
	else
	{
		// Using external PHY board on MII connector

		// If we select a Phy address higher than 31 then we want to search for the PHY
		if(Adapter->PHYAddress > 0x1f)
		{
			SMSC_TRACE0(DBG_INIT,"External PHY. Autodetecting address... \r\n");
			// Try to auto detect PHY address, if not found default to 1.
			for(Count=31;Count>=0;Count--)
			{
				Adapter->PHYAddress = (BYTE) Count;
				Status = NICReadPhy(Adapter, PHY_STATUS, &Value32);
				Status |= NICReadPhy(Adapter, PHY_STATUS, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
				{
					break;
				}

				if(
					((Value32 & 0xffff) != 0xffff) &&
					((Value32 & 0xffff) != 0x0000)
				  )
				{
					SMSC_TRACE1(DBG_INIT,"Detected PHY Address = %d\r\n", Count);
					break;
				}
			}

			if (Status != NIC_STATUS_SUCCESS)
			{
				goto Exit_NICInitializeMedia;
			}

			// if we did not find it default back to Id = 1.
			if (Count < 0)
			{
				Adapter->PHYAddress = 0x01;
				SMSC_TRACE1(DBG_INIT,"PHY Address not detected. Defaulting to %d\r\n", Adapter->PHYAddress);
			}
		} //if(pDevice->PHYAddress > 0x1f)
		else
		{
			BYTE i;
			DWORD SaveAddr = Adapter->PHYAddress;

			SMSC_TRACE1(DBG_INIT,"External PHY. Manually PHY Address = %d\r\n", Adapter->PHYAddress);

			// First isolate all PHYs we can find...

			for (i=0;i<0x20;i++)
			{
				Adapter->PHYAddress = i;
				Status = NICWritePhy(Adapter, PHY_STATUS, MII_CR_ISOLATE_);
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICInitializeMedia;
			
				Status = NICReadPhy(Adapter, PHY_STATUS, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICInitializeMedia;

				SMSC_TRACE2(DBG_INIT,"PHY %02d BMC = %04X\r\n", i, Value32);

			}

			// Now put the PHY at the registry parsed address out of isolation. 
			Adapter->PHYAddress = SaveAddr;

			Status = NICWritePhy(Adapter, PHY_STATUS, 0);	
			if (Status != NIC_STATUS_SUCCESS)
				goto Exit_NICInitializeMedia;

		} //if(pDevice->PHYAddress > 0x1f)
	}


	
	// Read PHY Id information

	Status = NICReadPhy(Adapter, PHY_ID1, &dwId1);
	if(Status != NIC_STATUS_SUCCESS)
		goto Exit_NICInitializeMedia;
	

	Status = NICReadPhy(Adapter, PHY_ID2, &dwId2);
	if(Status != NIC_STATUS_SUCCESS)
		goto Exit_NICInitializeMedia;
	

	Adapter->PHY_OUI = (dwId1 << 2) | ((dwId2 & 0xFC00) << 8);
    Adapter->PHY_Model = (UCHAR) ((dwId2 & 0x03F0UL) >> 4);
    Adapter->PHY_Revision = (UCHAR) (dwId2 & 0x000f);

	SMSC_TRACE3(DBG_INIT,"PHY OUI = %X	  PHY Model = %X	PHY Revision = %X\r\n",
						Adapter->PHY_OUI, Adapter->PHY_Model,
						Adapter->PHY_Revision);


	// Apply PHY Workaround for ANEG long completion bug in
	   // CL PHY Revision 0xB
	   if ((Adapter->PHY_OUI == 0xC0001CUL) &&
		   (Adapter->PHY_Model == 0xEUL) &&
		   (Adapter->PHY_Revision == 0xBUL))
	   {
		   int i;
	
		   typedef struct _PHY_STUFFING_
		   {
			   DWORD   dwAddr;
			   DWORD   dwVal;
		   } PHY_STUFFING;
	
		   PHY_STUFFING cStuffings[]=
		   {
			   {0x17UL, 0x006dUL},
			   {0x14UL, 0x40F6UL},
			   {0x17UL, 0x008dUL},
			   {0x14UL, 0x40F6UL},
			   {0x17UL, 0x3436UL},
			   {0x14UL, 0x40e4UL},
			   {0x14UL, 0x40e5UL},
			   {0x14UL, 0x40e6UL},
			   {0x14UL, 0x40e7UL},
			   {0x17UL, 0x74e0UL},
			   {0x14UL, 0x408eUL},
			   {0x17UL, 0x5b53UL},
			   {0x14UL, 0x408fUL},
			   {0x17UL, 0x02a3UL},
			   {0x14UL, 0x4091UL},
			   {0x17UL, 0x15ffUL},
			   {0x14UL, 0x40abUL},
			   {0x17UL, 0x6419UL},
			   {0x14UL, 0x40a1UL},
			   {0x17UL, 0x320cUL},
			   {0x14UL, 0x40a2UL},
			   {0x17UL, 0x0072UL},
			   {0x14UL, 0x4138UL},
			   {0x17UL, 0x0074UL},
			   {0x14UL, 0x4139UL},
			   {0x17UL, 0x0076UL},
			   {0x14UL, 0x413aUL},
			   {0x17UL, 0x0078UL},
			   {0x14UL, 0x413bUL},
			   {0x17UL, 0x007aUL},
			   {0x14UL, 0x413cUL},
			   {0x17UL, 0x007eUL},
			   {0x14UL, 0x413dUL},
			   {0x17UL, 0x0077UL},
			   {0x14UL, 0x413eUL},
			   {0x14UL, 0x413fUL},
			   {0x09UL, 0x0300UL},
			   {0x04UL, 0x01e1UL}
		   };
	
		   SMSC_TRACE0(DBG_INIT,"CL PHY Reset workaround in CL PHY Revision 0xB\r\n");
		   
		   for (i = 0; i < (sizeof(cStuffings)/sizeof(PHY_STUFFING)); i++)
		   {
			   Status = NICWritePhy(Adapter, cStuffings[i].dwAddr, cStuffings[i].dwVal);
			   if(Status != NIC_STATUS_SUCCESS)
			   {
				   SMSC_TRACE0(DBG_INIT,"Writing to PHY failed!!!\r\n");
				   break;
			   }
		   }
	   }

	shape_phy(Adapter);

	
	// Disable PHY to generate interrupts
	Status = NICSetPHYInterruptCapability(Adapter, 0);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICInitializeMedia;


	if (!Adapter->PhyIsExternal)
	{
		if (Adapter->AutoMdix >= AMDIX_USE_STRAP)
		{
		/*
			// Disabled, make sure it really is.

			Status = NICReadPhy(Adapter, PHY_THERA_CTL_STS_IND, &Value32);
			if(Status != NIC_STATUS_SUCCESS)
				goto Exit_NICInitializeMedia;
		 Value32 &= 0x1FFF; 
			Status = NICWritePhy(Adapter, PHY_THERA_CTL_STS_IND, Value32);

		   Status = NICReadRegister(Adapter, ETH_STRAP_DEBUG, &Value32 );
			if(Status != NIC_STATUS_SUCCESS)
				goto Exit_NICInitializeMedia;
			if (Value32 & ETH_STRAP_DEBUG_AMDIX_EN_)
			{
				SMSC_TRACE0(DBG_INIT,"Auto-MDIX enabled by strap default\r\n");
			}
			else 
			{
				SMSC_TRACE0(DBG_INIT,"Auto-MDIX disabled by strap default\r\n");
			}
		*/   
		}
		  else
        {
            // Enabled, set it up.

            Status = NICReadPhy(Adapter, PHY_MOD_CTL_STS, &Value32);
            if(Status != NIC_STATUS_SUCCESS)
                goto Exit_NICInitializeMedia;

            if (Adapter->AutoMdix & AMDIX_ENABLE) 
            {
            	  Value32=Value32 & (~PHY_MOD_CTL_STS_AMDIX_DIS_);
                Status = NICWritePhy(Adapter, PHY_MOD_CTL_STS, Value32 );
                if(Status != NIC_STATUS_SUCCESS)
                    goto Exit_NICInitializeMedia;

                SMSC_TRACE0(DBG_INIT,"Override Strap, Enable Auto-MDIX\r\n");
            } 
            else if (Adapter->AutoMdix & AMDIX_DISABLE_CROSSOVER) 
            {
            	Value32 |= PHY_MOD_CTL_STS_AMDIX_DIS_;
                Status = NICWritePhy(Adapter, PHY_MOD_CTL_STS, Value32 );
                if(Status != NIC_STATUS_SUCCESS)
                    goto Exit_NICInitializeMedia;

                Status = NICReadPhy(Adapter, PHY_EXT_CTL_STS, &Value32);
                if(Status != NIC_STATUS_SUCCESS)
                    goto Exit_NICInitializeMedia;

				Value32 |= PHY_EXT_CTL_STS_MDIX_AB_ | PHY_EXT_CTL_STS_MDIX_CD_;
                Status = NICWritePhy(Adapter, PHY_MOD_CTL_STS, Value32 );
                if(Status != NIC_STATUS_SUCCESS)
                    goto Exit_NICInitializeMedia;

                SMSC_TRACE0(DBG_INIT,"Override Strap, Disable Auto-MDIX, CrossOver Cable\r\n");
            } 
            else 
            {
            	Value32 |= PHY_MOD_CTL_STS_AMDIX_DIS_;
                Status = NICWritePhy(Adapter, PHY_MOD_CTL_STS, Value32);
                if(Status != NIC_STATUS_SUCCESS)
                    goto Exit_NICInitializeMedia;

                Status = NICReadPhy(Adapter, PHY_EXT_CTL_STS, &Value32);
                if(Status != NIC_STATUS_SUCCESS)
                    goto Exit_NICInitializeMedia;

			 Value32 &= (~(PHY_EXT_CTL_STS_MDIX_AB_ | PHY_EXT_CTL_STS_MDIX_CD_));
                Status = NICWritePhy(Adapter, PHY_MOD_CTL_STS, Value32 );
                if(Status != NIC_STATUS_SUCCESS)
                    goto Exit_NICInitializeMedia;

                SMSC_TRACE0(DBG_INIT,"Override Strap, Disable Auto-MDIX, Straight Cable\r\n");
            }
        }
    }
    

	// Now lets really setup the media.

	Adapter->LinkSpeed = LINK_SPEED_UNKNOWN;
	Adapter->LinkDuplex = LINK_DUPLEX_UNKNOWN;
	Adapter->bRespondToPause = FALSE;
    Adapter->bGeneratePause = FALSE;

	if(Adapter->MediaType < MEDIA_TYPE_10HD_FORCED)
	{
		switch (Adapter->MediaType)
		{
			case MEDIA_TYPE_AUTO_FULL:
				SMSC_TRACE0(DBG_INIT,"Auto 10/100/1000\r\n");

				Adapter->AdvertisedCapabilites = PHY_ANA_ALL_SUPPORTED_MEDIA;
				
				if (Adapter->UsbSpeedIsHigh)
                {
                    Adapter->AdvertisedCapabilites = PHY_ANA_ALL_SUPPORTED_MEDIA;
                }
                else
                {
                    Adapter->AdvertisedCapabilites = (PHY_ANA_10BASE_T_HD | PHY_ANA_10BASE_T_FD);
                    SMSC_TRACE1(DBG_INIT,"Device operating at USB full speed so reducing autonegotiation AdvertisedCapabilites to 10Mbps only. Overiding to 0x%08lx\n",
                                    Adapter->AdvertisedCapabilites);
                    
                }
				break;

			case MEDIA_TYPE_10HD_AUTO:
				SMSC_TRACE0(DBG_INIT,"Auto 10Half\r\n");
				Adapter->AdvertisedCapabilites = PHY_ANA_10BASE_T_HD;
				break;

			case MEDIA_TYPE_10FD_AUTO:
				SMSC_TRACE0(DBG_INIT,"Auto 10Full\r\n");
				Adapter->AdvertisedCapabilites = PHY_ANA_10BASE_T_FD;
				break;

			case MEDIA_TYPE_100HD_AUTO:
				SMSC_TRACE0(DBG_INIT,"Auto 100Half\r\n");
				Adapter->AdvertisedCapabilites = PHY_ANA_100BASE_TX_HD;
				break;

			case MEDIA_TYPE_100FD_AUTO:
				SMSC_TRACE0(DBG_INIT,"Auto 100Full\r\n");
				Adapter->AdvertisedCapabilites = PHY_ANA_100BASE_TX_FD;
				break;

			case MEDIA_TYPE_1000FD_AUTO:
				SMSC_TRACE0(DBG_INIT,"Auto 1000 Full\r\n");
				Adapter->AdvertisedCapabilites = PHY_ANA_1000BASE_T_FD;
				break;

			default:
				break;
		}

		 // Setup Advertisement register
        Status = NICReadPhy(Adapter, PHY_AUTONEG_ADV, &Value32);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICInitializeMedia;

        // Reset Capabilities
        Value32 &= (~NWAY_AR_ALL_CAPS_);

        // Set Media Type Caps. for 10/100
        Value32 |= (Adapter->AdvertisedCapabilites & 0x0F) << 5;

        // If flow control is enabled and a FC capability is 
        // advertised then also advertise flow control pause type
        if (
                (Adapter->FlowControl) &&
                (               
                    (Adapter->MediaType == MEDIA_TYPE_AUTO_FULL)  ||
                    (Adapter->MediaType == MEDIA_TYPE_10FD_AUTO)  ||
                    (Adapter->MediaType == MEDIA_TYPE_100FD_AUTO) ||
                    (Adapter->MediaType == MEDIA_TYPE_1000FD_AUTO)
                )
           )
        {
            Value32 |= ((Value32 & (~NWAY_AR_PAUSE_MASK)) | (Adapter->FlowControl << 10)); 
        }

		

        Status = NICWritePhy(
                                Adapter, 
                                PHY_AUTONEG_ADV, 
                                Value32
                            );
        if(Status != NIC_STATUS_SUCCESS)
            goto Exit_NICInitializeMedia;

		SMSC_TRACE1(DBG_INIT,"PHY_AUTONEG_ADV = 0x%x\r\n",Value32);
/*
        // 1000Base-T Capability
        if (Adapter->AdvertisedCapabilites & MEDIA_TYPE_1000FD_AUTO)
        {
            // Setup Advertisement register
            Status = NICReadPhy(Adapter, PHY_1000T_CTRL, &Value32);
            if (Status != NIC_STATUS_SUCCESS)
                goto Exit_NICInitializeMedia;

            // Only 1000BASE-T FD, Bothwell can not operate in 1000BASE-T HD
            Value32 &= (~CR_1000T_HD_CAPS_);
            Status = NICWritePhy(
                                    Adapter, 
                                    PHY_1000T_CTRL, 
                                    (Value32 | CR_1000T_FD_CAPS_)
                                );
            if(Status != NIC_STATUS_SUCCESS)
                goto Exit_NICInitializeMedia;
        }
*/

		// Setup Advertisement register for 1000FD
		Status = NICReadPhy(Adapter, PHY_1000T_CTRL, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICInitializeMedia;
		
		// Only 1000BASE-T FD, Bothwell can not operate in 1000BASE-T HD
		Value32 &= (~CR_1000T_HD_CAPS_);
		// 1000Base-T Capability
		if (Adapter->AdvertisedCapabilites & PHY_ANA_1000BASE_T_FD)
		{
			Value32 |= CR_1000T_FD_CAPS_;
		}
		else
		{
			Value32 &= (~CR_1000T_FD_CAPS_);
		}
				
		Status = NICWritePhy(
							Adapter, 
							PHY_1000T_CTRL, 
							Value32
				            );
		if(Status != NIC_STATUS_SUCCESS)
		    goto Exit_NICInitializeMedia;


		/* Auto detect link speed and duplex mode */
		dwTemp=MII_CR_AUTO_NEG_EN_ | MII_CR_RESTART_AUTO_NEG_;
		Status = NICWritePhy(
								Adapter,
								PHY_CTRL,
								dwTemp
							);
		if(Status != NIC_STATUS_SUCCESS)
			goto Exit_NICInitializeMedia;

		// Since NWAY usually takes several seconds and this is too much time
		// to wait for during UM_Initialize we will check for completion later
		// through "LinkCheck". Initialize PerformingNWAY status vars then.

		if (!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED))
		{
			MP_SET_FLAG(Adapter, fMP_DISCONNECTED);
			Status = NICIndicateMediaStatus(Adapter);
			if (Status != NIC_STATUS_SUCCESS)
				goto Exit_NICInitializeMedia;
		}

		Adapter->PerformingNWAY = TRUE;
		Adapter->PerformingNWAYTimerCounter = 0;
		SMSC_TRACE0(DBG_INIT,"Starting NWAY ...\r\n");

	}
	else // if(pDevice->MediaType <= MEDIA_TYPE_AUTO_CUSTOM)
	{
		/* Determine media type we will now force to */
		switch(Adapter->MediaType)
		{
			case MEDIA_TYPE_10HD_FORCED:
				SMSC_TRACE0(DBG_INIT,"Forcing 10Mbs UTP\r\n");
				Adapter->LinkSpeed = LINK_SPEED_10MBS;
				Adapter->LinkDuplex = LINK_DUPLEX_HALF;
				break;

			case MEDIA_TYPE_10FD_FORCED:
				SMSC_TRACE0(DBG_INIT,"Forcing 10Mbs UTP Full Duplex\r\n");
				Adapter->LinkSpeed = LINK_SPEED_10MBS;
				Adapter->LinkDuplex = LINK_DUPLEX_FULL;
				break;

			case MEDIA_TYPE_100HD_FORCED:
				SMSC_TRACE0(DBG_INIT,"Forcing 100Mbs UTP\r\n");
				Adapter->LinkSpeed = LINK_SPEED_100MBS;
				Adapter->LinkDuplex = LINK_DUPLEX_HALF;
				break;

			case MEDIA_TYPE_100FD_FORCED:
				SMSC_TRACE0(DBG_INIT,"Forcing 100Mbs UTP Full Duplex\r\n");
				Adapter->LinkSpeed = LINK_SPEED_100MBS;
				Adapter->LinkDuplex = LINK_DUPLEX_FULL;
				break;

			case MEDIA_TYPE_1000FD_MASTER_FORCED:
                		SMSC_TRACE0(DBG_INIT,"Forcing 1000Mbs Master Full Duplex\r\n");
                		Adapter->LinkSpeed = LINK_SPEED_1000MBS_MASTER;
                		Adapter->LinkDuplex = LINK_DUPLEX_FULL;
                		break;

            		case MEDIA_TYPE_1000FD_SLAVE_FORCED:
                		SMSC_TRACE0(DBG_INIT,"Forcing 1000Mbs Slave Full Duplex\r\n");
                		Adapter->LinkSpeed = LINK_SPEED_1000MBS_SLAVE;
                		Adapter->LinkDuplex = LINK_DUPLEX_FULL;
                		break;

			default:
				/* If we want to add suport for a non-MII PHY (for BNC ?) and	*/
				/* autonegotiate between it and the MII PHY based on whichever	*/
				/* criteria (power up link presence or not on MII?) this is the */
				/* place to do it. For now hardcode it to UTP					*/

				break;
		} /* switch */

		Status = NICSetMediaType(Adapter);
		if(Status != NIC_STATUS_SUCCESS)
			goto Exit_NICInitializeMedia;

		// -- fixme CL PHY LED_MODE workaround
        Status = NICReadPhy(Adapter, PHY_EXT_CTL_STS, &Value32);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICInitializeMedia;
        Status = NICWritePhy(Adapter, PHY_EXT_CTL_STS, (Value32 | PHY_EXT_CTL_STS_LED_MODE_));
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICInitializeMedia;
        


		// CL PHY LED_MODE workaround
        // CL PHY Revision 0xB
        if ((Adapter->PHY_OUI == 0xC0001CUL) &&
            (Adapter->PHY_Model == 0xEUL) &&
            (Adapter->PHY_Revision == 0xBUL))
        {
            SMSC_TRACE0(DBG_INIT,"Applying CL PHY LED_MODE workaround in CL PHY Revision 0xB\r\n");
            Status = NICReadPhy(Adapter, PHY_EXT_CTL_STS, &Value32);
            if (Status != NIC_STATUS_SUCCESS)
                goto Exit_NICInitializeMedia;
            Status = NICWritePhy(Adapter, PHY_EXT_CTL_STS, (Value32 | PHY_EXT_CTL_STS_LED_MODE_));
            if (Status != NIC_STATUS_SUCCESS)
                goto Exit_NICInitializeMedia;
        }
		
		// Signal Media connection status change if necessary. Read twice
		//	in case link down was latched.
		Status = NICReadPhy(Adapter, PHY_STATUS, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICInitializeMedia;

		Status = NICReadPhy(Adapter, PHY_STATUS, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICInitializeMedia;
		else
			SMSC_TRACE1(DBG_INIT,"PHY_STATUS,Value32 = 0x%x \r\n",Value32);

		if( MP_TEST_FLAG(Adapter, fMP_DISCONNECTED) && (Value32 & MII_SR_LINK_STATUS_) )
		{
			MP_CLEAR_FLAG(Adapter, fMP_DISCONNECTED);
			Status = NICIndicateMediaStatus(Adapter);
			if (Status != NIC_STATUS_SUCCESS)
				goto Exit_NICInitializeMedia;
		}

	}


	
	Status = NIC_STATUS_SUCCESS;

Exit_NICInitializeMedia:
	SMSC_TRACE1(DBG_INIT,"<-- NICInitializeMedia,Status = %d \r\n",Status);
	return Status;
} /* NICInitializeMedia*/


NIC_STATUS 
NICSetMediaType(
    IN PMP_ADAPTER Adapter    
    )
/*++

Routine Description:

    This routine initializes the PHY registers for the proper media type.

Arguments:

    Adapter     Pointer to our adapter

Return Value:

    NIC_STATUS.

--*/
{
    NIC_STATUS Status;
    DWORD Value32, Temp32;
    int Count = 0;

    SMSC_TRACE0(DBG_INIT,"--->NICSetMediaType()\r\n"); 
    if(Adapter->MediaType >= MEDIA_TYPE_10HD_FORCED)
    {
        /* Force link speed and duplex mode */
        SMSC_TRACE0(DBG_INIT,"Forcing media type: ");

        Value32 = 0;
        if((Adapter->LinkSpeed == LINK_SPEED_1000MBS_MASTER) ||
           (Adapter->LinkSpeed == LINK_SPEED_1000MBS_SLAVE))
        {
            Value32 |= (MII_CR_SPEED_SELECT_MSB_ + MII_CR_FULL_DUPLEX_);
            SMSC_TRACE0(DBG_INIT,"1000 Mbps, FD, ");
            
            Status = NICReadPhy(Adapter, PHY_1000T_CTRL, &Temp32);
            if (Status != NIC_STATUS_SUCCESS)
                goto Exit_NICSetMediaType;
            
            /* Force Master/Slave Manual Config */
            Temp32 |= CR_1000T_MS_ENABLE_;
            
            /* Config. PHY as Master/Slave */
            if(Adapter->LinkSpeed == LINK_SPEED_1000MBS_MASTER)
            {
                Temp32 |= CR_1000T_MS_VALUE_;
                SMSC_TRACE0(DBG_INIT,"PHY Master");
            }
            else
            {
                Temp32 &= (~CR_1000T_MS_VALUE_);
                SMSC_TRACE0(DBG_INIT,"PHY Slave");
            }

            Status = NICWritePhy(Adapter, PHY_1000T_CTRL, Temp32);
            Adapter->PhyRxClkWorkaround = TRUE;
            
            if (Status != NIC_STATUS_SUCCESS)
                goto Exit_NICSetMediaType;
        }
        else
        {
            if(Adapter->LinkDuplex == LINK_DUPLEX_FULL) 
            {
                Value32 |= MII_CR_FULL_DUPLEX_;
                SMSC_TRACE0(DBG_INIT,"Full Duplex, ");
            }
            else
                SMSC_TRACE0(DBG_INIT,"Half Duplex, ");

            if(Adapter->LinkSpeed == LINK_SPEED_100MBS)
            {
                Value32 |= MII_CR_SPEED_SELECT_LSB_;
                SMSC_TRACE0(DBG_INIT,"100Mbs\r\n");
            }
            else
                SMSC_TRACE0(DBG_INIT,"10Mbs\r\n");
        }
        
        // Setup Flow Control options
        switch(Adapter->FlowControl)
        {
            case DISABLED:
                Adapter->bGeneratePause = FALSE;
                Adapter->bRespondToPause = FALSE;

            case FC_SYMM:
                Adapter->bRespondToPause = TRUE;
                Adapter->bGeneratePause = TRUE;
                break;

            case FC_ASYM_GEN:
                Adapter->bGeneratePause = TRUE;
                Adapter->bRespondToPause = FALSE;
                break;

            case FC_ASYM_RESP:
                Adapter->bGeneratePause = FALSE;
                Adapter->bRespondToPause = TRUE;
                break;
        }

        // Force the link partner to restart autonegotiation by dropping
        // the link (put the PHY into loopback mode)

        Status = NICWritePhy(Adapter, PHY_CTRL, Value32 | MII_CR_LOOPBACK_);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICSetMediaType;

        NdisMSleep(NDIS_SLEEP_1_MS);

        Status = NICWritePhy(Adapter, PHY_CTRL, Value32);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICSetMediaType;

        // Wait for the link to be down up to 5 secs
        while (Count++ < NIC_5_SEC_TIMEOUT)
        {
            NdisMSleep(NDIS_SLEEP_1_MS); // Wait 1ms

            Status = NICReadPhy(Adapter, PHY_STATUS, &Value32);
            if (Status != NIC_STATUS_SUCCESS)
                goto Exit_NICSetMediaType;

            if((Value32 == 0xffff) || (Value32 == 0x0000))
                continue;

            if(Value32 & MII_SR_LINK_STATUS_)
            {
                break;
            }
        }

        if (Count == NIC_5_SEC_TIMEOUT)
        {
            goto Exit_NICSetMediaType;
        }

        // If the media type wasn't set to disconnected do so and and indicate 
        // disconnection.
        if (!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED))
        {
            MP_SET_FLAG(Adapter, fMP_DISCONNECTED);
            Status = NICIndicateMediaStatus(Adapter);
            if(Status != NIC_STATUS_SUCCESS)
                goto Exit_NICSetMediaType;
        }
    }

    
    Status = NICReadRegister(Adapter, ETH_MAC_CR, &Value32);
    if (Status != NIC_STATUS_SUCCESS)
        goto Exit_NICSetMediaType;
    SMSC_TRACE2(DBG_INIT,"%s. ETH_MAC_CR = 0x%08lx\r\n", (Adapter->LinkDuplex == LINK_DUPLEX_FULL) ? "FULL" : "HALF", Value32 );
    

    // Setup Flow Control options
    if(Adapter->FlowControl != DISABLED)
    {
        if (Adapter->bGeneratePause) 
        {
            // Handle the confusing case of advertising FC_ASYM_RESP
            // which is "Asymmetric towards local device AND symmetric" 
            // so it allows to generate pause frames as per table 28B-2 
            // and 28B-3. Since we did not setup FCTFlowReg & FCPause for 
            // FC_ASYM_RESP during init we need to do it now.
            if (Adapter->FCTFlowReg == 0)
            {
                DWORD dwWaterMark;
                
                dwWaterMark = (((Adapter->RxFifoSize * 1)/(10 * 512)) & 0x7FUL);  // Set Hi Water Mark to 20% of the Rx Fifo Size (Used Space)
                dwWaterMark <<= 8UL;
                dwWaterMark |= (((Adapter->RxFifoSize * 6)/(10 * 512)) & 0x7FUL); // Set Lo Water Mark to 80% of the Rx Fifo Size (Used Space)
                Adapter->FCTFlowReg = dwWaterMark;
            }

            if (Adapter->FCPause == 0)
            {
                Adapter->FCPause = 0xFFFF;  // Use the max, since we'll send a 0 to overide it when we don't 
                                            // need it anymore
            }

            // Enable device to send pauses
            SMSC_TRACE2(DBG_INIT,"Enabling sending Flow Control frames with Pause = 0x%04X and AFC_REG = 0x%08lx\r\n", 
                            Adapter->FCPause, Adapter->FCTFlowReg);
        }
        else
        {
            SMSC_TRACE0(DBG_INIT,"Disabling sending Flow Control pause frames\r\n"); 
            Adapter->FCTFlowReg = 0;
        }

        if (Adapter->bRespondToPause)
        {
            // Enable device to respond to pauses
            SMSC_TRACE0(DBG_INIT,"Enabling responding to Flow Control pause frames\r\n");
        }
        else
        {
            // Disable device to respond to pauses
            SMSC_TRACE0(DBG_INIT,"Disabling responding to Flow Control pause frames\r\n");
        }

        /* Set ETH_FLOW register  */
        Status = NICWriteRegister(Adapter, ETH_FCT_FLOW, &(Adapter->FCTFlowReg));
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICSetMediaType;

        Value32 = Adapter->FCPause;
        Value32 |= (Adapter->bRespondToPause ? ETH_FLOW_CR_RX_FCEN_ : 0);
        Value32 |= (Adapter->bGeneratePause ? ETH_FLOW_CR_TX_FCEN_ : 0);
        Status = NICWriteRegister(Adapter, ETH_FLOW, &Value32 );
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICSetMediaType;

        
        Status = NICReadRegister(Adapter, ETH_FLOW, &Value32);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICSetMediaType;
        SMSC_TRACE1(DBG_INIT,"ETH_FLOW    = 0x%08lx\r\n", Value32);
        Status = NICReadRegister(Adapter, ETH_FCT_FLOW, &Value32);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICSetMediaType;
        SMSC_TRACE1(DBG_INIT,"ETH_FCT_FLOW = 0x%08lx\r\n", Value32);
        
    }

    Status = NIC_STATUS_SUCCESS;

Exit_NICSetMediaType:
    SMSC_TRACE0(DBG_INIT,"<--- NICSetMediaType() \r\n"); 
    return Status;
}


/*++

Routine Description:

	This routine checks to see if autonegotiation completed

Arguments:

	Adapter 	Pointer to our adapter

Return Value:

	TRUE or FALSE

--*/
BOOL NICIsNWAYComplete(
	IN PMP_ADAPTER Adapter
	)
{
	UINT Count;
	DWORD Value32 = 0x0000;
	NIC_STATUS Status;
	BOOLEAN RetVal=FALSE;

	SMSC_TRACE0(DBG_LINK,"--> NICIsNWAYComplete\r\n");

	for(Count=0;Count<4;Count++)
	{
		Status = NICReadPhy(Adapter, PHY_STATUS, &Value32);
		
		SMSC_TRACE1(DBG_LINK,"PHY_STATUS = 0x%x\r\n", Value32);
		
		if (Status != NIC_STATUS_SUCCESS)
		{			
			goto Exit_NICIsNWAYComplete;
			
		}

		if((Value32 == 0xffff) || (Value32 == 0x0000))
			continue;

		if(Value32 & MII_SR_AUTONEG_COMPLETE_)
		{
			RetVal = TRUE;
			break;
		}
	} 

Exit_NICIsNWAYComplete:
	
	if (RetVal)
	{
		SMSC_TRACE0(DBG_LINK,"NWAY Completed!\r\n");
	}
	else
	{
		SMSC_TRACE0(DBG_LINK,"NWAY Still In progress...\r\n");
	}
	

	SMSC_TRACE0(DBG_LINK,"<-- NICIsNWAYComplete\r\n");

	return RetVal;
}


/*++

Routine Description:

	This routine indicates completes pending media status and link queries

Arguments:

	Adapter 	Pointer to our adapter

Return Value:

	None.

--*/

void
NICCompleteOidQueryIfPending(
	IN PMP_ADAPTER Adapter
	)
{
	PWDM_NDIS_REQUEST	wdmNdisRequest = &Adapter->WdmNdisRequest;

	NdisAcquireSpinLock(&Adapter->LCSpinlock);

	// Let's complete the pending request first if it's a query

	if (
		MP_TEST_FLAG(Adapter, fMP_REQUEST_PENDING) &&
		(wdmNdisRequest->RequestType == NdisRequestQueryInformation)
	   )
	{
		NDIS_STATUS OidStatus = NDIS_STATUS_SUCCESS;
		*wdmNdisRequest->BytesReadOrWritten = sizeof(ULONG);

		switch (wdmNdisRequest->Oid)
		{


			case OID_GEN_MEDIA_CONNECT_STATUS:
				SMSC_TRACE0(DBG_REQUEST,"Completing Pending OID_GEN_MEDIA_CONNECT_STATUS\r\n");
				*(DWORD *)wdmNdisRequest->InformationBuffer = NICGetMediaConnectStatus(Adapter);
				break;

			case OID_GEN_LINK_SPEED:
				SMSC_TRACE0(DBG_REQUEST,"Completing Pending OID_GEN_LINK_SPEED\r\n");
				*(DWORD *)wdmNdisRequest->InformationBuffer = NICGetLinkSpeed(Adapter);
				break;
			default:
				// Since we did not pend anything other than media status requests,
				// this should never happen !!!
				ASSERT (FALSE);
				*wdmNdisRequest->BytesReadOrWritten = 0;
				OidStatus = NDIS_STATUS_FAILURE;
				break;
		}

		
		if (OidStatus == NDIS_STATUS_SUCCESS)
		{
			SMSC_TRACE2(DBG_REQUEST,"Completing Pending 0x%x successfully with value 0x%08lx\r\n",
							wdmNdisRequest->Oid,
							*(DWORD *)wdmNdisRequest->InformationBuffer);
		}
		else
		{
			SMSC_TRACE1(DBG_REQUEST,"Completing Pending 0x%x with NDIS_STATUS_FAILURE, we never pended this OID!\r\n",
							wdmNdisRequest->Oid);
		}
		

		MP_CLEAR_FLAG(Adapter, fMP_REQUEST_PENDING);
		NdisMQueryInformationComplete(Adapter->AdapterHandle, OidStatus);
	}
	
	NdisReleaseSpinLock(&Adapter->LCSpinlock);
}


/*++

Routine Description:

	Configures wake up events, which allows the computer to wake up on receipt
	of a magic packet / link change / pattern match

Arguments:

	Adapter 	Pointer to our adapter

Return Value:

	NIC_STATUS.

--*/

NIC_STATUS
NICConfigureWakeEvents(
		IN PMP_ADAPTER Adapter
	)
{
	NIC_STATUS Status = NIC_STATUS_SUCCESS;
	DWORD Value32,dwErr,dwTemp;
  
	SMSC_TRACE0(DBG_POWER, "--> NICConfigureWakeEvents\r\n");


	if ((Adapter->CurrentWakeEvents != 0UL) || (Adapter->WolMagicPacketSettings == FORCED) || (Adapter->WoLinkChgSettings == FORCED))
	{

		// If EDPD is enabled, get out of it first
		if (Adapter->PhyLinkPowerDown == TRUE)
		{
			SMSC_TRACE0(DBG_POWER,"Turning OFF EDPD.\r\n");

			Status = NICReadPhy(Adapter, PHY_MOD_CTL_STS, &Value32);
			if (Status != NIC_STATUS_SUCCESS)
				goto Exit_NICConfigureWakeEvents;
			SMSC_TRACE1(DBG_POWER,"PHY__MOD_CTL_STS=0x%08lx\r\n", Value32);

			Value32 &= ~PHY_MOD_CTL_STS_EDPWRDOWN_;
			Status = NICWritePhy(Adapter, PHY_MOD_CTL_STS, Value32 );
			if (Status != NIC_STATUS_SUCCESS)
				goto Exit_NICConfigureWakeEvents;

			Adapter->PhyLinkPowerDown = FALSE;

			if (!Adapter->PhyIsExternal)
            {
                // Restore Auto MDIX
                Status = NICWritePhy(Adapter, PHY_MOD_CTL_STS, Adapter->SaveAutoMdix);
                if(Status != NIC_STATUS_SUCCESS)
                    goto Exit_NICConfigureWakeEvents;

                Status = NICWritePhy(Adapter, PHY_EXT_CTL_STS, Adapter->SaveMdixState);
                if(Status != NIC_STATUS_SUCCESS)
                    goto Exit_NICConfigureWakeEvents;

            }

			// If we were doing NWAY turn it ON now
            if(Adapter->MediaType < MEDIA_TYPE_10HD_FORCED)
            {
                Status = NICWritePhy(  
                               Adapter, 
                               PHY_CTRL, 
                               MII_CR_AUTO_NEG_EN_ || MII_CR_RESTART_AUTO_NEG_// --fixme Don't we need to set MII_CR_RESTART_AUTO_NEG_ as well?
                               );
                if (Status != NIC_STATUS_SUCCESS)
                    	goto Exit_NICConfigureWakeEvents;

            }
		}
	
		SMSC_TRACE0(DBG_POWER,"Enabling Wake events in ESS Regs. \r\n");
		Status = NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICConfigureWakeEvents;
		Value32 &= (~ETH_PMT_CTL_RES_CLR_WKP_MASK_);
		SMSC_TRACE1(DBG_POWER,"Setting Wake Auto Clear to %d\r\n", Adapter->WakeAutoClear);
		Value32 = Value32 | (Adapter->WakeAutoClear << 8);
		Status = NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32 );

		// Disable Wake up packet generation, will reenable right below if needed
		Status = NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
		Value32 &= ~ETH_PMT_CTL_WOL_EN_;
		Status =NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICConfigureWakeEvents;
			

		if(
			(Adapter->CurrentWakeEvents & (NIC_FEATURE_WAKE_ON_LAN | NIC_FEATURE_MAGIC_PACKET)) ||
			(Adapter->WolMagicPacketSettings == FORCED)
		  )
		{
			// Enable or disable MP wakes
			if (
				(Adapter->CurrentWakeEvents & NIC_FEATURE_MAGIC_PACKET) ||
				(Adapter->WolMagicPacketSettings == FORCED)
			   )
			{

				
				if ((Adapter->CurrentWakeEvents & NIC_FEATURE_MAGIC_PACKET) != NIC_FEATURE_MAGIC_PACKET)
				{
					SMSC_TRACE0(DBG_POWER, "MAGIC PACKET. user forced thru registy key 'WolMPSettings'.\r\n");
				}
				

				SMSC_TRACE0(DBG_POWER,"Set magic Packet !!!\r\n");

				// Clear any pending magic packet status
				Status =NICReadRegister(Adapter, ETH_WUCSR, &Value32);
				Value32 |= ETH_WUCSR_MPR_;
				Status =NICWriteRegister(Adapter, ETH_WUCSR, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICConfigureWakeEvents;

				// Enable MP wake
				SMSC_TRACE0(DBG_POWER,"Now enabling WUCSR_MPR_!!!\r\n");
				Status =NICReadRegister(Adapter, ETH_WUCSR, &Value32);
				Value32 |= ETH_WUCSR_MPEN_;
				Status =NICWriteRegister(Adapter, ETH_WUCSR, &Value32);
				Status =NICReadRegister(Adapter, ETH_WUCSR, &dwTemp);
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICConfigureWakeEvents;
				SMSC_TRACE1(DBG_POWER, "ETH_WUCSR= 0x%x \r\n",dwTemp);
				
				Status =NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
				Value32 |= ETH_PMT_CTL_WOL_EN_;
				Status =NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
				Status =NICReadRegister(Adapter, ETH_PMT_CTL, &dwTemp);
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICConfigureWakeEvents;
				SMSC_TRACE1(DBG_POWER, "ETH_PMT_CTL= 0x%x \r\n",dwTemp);
			}
			else
			{
				Status =NICReadRegister(Adapter, ETH_WUCSR, &Value32);
				Value32 &= ~ETH_WUCSR_MPEN_;
				Status =NICWriteRegister(Adapter, ETH_WUCSR, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICConfigureWakeEvents;
			}

			// Enable or disable Wake on pattern match wakes
			if (Adapter->CurrentWakeEvents & NIC_FEATURE_WAKE_ON_LAN)
			{
				DWORD Temp32;
				USHORT i=0;
                		// Program WOL Configs
                		for (i=0;i<Adapter->NumOfWolPatterns;i++)
                		{
                			Temp32=Adapter->dwWUFFCfg[i];
                    			Status = NICWriteRegister(Adapter, (ETH_WUF_CFGx + i*4), &Temp32);
                    			if (Status != NIC_STATUS_SUCCESS)
                        			goto Exit_NICConfigureWakeEvents;
                		}

                		// Program WOL Filters
                		for (i=0;i<(Adapter->NumOfWolPatterns*4);i++)
               		{
               			Temp32=Adapter->dwWUFFMask[i];
                    			Status = NICWriteRegister(Adapter, (ETH_WUF_MASKx + i*4), &Temp32);
                    			if (Status != NIC_STATUS_SUCCESS)
                        			goto Exit_NICConfigureWakeEvents;
                		}

				
				// Clear any pending pattern match packet status
				Status =NICReadRegister(Adapter, ETH_WUCSR, &Value32);
				Value32 |= ETH_WUCSR_WUFR_;
				Status =NICWriteRegister(Adapter, ETH_WUCSR, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;

				// Enable pattern match packet wake
				SMSC_TRACE0(DBG_POWER, "Now enabling WUCSR_WAKE_EN_\r\n");
				Status =NICReadRegister(Adapter, ETH_WUCSR, &Value32);
				Value32 |= ETH_WUCSR_WAKE_EN_;
				Status =NICWriteRegister(Adapter, ETH_WUCSR, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;
				
				Status =NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
				Value32 |= ETH_PMT_CTL_WOL_EN_;
				Status =NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;
			}
			else
			{
				Status =NICReadRegister(Adapter, ETH_WUCSR, &Value32);
				Value32 &= ~ETH_WUCSR_WAKE_EN_;
				Status =NICWriteRegister(Adapter, ETH_WUCSR, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;
			}

			SMSC_TRACE1(DBG_POWER, "CurrentWakeEvents = 0x%08lx. Reenabling Receiver!\r\n", Adapter->CurrentWakeEvents);
			Status =NICReadRegister(Adapter, ETH_MAC_RX, &Value32);
			Value32 |= ETH_MAC_RX_RXEN_;
			Status =NICWriteRegister(Adapter, ETH_MAC_RX, &Value32);
			Status =NICReadRegister(Adapter, ETH_MAC_RX, &dwTemp);
			if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;
			SMSC_TRACE1(DBG_POWER, "MAC_CR= 0x%x \r\n",dwTemp);

			// Put Thera in Suspend0
			SMSC_TRACE0(DBG_POWER, "Setting Suspend0 mode\r\n");
			Status =NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
			Value32 &= (~(ETH_PMT_CTL_SUS_MODE_ | ETH_PMT_CTL_WUPS_MASK_| ETH_PMT_CTL_PHY_RST_));
			Value32 |= ETH_PMT_CTL_SUS_MODE_0_;
			Status =NICWriteRegister(Adapter, ETH_PMT_CTL,	&Value32);
			Status =NICReadRegister(Adapter, ETH_PMT_CTL, &dwTemp);
			if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;
			SMSC_TRACE1(DBG_POWER, "ETH_PMT_CTL= 0x%x \r\n",dwTemp);


			dwErr = ClearOrSetFeature(
				Adapter->UsbFuncs,
				Adapter->hUsbDevice,
				NULL,
				NULL,
				USB_SEND_TO_DEVICE,
				USB_FEATURE_REMOTE_WAKEUP,
				 0,
				 4000,
				 TRUE
					);
	
			if (dwErr != ERROR_SUCCESS) {
				SMSC_TRACE1(DBG_POWER,"setfeature Failed, dwErr= %d \r\n",dwErr);
			}
			else {
				Adapter->SetFeatureRemoteWake=TRUE;
			}
			
          }
			
				
			

		// Enable Energy detect (Link up) to generate wake
		if ((Adapter->CurrentWakeEvents & NIC_FEATURE_LINK_CHANGE) || (Adapter->WoLinkChgSettings == FORCED))
		{
			
			if ((Adapter->CurrentWakeEvents & NIC_FEATURE_LINK_CHANGE) != NIC_FEATURE_LINK_CHANGE)
			{
				SMSC_TRACE0(DBG_POWER, "LINK CHG. Note Windows did NOT enable it thru Ndis, user forced thru registy key 'WoLinkChgSettings'.\r\n");
			}
			

			// If there's currently no link we can use Suspend1
			Status =NICReadPhy(Adapter, PHY_STATUS, &Value32);
			if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;
			
			if (!(Value32 & MII_SR_LINK_STATUS_))
			{
				SMSC_TRACE0(DBG_POWER, "Setting PHY in PD/ED Mode.\r\n");

				Status = NICReadRegister(Adapter, ETH_MAC_RX, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICConfigureWakeEvents;

				SMSC_TRACE0(DBG_POWER, "Stop receiver\r\n");
				Value32 &= ~ETH_MAC_RX_RXEN_;
				Status = NICWriteRegister(Adapter, ETH_MAC_RX, &Value32 );
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICConfigureWakeEvents;

				// If it was enabled, disable Wake up packet generation
				Status = NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICConfigureWakeEvents;
				
				Value32 &=	~ETH_PMT_CTL_WOL_EN_;
				Status = NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICConfigureWakeEvents;

				// If we were doing NWAY turn it OFF now
				if(Adapter->MediaType < MEDIA_TYPE_10HD_FORCED)
				{
					Value32=0;
					Status = NICWritePhy(Adapter, PHY_CTRL, Value32);
					if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;

					SMSC_TRACE0(DBG_POWER,"NWAY Disabled\r\n");
				}
				
				if (!Adapter->PhyIsExternal)
				{
					// Turn off Auto MDIX
					Status = NICReadPhy(Adapter, PHY_MOD_CTL_STS, &Value32);
					if(Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;
					
					Value32 &= 0x1FFF;		
					Status = NICWritePhy(Adapter, PHY_MOD_CTL_STS,Value32 );
					if(Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;

					SMSC_TRACE0(DBG_POWER,"Turn off AMDIX\r\n");
				}

				 // Clear any pending Phy interrupt and enable the mask for it
				Status =NICSetPHYInterruptCapability(Adapter, PHY_INT_MASK_ENERGY_ON_);
				if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;
					
				if (!Adapter->PhyIsExternal)
				{
					// Enable ED (really "Phy interrupt") Wake at Mac
					Status = NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
					if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;
					
					Value32 |= ETH_PMT_CTL_ED_EN_;
					Status = NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32 );
					if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;

							   
					// Set phy in ED low power mode.
					Status =NICReadPhy(Adapter, PHY_MOD_CTL_STS, &Value32);
					Value32 |= PHY_MOD_CTL_STS_EDPWRDOWN_;
					Status =NICWritePhy(Adapter, PHY_MOD_CTL_STS, Value32);
					if (Status != NIC_STATUS_SUCCESS)
						 goto Exit_NICConfigureWakeEvents;
				}

				Adapter->PhyLinkPowerDown =TRUE;
 
				// Put Thera in Suspend1
				SMSC_TRACE0(DBG_POWER, "Setting Suspend1 mode\r\n");
				Status =NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
				Value32 &= (~(ETH_PMT_CTL_SUS_MODE_ | ETH_PMT_CTL_WUPS_MASK_ | ETH_PMT_CTL_PHY_RST_));
				Value32 |= ETH_PMT_CTL_SUS_MODE_1_;
				Status =NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;
			}
			else
			{

			 		// Clear any pending Phy interrupt and enable the mask for it
					Status =NICSetPHYInterruptCapability(Adapter, PHY_INT_MASK_ENERGY_ON_);
					if (Status != NIC_STATUS_SUCCESS)
							goto Exit_NICConfigureWakeEvents;
					
					if (!Adapter->PhyIsExternal) {
						Status =NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
						Value32 |= ETH_PMT_CTL_ED_EN_;
						Status =NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
						if (Status != NIC_STATUS_SUCCESS)
								goto Exit_NICConfigureWakeEvents;
					}
			
				 	// Put Thera in Suspend0
					SMSC_TRACE0(DBG_POWER, "Setting Suspend0 mode\r\n");
					Status =NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
					Value32 &= (~(ETH_PMT_CTL_SUS_MODE_ | ETH_PMT_CTL_WUPS_MASK_ | ETH_PMT_CTL_PHY_RST_));
					Value32 |= ETH_PMT_CTL_SUS_MODE_0_;
					Status =NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
					if (Status != NIC_STATUS_SUCCESS)
							goto Exit_NICConfigureWakeEvents;
			}

			dwErr = ClearOrSetFeature(
					 Adapter->UsbFuncs,
					Adapter->hUsbDevice,
					NULL,
					NULL,
					USB_SEND_TO_DEVICE,
					USB_FEATURE_REMOTE_WAKEUP,
					0,
					0,
					TRUE
					);
	
			if (dwErr != ERROR_SUCCESS) {
				SMSC_TRACE1(DBG_POWER,"setfeature Failed, dwErr= %d \r\n",dwErr);
			}
			else {
				Adapter->SetFeatureRemoteWake=TRUE;
			}
			
		}
		else
		{
			// Clear any pending Phy interrupt and disable the mask for it
			Status =NICSetPHYInterruptCapability(Adapter, 0);
			if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;

			Status =NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
			Value32 &= ~ETH_PMT_CTL_ED_EN_;
			Status =NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
			if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;
		}
	}
	else
	{
		SMSC_TRACE0(DBG_POWER,"Disabling Wake events in ESS Regs.\r\n ");

		// Disable Energy detect (Link up) & Wake up events to do USB wake
		Status =NICReadRegister(Adapter, ETH_WUCSR, &Value32);
		Value32 &= ~(ETH_WUCSR_MPEN_ | ETH_WUCSR_WAKE_EN_ );
		Status =NICWriteRegister(Adapter, ETH_WUCSR, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;

		Status =NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
		Value32 &= ~(ETH_PMT_CTL_ED_EN_ | ETH_PMT_CTL_WOL_EN_);
		Status =NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;

		// Put Thera in Suspend2
		SMSC_TRACE0(DBG_POWER, "Setting Suspend2 mode\r\n");
		Status =NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
		Value32 &= (~(ETH_PMT_CTL_SUS_MODE_ | ETH_PMT_CTL_WUPS_MASK_ | ETH_PMT_CTL_PHY_RST_));
		Value32 |= ETH_PMT_CTL_SUS_MODE_2_;
		Status =NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICConfigureWakeEvents;
	}

Exit_NICConfigureWakeEvents:

	SMSC_TRACE0(DBG_POWER, "<-- NICConfigureWakeEvents\r\n");

	return Status;
}

/*++

Routine Description:

   Workitem to process the USB Error Recovery. This workitme is scheduled
   in completion functions of send, receive and interrupt after a failed
   URB

Arguments:

	WorkItem - Pointer to workitem

	Context - Pointer to our Adapter

Return Value:

	None.

--*/
VOID
NICUsbErrorRecoveryWorkItemCallBack(
	PNDIS_WORK_ITEM WorkItem,
	PVOID Context
	)

{
	PMP_ADAPTER 	Adapter = (PMP_ADAPTER)Context;
	NIC_STATUS		nStatus;
	DWORD Value32;
	
	SMSC_TRACE0(USB_RECOVERY, "--> NICUsbErrorRecoveryWorkItemCallBack\n");

	MP_SET_FLAG(Adapter, fMP_USB_RESET_IN_PROGRESS);
	
	// If there is a reset or halt or surprise remove in progress just get out
	if (MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS || fMP_HALT_IN_PROGRESS) || \
		MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED))
	{
		goto Exit_NICUsbErrorRecoveryWorkItemCallBack;
	}

	// If there was a TXE, no need to do anything here since we will need a reset 
	// anyway to recover, just set the flag.
	if (NICReadRegister(Adapter, ETH_INT_STS, &Value32) != NIC_STATUS_SUCCESS)
		goto Exit_NICUsbErrorRecoveryWorkItemCallBack;

	if (Value32 & ETH_INT_STS_TXE_)
	{
		SMSC_TRACE0(DBG_ERROR, "ETH_INT_STS_TXE_ found set in USB Error Recovery: TX Error. Will request reset in next check for hang !!!\r\n");
		MP_SET_FLAG(Adapter, fMP_NEED_FULL_RESET);
		goto Exit_NICUsbErrorRecoveryWorkItemCallBack;
	}

	NdisAcquireSpinLock(&Adapter->RequestLock);
	if (MP_TEST_FLAG(Adapter, fMP_REQUEST_PENDING))
	{
		// We cannot do our reset here to avoid a race condition
		// with the requestworkitem, just set a flag for the request
		// workitem to schedule our workitem back when it's done.
		MP_SET_FLAG(Adapter, fMP_NEED_USB_RESET);

		//
		// Clear the flag to let the WorkItem structure be reused for
		// scheduling another one. Lets do this first to avoid raise
		// conditions
		//
		InterlockedExchange(&Adapter->IsUsbErrorRecoveryWorkItemQueued, FALSE);

		MP_CLEAR_FLAG(Adapter, fMP_USB_RESET_IN_PROGRESS);

		NdisReleaseSpinLock(&Adapter->RequestLock);

		MP_DEC_REF(Adapter);

		SMSC_TRACE0(USB_RECOVERY, "<-- NICUsbErrorRecoveryWorkItemCallBack\r\n");
		return;
	}
	else
	{
		NdisReleaseSpinLock(&Adapter->RequestLock);
	}

	NICStopAndFlushRxPath(Adapter); 	
	NICStopAndFlushTxPath(Adapter); 
	NICFreeQueuedSendPackets(Adapter);
	#ifdef USE_INT_PIPE
	if (Adapter->StopBulkInOnRxEmpty) {
		NICStopInterruptPipe(Adapter);		
	}
	#endif

	nStatus = NICStartTxPath(Adapter);
					
	nStatus |= NICStartRxPath(Adapter);
					
	#ifdef USE_INT_PIPE
	if (Adapter->StopBulkInOnRxEmpty) 
	{
		{
			nStatus |= NICStartInterruptPipe(Adapter);
							
		}
	}
	#endif
	
Exit_NICUsbErrorRecoveryWorkItemCallBack:
	// OK, we're done

	//
	// Clear the flag to let the WorkItem structure be reused for
	// scheduling another one. Lets do this first to avoid raise
	// conditions
	//
	InterlockedExchange(&Adapter->IsUsbErrorRecoveryWorkItemQueued, FALSE);

	MP_CLEAR_FLAG(Adapter, fMP_USB_RESET_IN_PROGRESS);

	MP_DEC_REF(Adapter);

	SMSC_TRACE0(USB_RECOVERY,"<-- NICUsbErrorRecoveryWorkItemCallBack\n");
	
}


/*++

Routine Description:

   Function  to process the reset request. This  is called in the MPReset function.

Arguments:

	Adapter - Pointer to our Adapter

Return Value:

	None.

--*/
NIC_STATUS
NICReset(
	PMP_ADAPTER Adapter
	)

{
	int Count = 0;
	NIC_STATUS nStatus = NIC_STATUS_SUCCESS;

	SMSC_TRACE0(DBG_RESET,"--> NICReset\r\n");
	
	if ((MP_TEST_FLAG(Adapter,fMP_DEVICE_NOT_ATTACHED)))
	{
		SMSC_TRACE0(DBG_ERROR,"Device Not Attached yet!!! \r\n");
		nStatus = NIC_STATUS_FAILURE;
		goto Exit_NICReset;
	}

	if (MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED))
	{
		SMSC_TRACE0(DBG_ERROR,"Device already removed!!! \r\n");
		nStatus = NIC_STATUS_FAILURE;
		goto Exit_NICReset;
	}
	
	// First cancel the Link Check Timer / Wait for its in-progress workitem to finish
	SafeCancelLinkCheck(Adapter);

//    We don't need to complete those pending OID here since Ndis will send the request again after the MPReset finish. 
//	if(MP_TEST_FLAG(Adapter, fMP_REQUEST_PENDING))
//		NICCompleteOidQueryIfPending(Adapter);

	// If there was another non query request pending, wait for it to complete
	// Before we start the reset activity
	while (MP_TEST_FLAG(Adapter, fMP_REQUEST_PENDING || fMP_USB_RESET_IN_PROGRESS)) 
	{
		NdisMSleep(NDIS_SLEEP_5_MS);	// Wait 5 msec and try again.
		if (++Count == 1000UL)
		{
			// If we spent 5 secs or more then something failed.
			SMSC_TRACE0(DBG_ERROR,"Timeout waiting for OID request completion during reset workitem\r\n");
			nStatus = NIC_STATUS_FAILURE;
			goto Exit_NICReset;
		}
	}


	#ifdef USE_INT_PIPE
	if (Adapter->StopBulkInOnRxEmpty) {
		NICStopInterruptPipe(Adapter);
	}
	#endif	
	NICStopRxPath(Adapter);
	NICStopTxPath(Adapter);
	NICFreeQueuedSendPackets(Adapter);

	// Wait until our reset workitem and the initialize one are the only refs left
	Count = 0;
	while (Adapter->RefCount > 2)
	{
		NdisMSleep(NDIS_SLEEP_1_MS);
		if (++Count == NIC_1_SEC_TIMEOUT)
		{
			// If we spent 1 secs or more then something failed.
			SMSC_TRACE0(DBG_ERROR,"Timeout waiting for all completions during reset workitem\r\n");
			nStatus = NIC_STATUS_FAILURE;
			break;
		}
	}

	MP_SET_FLAG(Adapter, fMP_DISCONNECTED);
	// Now reset the hardware.
	if( (nStatus = NICResetAdapter(Adapter)) != NIC_STATUS_SUCCESS)
	{
		goto Exit_NICReset;
	}

	MP_CLEAR_FLAG(Adapter, fMP_RESET_IN_PROGRESS);

	if (NIC_STATUS_FAILURE == (nStatus = NICStartTxPath(Adapter)))
	{
		goto Exit_NICReset;
	}

	if (NIC_STATUS_FAILURE == (nStatus = NICStartRxPath(Adapter)))
	{
		goto Exit_NICReset;
	}

	#ifdef USE_INT_PIPE
	if (Adapter->StopBulkInOnRxEmpty) {
		if (NIC_STATUS_FAILURE == (nStatus = NICStartInterruptPipe(Adapter)))
		{
			goto Exit_NICReset;
		}
	}
	#endif
	// Even though our hardware would be ready at this time if we are not doing NWAY,
	// the device on the other side may be performing NWAY. So we should actually
	// wait a prudent amount of time for real link status before completing the reset.
	// We reset the link check status to call us back periodically until link up or
	// timeout at which time we asynchronously complete the reset.

	Adapter->SecondsElapsedAfterMPResetExit = 0;

	// Start the periodic Link check.
	StartLinkCheck(Adapter);

Exit_NICReset:
	// Clear the flag again in case we failed before the clear above.
	MP_CLEAR_FLAG(Adapter, fMP_RESET_IN_PROGRESS); 
/*
	NdisAcquireSpinLock(&Adapter->RequestLock);
	if (nStatus != NIC_STATUS_SUCCESS)
	{


		NdisMResetComplete(
							Adapter->AdapterHandle,
							NDIS_STATUS_HARD_ERRORS,
							FALSE
						  );
	}
	NdisReleaseSpinLock(&Adapter->RequestLock);
*/
	SMSC_TRACE0(DBG_RESET,"<-- NICReset\r\n");
	return nStatus;
}



/*++

Routine Description:

   Reset event

Arguments:

	hEvent - Event needs to be reset


Return Value:

	None.

--*/
DWORD
USB_ResetEvent(
	HANDLE hEvent
	)
{

	DWORD dwErr = ERROR_SUCCESS;
   
	SMSC_TRACE0(DBG_USB_RESET,"--> USB_ResetEvent\r\n");
	if ( hEvent && !ResetEvent(hEvent) ) {
		dwErr = GetLastError();
		SMSC_TRACE1(DBG_ERROR,"*** ResetEvent ERROR:%d ***\r\n", dwErr);
		return dwErr;
	}
	SMSC_TRACE0(DBG_USB_RESET,"<-- USB_ResetEvent\r\n");
	return dwErr;
}



/*++

Routine Description:

   Reset all pipes

Arguments:

	pUsbDevice -Pointer to the adapter

Return Value:

	TRUE if success
	FALSE if any errors

--*/

BOOL
USB_ResetAllPipes(
	PMP_ADAPTER pUsbDevice
	)
{
   BOOL bRc = FALSE;
   DWORD dwErr;

   SMSC_TRACE0(DBG_USB_RESET,"---->USB_ResetAllPipes\r\n");

   // 0. reset endpoint 0
   dwErr = ResetDefaultEndpoint(pUsbDevice->UsbFuncs,
								pUsbDevice->hUsbDevice);

   if ( ERROR_SUCCESS == dwErr )
   {
		// 1. reset BulkIn endpoint
		if ( NICResetPipe(pUsbDevice,pUsbDevice->BulkIn) )
		{
			// 2. reset BulkOut endpoint
			if ( NICResetPipe(pUsbDevice, pUsbDevice->BulkOut) )
			{
				// 3. reset Interrupt endpoint
				bRc = pUsbDevice->Interrupt.hPipe ? NICResetPipe(pUsbDevice,pUsbDevice->Interrupt) : TRUE;
			}
		}
	}

	SMSC_TRACE1(DBG_USB_RESET,"<----USB_ResetAllPipes:%d\r\n", bRc);

	return bRc;
}


/*++
// Description:
//		This routine calculates a 32 bit CRC on a sequence of bytes
//		and returns the 6 lower bits suitable for indexing the multicast
//		table
//
// Parameters:
//		BYTE *pBuffer	points to an array of bytes to calculate the CRC on.
//		WORD wLength	the size (in bytes) of the array
//
// Returns:
//		6 hash bits from the 32 bit CRC which are used to index the mcast regs
--*/
DWORD
NICCalculateHash(
	BYTE *pBuffer,
	WORD wLength
	)

{
	UINT i;
	DWORD crc = 0xFFFFFFFFUL;
	DWORD result = 0UL;
	const DWORD poly = 0xEDB88320UL;

	for(i=0U; i<wLength; i++)
	{
		int bit;
		DWORD data=((DWORD)pBuffer[i]);
		for(bit=0; bit<8; bit++)
		{
			const DWORD p = (crc^((DWORD)data))&1UL;
			crc >>= 1;
			if(p != 0UL) {
				crc ^= poly;
			}
			data >>=1;
		}
	}
	result=((crc&0x001UL)<<8)|
           ((crc&0x002UL)<<6)|
           ((crc&0x004UL)<<4)|
           ((crc&0x008UL)<<2)|
           ((crc&0x010UL))|
           ((crc&0x020UL)>>2)|
           ((crc&0x040UL)>>4)|
           ((crc&0x080UL)>>6)|
           ((crc&0x100UL)>>8);

    result &= 0x1FFUL; // hash result has to be a number between 0 and 511

	return (result);
}

NIC_STATUS 
NICSetHwMulticastTable(
    PMP_ADAPTER Adapter,
    DWORD   NumOfAdresses,
    BYTE    *Addresses
    )
/*++

Routine Description:

    Sets the multicast hash table at the MAC
 
Arguments:

    Adapter     Pointer to our adapter
    NumOfAdresses   number of elements in the multicast table 
    Addresses       Pointer to the multicast table
    
    Should be called at IRQL = PASSIVE_LEVEL.
    
Return Value:

    VOID
    
--*/    
{
    UINT i,j;
    DWORD hash;
    NIC_STATUS Status = NIC_STATUS_SUCCESS;

    /* Reset the multicast table */
    {
        int hIndex;
        
        for (hIndex=0; hIndex<16; hIndex++)
        {
            Adapter->MulticastHashTable[hIndex]=0UL;
        }
    }

    /* Now parse the addresses and set the hash bits */
    SMSC_TRACE0(DBG_MULTICAST, "Multicast Table:");
    for (i=0;i<NumOfAdresses;i++)
    {
        for (j=0;j<ETH_LENGTH_OF_ADDRESS;j++)
        {
            SMSC_TRACE1(DBG_MULTICAST,"%02x ",*(Addresses + j));
        }

        hash = NICCalculateHash(Addresses, ETH_LENGTH_OF_ADDRESS);
        SMSC_TRACE1(DBG_MULTICAST,"hash index = %d\r\n", hash);

        Adapter->MulticastHashTable[(hash/32)] |= (1UL << (hash % 32));
        Addresses += ETH_LENGTH_OF_ADDRESS;
    }
    
    // Now set the actual hardware
    {
        int hIndex;
        
        for (hIndex=0; hIndex<16; hIndex++)
        {
            SMSC_TRACE2(DBG_MULTICAST,"MulticastHashTable[%d] = 0x%08lx\n", hIndex, Adapter->MulticastHashTable[hIndex]);
            NICWriteHashRam32(Adapter, hIndex, Adapter->MulticastHashTable[hIndex]);
        }
    }

   return (Status);
}



/*++

Routine Description:

		Sets the PowerState at the MAC
 
Arguments:

		Adapter 	Pointer to our adapter
	PowerState		Power state to switch to.
	
	
	
Return Value:

	NIC_STATUS_XXX
	
--*/
NIC_STATUS 
NICSetPowerState(
	PMP_ADAPTER Adapter,
	BYTE	PowerState
	)
{
	NIC_STATUS		Status = NIC_STATUS_SUCCESS;
	DWORD			Value32, Value32_2,dwTemp;
	DWORD			OldPowerState = Adapter->CurrentPowerState;
	DWORD dwErr=0;
	BOOLEAN         bPhyInit = FALSE;

	

	SMSC_TRACE1(DBG_POWER, "--> NICSetPowerState (D%d)\r\n", PowerState);


	if ((Adapter->CurrentPowerState == 0) && (PowerState > 0))
	{
		SMSC_TRACE1(DBG_POWER, "We were in D0. Transitioning to power down state D%d.\r\n", PowerState);

		// Turn Off Link GPIO LED
//		NICSetGpo(Adapter, Adapter->LinkLedOnGpio, LED_OFF);
		
		// Stop the Link check Timer
		SafeCancelLinkCheck(Adapter);

		// Stop everything on USB, FEC and MAC
		#ifdef USE_INT_PIPE
		if (Adapter->StopBulkInOnRxEmpty) {
			Status = NICStopInterruptPipe(Adapter);
			
		}
		#endif
		Status |= NICStopAndFlushRxPath(Adapter);
		
		Status |= NICStopAndFlushTxPath(Adapter);
		
		NICFreeQueuedSendPackets(Adapter);
		

		// Now that we are actually going into power down we can enable
		// wake events safely if needed (we could not do this earlier since 
		// it would disable normal reception).
		Status |= NICConfigureWakeEvents(Adapter);

		if(Status != NIC_STATUS_SUCCESS)
		{
			goto Exit_NICSetPowerState;
		}
	
		Status =NICReadRegister(Adapter, ETH_PMT_CTL, &dwTemp);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICSetPowerState;
		SMSC_TRACE1(DBG_POWER, "ETH_PMT_CTL= 0x%x \r\n",dwTemp);

		Status =NICReadRegister(Adapter, ETH_WUCSR, &dwTemp);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICSetPowerState;
		SMSC_TRACE1(DBG_POWER, "ETH_WUCSR= 0x%x \r\n",dwTemp);

		Status =NICReadRegister(Adapter, ETH_MAC_RX, &dwTemp);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICSetPowerState;
		SMSC_TRACE1(DBG_POWER, "ETH_MAC_RX= 0x%x \r\n",dwTemp);

		// Set Power state variable
		Adapter->CurrentPowerState = PowerState;
		
	}

	if ((Adapter->CurrentPowerState != 0) && (PowerState == 0))
	{
		SMSC_TRACE1(DBG_POWER, "We were in power down state D%d before... waking up\r\n", Adapter->CurrentPowerState);
		
		// Disable all WoL Patterns
		dwTemp=0UL;
		NICWriteRegister(Adapter, ETH_WUCSR2,&dwTemp);
        NICWriteRegister(Adapter, ETH_WUCSR,&dwTemp);
		// PHY_PWRUP bit in PMT_CTL should be set after resuming from SUSPEND2
			   Status = NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
			   if (Status == NIC_STATUS_SUCCESS)
			   {
				   if ((Value32 & ETH_PMT_CTL_SUS_MODE_) == ETH_PMT_CTL_SUS_MODE_2_)
				   {
					   DWORD   Count = 0, dwTemp32;
					   dwTemp32=ETH_PMT_CTL_PHY_PWRUP_;
					   NICWriteRegister(Adapter, ETH_PMT_CTL, &dwTemp32);
					   do
					   {
						   NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
						   if (!(Value32 & ETH_PMT_CTL_PHY_PWRUP_))
							   break;
							   
						   NdisMSleep(200); //200 usec
						   Count++;
					   } while (Count <= 50); // don't wait more than a 10 msec

					   // PHY is reset and has to be re-initialized
                	bPhyInit = TRUE;
                	SMSC_TRACE0(DBG_POWER,"Resuming from Suspend2, PHY powerup and init done.\r\n");
                	Adapter->ReinitStatCounter = TRUE;
					   
					}
			   }

		MP_SET_FLAG(Adapter, fMP_DISCONNECTED);

		Adapter->CurrentPowerState = 0; // If not NICIsCardPresent() will fail

		if(!NICIsCardPresent(Adapter))
		{
			Status = NIC_STATUS_FAILURE;
			Adapter->CurrentPowerState = OldPowerState;
			goto Exit_NICSetPowerState;
		}

		// Clear all flags
		dwTemp=0xF0UL;
		NICWriteRegister(Adapter, ETH_WUCSR2, &dwTemp);
        NICWriteRegister(Adapter, ETH_WUCSR, &dwTemp);

		Status = NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
				goto Exit_NICSetPowerState;
		SMSC_TRACE1(DBG_POWER, "ETH_PMT_CTL = %08lx\r\n", Value32);

		if(Value32 & ETH_PMT_CTL_WUPS_MASK_)
		{
			Adapter->WakeupGlobalEventCounter++;
			SMSC_TRACE0(DBG_POWER, "There was a wake event. The source was:\r\n");

			if(Value32 & ETH_PMT_CTL_WUPS_PHY_)
			{
				Adapter->WakeupLinkEventCounter++;
				Value32 = ETH_PMT_CTL_WUPS_PHY_;
				Status = NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICSetPowerState;
				SMSC_TRACE0(DBG_POWER, "  Link Change\r\n");
			}

			if(Value32 & ETH_PMT_CTL_WUPS_MAC_)
			{
				SMSC_TRACE0(DBG_POWER, "  WOL and/or MP\r\n");
				Status = NICReadRegister(Adapter, ETH_WUCSR, &Value32_2 );
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICSetPowerState;

				if(Value32_2 & ETH_WUCSR_WUFR_)
				{
					SMSC_TRACE0(DBG_POWER, "	  WOL\r\n");
					Adapter->WakeupWOLEventCounter++;
					Value32 = ETH_WUCSR_WUFR_;
					Status = NICWriteRegister(Adapter, ETH_WUCSR, &Value32);
					if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICSetPowerState;
				}

				if(Value32_2 & ETH_WUCSR_MPR_)
				{
					Adapter->WakeupMPEventCounter++;
					Value32 = ETH_WUCSR_MPR_;
					Status = NICWriteRegister(Adapter, ETH_WUCSR, &Value32);
					if (Status != NIC_STATUS_SUCCESS)
						goto Exit_NICSetPowerState;
					SMSC_TRACE0(DBG_POWER, "	  Magic Packet\r\n");
				}

				SMSC_TRACE1(DBG_POWER, "\nWUCSR = %08lx\r\n", Value32_2);
				Value32 = ETH_WUCSR, ETH_WUCSR_MPR_;
				Status = NICWriteRegister(Adapter, ETH_PMT_CTL, &Value32);
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICSetPowerState;
			}

			
			NICReadRegister(Adapter, ETH_PMT_CTL, &Value32);
			if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICSetPowerState;
			SMSC_TRACE1(DBG_POWER, "ETH_PMT_CTL = %08lx\r\n", Value32);
			

			SMSC_TRACE0(DBG_POWER, "\nWake up counters:\r\n");
			SMSC_TRACE1(DBG_POWER, " Global 	  = %08lx\r\n",Adapter->WakeupGlobalEventCounter);
			SMSC_TRACE1(DBG_POWER, " Magic Packet = %08lx\r\n",Adapter->WakeupMPEventCounter);
			SMSC_TRACE1(DBG_POWER, " Wake On LAN  = %08lx\r\n",Adapter->WakeupWOLEventCounter);
			SMSC_TRACE1(DBG_POWER, " Link change  = %08lx\r\n",Adapter->WakeupLinkEventCounter);
		}

		if (Adapter->SetFeatureRemoteWake) {
			dwErr = ClearOrSetFeature(
						Adapter->UsbFuncs,
						Adapter->hUsbDevice,
						NULL,
						NULL,
						USB_SEND_TO_DEVICE,
						USB_FEATURE_REMOTE_WAKEUP,
						0,
						4000,
						FALSE
						);
	
			if (dwErr != ERROR_SUCCESS)
				SMSC_TRACE1(DBG_POWER,"Clear feature Remote Wakeup Failed, dwErr= %d \r\n",dwErr);
		}

		
		if (bPhyInit == TRUE)
        {
            Status = NICResetAdapter(Adapter);
            if(Status != NIC_STATUS_SUCCESS)
            {
                SMSC_TRACE0(DBG_POWER,"Failed to Reset Adapter\r\n");
                Adapter->CurrentPowerState = OldPowerState;
                goto Exit_NICSetPowerState;
            }
        }
        else
        {
            Status = NICInitializeMedia(Adapter);
            
        }

		Status = NICStartTxPath(Adapter);
		if (Status != NIC_STATUS_SUCCESS)
		{
			Adapter->CurrentPowerState = OldPowerState;
			goto Exit_NICSetPowerState;
		}

		Status = NICStartRxPath(Adapter);
		if (Status != NIC_STATUS_SUCCESS)
		{
			Adapter->CurrentPowerState = OldPowerState;
			goto Exit_NICSetPowerState;
		}
		#ifdef USE_INT_PIPE
		Status = NICStartInterruptPipe(Adapter);
		if (Status != NIC_STATUS_SUCCESS)
		{
			Adapter->CurrentPowerState = OldPowerState;
			goto Exit_NICSetPowerState;
		}
		#endif

		// Start the periodic Link check.
		StartLinkCheck(Adapter);
	}
	

Exit_NICSetPowerState:
	SMSC_TRACE0(DBG_POWER, "<-- NICSetPowerState\r\n");

	return(Status);
}


/*++

Routine Description:

	This routine check to see whether this device supports MEF or not

Arguments:

		Adapter    Pointer to our adapter

Return Value:

	TRUE if it does, fALSE if it does not

--*/
BOOLEAN
NICSupportsMEF(
	PMP_ADAPTER Adapter)

{
	DWORD Temp32, Value32;
	NIC_STATUS Status = NIC_STATUS_FAILURE;

	SMSC_TRACE0(DBG_INIT,"--> NICSupportsMEF\r\n");

	// Read HW_CFG value
	Status = NICReadRegister(Adapter, ETH_HW_CFG, &Value32);


	// Attempt to set MEF in HW_CFG
	Temp32 = Value32| ETH_HW_CFG_MEF_;
	Status |= NICWriteRegister(Adapter, ETH_HW_CFG, &Temp32 );

	// Read back HW_CFG
	Status |= NICReadRegister(Adapter, ETH_HW_CFG, &Temp32);

	SMSC_TRACE1(DBG_INIT,"Device has %sMEF support\r\n", (Temp32 & ETH_HW_CFG_MEF_) ? " " : "NO ");


	// Restore original value
	Status |= NICWriteRegister(Adapter, ETH_HW_CFG, &Value32);

	SMSC_TRACE0(DBG_INIT,"<-- NICSupportsMEF\r\n");

	return (((Temp32 & ETH_HW_CFG_MEF_) == ETH_HW_CFG_MEF_) && (Status == NIC_STATUS_SUCCESS));
}

